Package: triplediff
Title: Triple-Difference Estimators
Version: 0.2.4
Authors@R: c(person("Marcelo", "Ortiz-Villavicencio", email = "marcelo.ortiz@emory.edu", role = c("aut", "cre")),
              person("Pedro H. C.", "Sant'Anna", email = "pedro.santanna@emory.edu", role = c("aut"))
            )
Description: Implements triple-difference (DDD) estimators for both average
              treatment effects and event-study parameters. Methods include regression
              adjustment, inverse-probability weighting, and doubly-robust estimators,
              all of which rely on a conditional DDD parallel-trends assumption and
              allow covariate adjustment across multiple pre- and post-treatment
              periods. The methodology is detailed in Ortiz-Villavicencio and
              Sant'Anna (2025) <doi:10.48550/arXiv.2505.09942>.
License: MIT + file LICENSE
Encoding: UTF-8
Depends: R (>= 4.3)
RoxygenNote: 7.3.3
Imports: BMisc (>= 1.4.9), data.table (>= 1.15.0), fastglm (>= 0.0.3),
        Matrix (>= 1.6.1), parallel (>= 1.4.0), Rcpp (>= 1.0.12)
LinkingTo: Rcpp (>= 1.0.12)
Suggests: testthat (>= 3.0.0)
Config/testthat/edition: 3
URL: http://marcelortiz.com/triplediff/
NeedsCompilation: yes
Packaged: 2026-06-13 20:03:54 UTC; marcelortiz
Author: Marcelo Ortiz-Villavicencio [aut, cre],
  Pedro H. C. Sant'Anna [aut]
Maintainer: Marcelo Ortiz-Villavicencio <marcelo.ortiz@emory.edu>
Repository: CRAN
Date/Publication: 2026-06-13 20:20:02 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-06-13 23:32:37 UTC; unix
Archs: triplediff.so.dSYM

Package: triplediff
Title: Triple-Difference Estimators
Version: 0.2.4
Authors@R: c(person("Marcelo", "Ortiz-Villavicencio", email = "marcelo.ortiz@emory.edu", role = c("aut", "cre")),
              person("Pedro H. C.", "Sant'Anna", email = "pedro.santanna@emory.edu", role = c("aut"))
            )
Description: Implements triple-difference (DDD) estimators for both average
              treatment effects and event-study parameters. Methods include regression
              adjustment, inverse-probability weighting, and doubly-robust estimators,
              all of which rely on a conditional DDD parallel-trends assumption and
              allow covariate adjustment across multiple pre- and post-treatment
              periods. The methodology is detailed in Ortiz-Villavicencio and
              Sant'Anna (2025) <doi:10.48550/arXiv.2505.09942>.
License: MIT + file LICENSE
Encoding: UTF-8
Depends: R (>= 4.3)
RoxygenNote: 7.3.3
Imports: BMisc (>= 1.4.9), data.table (>= 1.15.0), fastglm (>= 0.0.3),
        Matrix (>= 1.6.1), parallel (>= 1.4.0), Rcpp (>= 1.0.12)
LinkingTo: Rcpp (>= 1.0.12)
Suggests: testthat (>= 3.0.0)
Config/testthat/edition: 3
URL: http://marcelortiz.com/triplediff/
NeedsCompilation: yes
Packaged: 2026-06-13 20:03:54 UTC; marcelortiz
Author: Marcelo Ortiz-Villavicencio [aut, cre],
  Pedro H. C. Sant'Anna [aut]
Maintainer: Marcelo Ortiz-Villavicencio <marcelo.ortiz@emory.edu>
Repository: CRAN
Date/Publication: 2026-06-13 20:20:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(print,agg_ddd)
S3method(print,ddd)
S3method(summary,agg_ddd)
S3method(summary,ddd)
export(agg_ddd)
export(att_dr)
export(att_dr_rc)
export(compute_aggregation)
export(ddd)
export(gen_dgp_2periods)
export(gen_dgp_mult_periods)
export(process_attgt)
import(BMisc)
import(data.table)
import(fastglm)
import(stats)
importFrom(Matrix,Matrix)
importFrom(Rcpp,sourceCpp)
importFrom(stats,model.matrix)
importFrom(stats,update)
importFrom(utils,write.table)
useDynLib(triplediff)

# Triple Differences Estimators <img src="man/figures/triplediff-logo.png" align="right" alt="" width="155" />

![](https://img.shields.io/badge/release%20lifecycle-alpha-orange.svg)
[![](https://img.shields.io/badge/devel%20version-0.2.0-blue.svg)](https://github.com/marcelortizv/triplediff)
[![](https://img.shields.io/badge/doi-10.48550/arXiv.2505.09942-yellow.svg)](https://doi.org/10.48550/arXiv.2505.09942)

<!-- README.md is generated from README.Rmd. Please edit that file -->

**triplediff** is an R package for computing average treatment effects
in Triple Differences Designs (also known as
Difference-in-Differences-in-Differences or DDD). DDD designs are widely
used in empirical work to relax parallel trends assumptions in
Difference-in-Differences settings. This package provides functions to
estimate group-time average treatment effect and event-study type
estimands associated with DDD designs. The setups allowed are:

- Two-periods or Multiple Periods DDD.
- Single Treatment Date and Variation in Treatment Timing (i.e.,
  staggered adoption).
- Conditional and Unconditional DDD parallel trends assumptions.

The `triplediff` package implements the framework proposed in:

- [Ortiz-Villavicencio, Marcelo and Pedro H.C. Sant’Anna. "Better
  Understanding Triple Differences Estimators",
  2025.](https://arxiv.org/abs/2505.09942)

## Installation

You can install the latest *development* version of `triplediff` from
Github with:

``` r
# install.packages("devtools")
devtools::install_github("marcelortizv/triplediff")

library(triplediff)
```

## Quick Start: How to use `triplediff`

We provide a quick start example to illustrate the main features of the
package. The UX is designed to be similar to the `did` package, so users
familiar with it should feel comfortable using `triplediff`.

`ddd` is the main function for computing the DDD estimators proposed in
[Ortiz-Villavicencio and Sant’Anna
(2025)](https://arxiv.org/abs/2505.09942). At its core, `ddd` employs
regression adjustment, inverse probability weighing, and doubly robust
estimators that are valid under conditional DDD parallel trends.

To use `ddd`, the minimal data requirements include:

- `id`: Unit unique identifier (e.g., firm, individual, etc.). Its
  column name is associated with parameter `idname`.

- `period`: Time identifier (e.g., year, month, etc.). Its column name
  is associated with parameter `tname`.

- `outcome`: Outcome variable of interest. Its column name is associated
  with parameter `yname`.

- `state`: First period when treatment is enabled for a particular unit.
  It is a positive number for treated units and defines which group the
  unit belongs to. It takes value `0` or `Inf` for never-enabling units.
  Its column name is associated with parameter `gname`.

- `partition`: This is an indicator variable that is 1 for the units
  eligible for treatment and 0 otherwise. Its column name is associated
  with parameter `pname`.

The following are two simplified examples of how to use the `ddd`
function. The first one is for a two-period DDD setup, while the second
one is for a multiple-period DDD setup with staggered treatment
adoption.

#### Case: Two-Periods DDD with covariates

First, we simulate some data with a built-in function `gen_dgp_2periods`
that generates a two-period DDD setup with a single treatment date. This
function receives the number of units and the type of DGP design to
generate. The `gen_dgp_2periods` function returns a data frame with the
required columns for the `ddd` function with 4 covariates.

``` r
set.seed(1234) # Set seed for reproducibility
# Simulate data for a two-periods DDD setup
df <- gen_dgp_2periods(
  size = 5000, # Number of units
  dgp_type = 1 # Type of DGP design
)$data

head(df)
#> Key: <id>
#>       id state partition  time        y        cov1       cov2      cov3
#>    <int> <num>     <num> <int>    <num>       <num>      <num>     <num>
#> 1:     1     0         0     1 209.9152 -0.97080934 -1.1726958 2.3893945
#> 2:     1     0         0     2 417.5260 -0.97080934 -1.1726958 2.3893945
#> 3:     2     0         0     1 211.4919  0.02591115  0.2763066 0.1063123
#> 4:     2     0         0     2 420.3656  0.02591115  0.2763066 0.1063123
#> 5:     3     0         0     1 221.9431  0.97147321 -0.4292088 0.5012794
#> 6:     3     0         0     2 440.9623  0.97147321 -0.4292088 0.5012794
#>          cov4 cluster
#>         <num>   <int>
#> 1:  0.2174955      39
#> 2:  0.2174955      39
#> 3: -0.1922253      29
#> 4: -0.1922253      29
#> 5:  1.1027248      44
#> 6:  1.1027248      44
```

Now we can estimate the average treatment effect on the treated (ATT)
using the `ddd` function.

``` r
# Estimate the average treatment effect on the treated, ATT(2,2)

att_22 <- ddd(yname = "y", tname = "time", idname = "id", gname = "state",
             pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
             data = df, control_group = "nevertreated", est_method = "dr")

summary(att_22)
#>  Call:
#> ddd(yname = "y", tname = "time", idname = "id", gname = "state", 
#>     pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4, 
#>     data = df, control_group = "nevertreated", est_method = "dr")
#> =========================== DDD Summary ==============================
#>  DR-DDD estimation for the ATT: 
#>      ATT       Std. Error    Pr(>|t|)  [95% Ptwise. Conf. Band]              
#>     -0.0780       0.0828       0.3463      -0.2404       0.0843              
#> 
#>  Note: * indicates that the confidence interval does not contain zero.
#>  --------------------------- Data Info   -----------------------------
#>  Panel Data
#>  Outcome variable: y
#>  Qualification variable: partition
#>  No. of units at each subgroup:
#>    treated-and-eligible: 1232
#>    treated-but-ineligible: 1285
#>    eligible-but-untreated: 1256
#>    untreated-and-ineligible: 1227
#>  --------------------------- Algorithms ------------------------------
#>  Outcome Regression estimated using: OLS
#>  Propensity score estimated using: Maximum Likelihood
#>  --------------------------- Std. Errors  ----------------------------
#>  Level of significance:  0.05
#>  Analytical standard errors.
#>  Type of confidence band:  Pointwise Confidence Interval
#>  =====================================================================
#>  See Ortiz-Villavicencio and Sant'Anna (2025) for details.
```

#### Case: Multiple Periods DDD with staggered treatment adoption

Next, we can simulate some data with a built-in function
`gen_dgp_mult_periods` that generates a multiple-period DDD setup with
staggered treatment adoption. This function receives the number of units
and the type of DGP design to generate. The `gen_dgp_mult_periods`
function returns a data frame with the required columns for the `ddd`
function with 4 covariates.

``` r
set.seed(1234) # Set seed for reproducibility
# Simulate data for a multiple-period DDD setup with staggered treatment adoption
data <- gen_dgp_mult_periods(size = 500, dgp_type = 1)$data

head(data)
#> Key: <id>
#>       id state partition  time         y        cov1       cov2       cov3
#>    <int> <num>     <num> <int>     <num>       <num>      <num>      <num>
#> 1:     1     3         1     1 1371.1923 -0.97080934  1.3995704 1.48834130
#> 2:     1     3         1     2 1690.9157 -0.97080934  1.3995704 1.48834130
#> 3:     1     3         1     3 2034.5661 -0.97080934  1.3995704 1.48834130
#> 4:     2     2         1     1  934.0812  0.02591115 -0.9747527 0.01714001
#> 5:     2     2         1     2 1189.5495  0.02591115 -0.9747527 0.01714001
#> 6:     2     2         1     3 1444.1072  0.02591115 -0.9747527 0.01714001
#>          cov4 cluster
#>         <num>   <int>
#> 1:  0.3853346      27
#> 2:  0.3853346      27
#> 3:  0.3853346      27
#> 4: -0.7822000      45
#> 5: -0.7822000      45
#> 6: -0.7822000      45
```

Now we can estimate the group-time average treatment effect on the
treated using the `ddd` function.

``` r
# Estimate the group-time average treatment effect in DDD

att_gt <- ddd(yname = "y", tname = "time", idname = "id", gname = "state", 
              pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
              data = data, control_group = "nevertreated", 
              base_period = "universal", est_method = "dr")

summary(att_gt)
#>  Call:
#> ddd(yname = "y", tname = "time", idname = "id", gname = "state", 
#>     pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4, 
#>     data = data, control_group = "nevertreated", base_period = "universal", 
#>     est_method = "dr")
#> =========================== DDD Summary ==============================
#>  DR-DDD estimation for the ATT(g,t): 
#> Group Time  ATT(g,t)  Std. Error [95% Pointwise  Conf. Band]  
#>   2    1       0.0000         NA           NA            NA   
#>   2    2      10.4901     0.3613       9.7819       11.1983  *
#>   2    3      20.5986     0.3868      19.8405       21.3568  *
#>   3    1      -0.4593     0.3660      -1.1767        0.2580   
#>   3    2       0.0000         NA           NA            NA   
#>   3    3      24.5983     0.3333      23.9450       25.2515  *
#> 
#>  Note: * indicates that the confidence interval does not contain zero.
#>  --------------------------- Data Info   -----------------------------
#>  Panel Data
#>  Outcome variable: y
#>  Qualification variable: partition
#>  Control group: Never Treated
#>  No. of units per treatment group:
#>   Units enabling treatment at period 3: 198
#>   Units enabling treatment at period 2: 195
#>   Units never enabling treatment: 107
#>  --------------------------- Algorithms ------------------------------
#>  Outcome Regression estimated using: OLS
#>  Propensity score estimated using: Maximum Likelihood
#>  --------------------------- Std. Errors  ----------------------------
#>  Level of significance:  0.05
#>  Analytical standard errors.
#>  Type of confidence band:  Pointwise Confidence Interval
#>  =====================================================================
#>  See Ortiz-Villavicencio and Sant'Anna (2025) for details.
```

Then, we can aggregate the group-time average treatment effect on the
treated to obtain the event-study type estimates. The `agg_ddd` function
allows us to aggregate the results from the `ddd` function.

``` r
es_e <- agg_ddd(att_gt, type = "eventstudy")

summary(es_e)
#>  Call:
#> agg_ddd(ddd_obj = att_gt, type = "eventstudy")
#> ========================= DDD Aggregation ============================
#>  Overall summary of ATT's based on event-study/dynamic aggregation: 
#>      ATT    Std. Error     [ 95%  Conf. Int.]  
#>  19.0983        0.3213    18.4685     19.7281 *
#> 
#>  Event Study:
#>  Event time Estimate Std. Error [95% Ptwise.  Conf. Band]  
#>          -2  -0.4593     0.3660       -1.1767      0.2580  
#>          -1   0.0000         NA            NA          NA  
#>           0  17.5980     0.4095       16.7955     18.4005 *
#>           1  20.5986     0.3868       19.8405     21.3568 *
#> 
#>  Note: * indicates that the confidence interval does not contain zero.
#>  --------------------------- Data Info   -----------------------------
#>  Outcome variable: y
#>  Qualification variable: partition
#>  Control group:  Never Treated
#>  --------------------------- Std. Errors  ----------------------------
#>  Level of significance:  0.05
#>  Analytical standard errors.
#>  =====================================================================
#>  See Ortiz-Villavicencio and Sant'Anna (2025) for details.
```

From these event-study type estimates, we can plot the results using the
`ggplot` library to create crisp visualizations.

``` r
es_df <- with(es_e$aggte_ddd, {
    tibble::tibble(period   = egt,
                   estimate = att.egt,
                   ci_lower = estimate - crit.val.egt * se.egt,
                   ci_upper = estimate + crit.val.egt * se.egt)
  })


all_per <- sort(unique(es_df$period))

p <- ggplot(es_df, aes(period, estimate)) +
    geom_errorbar(aes(ymin = ci_lower, ymax = ci_upper),
                  width = .15,
                  position = position_dodge(width = .3)) +
    geom_point(size = 2.2,
               position = position_dodge(width = .3)) +
    geom_hline(yintercept = 0, linewidth = .3) +
    geom_vline(xintercept = -1, linetype = "dashed",
               linewidth = .3) +
    scale_x_continuous(
      breaks = all_per
    ) +
         labs(title = "",
         x = "Event Time",
         y = "Treatment Effect") +
    theme_bw(base_size = 12) +
    theme(plot.title   = element_text(hjust = .25),
          legend.position = "right")
  
  p
```

![](man/figures/README-unnamed-chunk-10-1.png)<!-- -->

Finally, we can also estimate the group-time average treatment effect
using a GMM-based estimator with not-yet-treated units as comparison
group. This is done by setting the `control_group` parameter to
`"notyettreated"` in the `ddd` function.

``` r
# do the same but using GMM-based notyettreated
att_gt_nyt <- ddd(yname = "y", tname = "time", idname = "id", gname = "state", 
              pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
              data = data, control_group = "notyettreated", 
              base_period = "universal", est_method = "dr")

summary(att_gt_nyt)
#>  Call:
#> ddd(yname = "y", tname = "time", idname = "id", gname = "state", 
#>     pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4, 
#>     data = data, control_group = "notyettreated", base_period = "universal", 
#>     est_method = "dr")
#> =========================== DDD Summary ==============================
#>  DR-DDD estimation for the ATT(g,t): 
#> Group Time  ATT(g,t)  Std. Error [95% Pointwise  Conf. Band]  
#>   2    1       0.0000         NA           NA            NA   
#>   2    2      10.1894     0.2759       9.6486       10.7302  *
#>   2    3      20.5986     0.3868      19.8405       21.3568  *
#>   3    1      -0.4593     0.3660      -1.1767        0.2580   
#>   3    2       0.0000         NA           NA            NA   
#>   3    3      24.5983     0.3333      23.9450       25.2515  *
#> 
#>  Note: * indicates that the confidence interval does not contain zero.
#>  --------------------------- Data Info   -----------------------------
#>  Panel Data
#>  Outcome variable: y
#>  Qualification variable: partition
#>  Control group: Not yet Treated (GMM-based)
#>  No. of units per treatment group:
#>   Units enabling treatment at period 3: 198
#>   Units enabling treatment at period 2: 195
#>   Units never enabling treatment: 107
#>  --------------------------- Algorithms ------------------------------
#>  Outcome Regression estimated using: OLS
#>  Propensity score estimated using: Maximum Likelihood
#>  --------------------------- Std. Errors  ----------------------------
#>  Level of significance:  0.05
#>  Analytical standard errors.
#>  Type of confidence band:  Pointwise Confidence Interval
#>  =====================================================================
#>  See Ortiz-Villavicencio and Sant'Anna (2025) for details.
```

From this result, we can observe the standard errors for $ATT(2,2)$ are
lower than the ones obtained with the `control_group = "nevertreated"`
option since we leverage the additional information from the
not-yet-treated units in the estimation.

- ⚠️ We recommend users to read our paper for further details on the
  GMM-based procedure and its difference with other estimators relying
  in not-yet-treated units as comparison group.

## Release Note 📢

The current version is released in an **alpha** stage: the core features
are implemented and made available so users can try them out and provide
feedback. Please, be aware that it remains under active development, so
long-term stability is not guaranteed—APIs may evolve, adding more
features and/or breaking changes can occur at any time, without a
deprecation cycle.

Use [GitHub Issues](https://github.com/marcelortizv/triplediff/issues)
to report bugs, request features or provide feedback.

### 🚀 Currently Supported

- ️✅ Two-period DDD with single treatment date.
- ️✅ Multiple-period DDD with single treatment date.
- ️✅ DDD with staggered treatment adoption (i.e., more than two periods
  and variation in treatment timing).
- ️✅ Aggregations procedures (e.g., event-study type estimates).
- ️✅ GMM-based estimations with not-yet-treated units as comparison
  group.
- ✅ Repeated cross-sectional data.
- ✅ Unbalanced panel data.

### ⚠️ Not Yet Supported

- 🔲 Built-in plotting capabilities for visualizing results.
  - This can be done easily by users. E.g., event-study type estimates
    can be plotted using `ggplot2`. See the quick start example.
- 🔲 User-specified data-driven models for estimations of nuisance
  parameters.

# Triple Differences Estimators <img src="man/figures/triplediff-logo.png" align="right" alt="" width="155" />

![](https://img.shields.io/badge/release%20lifecycle-alpha-orange.svg)
[![](https://img.shields.io/badge/devel%20version-0.2.3-blue.svg)](https://github.com/marcelortizv/triplediff)
[![](https://img.shields.io/badge/doi-10.48550/arXiv.2505.09942-yellow.svg)](https://doi.org/10.48550/arXiv.2505.09942)

<!-- README.md is generated from README.Rmd. Please edit that file -->

**triplediff** is an R package for computing average treatment effects
in Triple Differences Designs (also known as
Difference-in-Differences-in-Differences or DDD). DDD designs are widely
used in empirical work to relax parallel trends assumptions in
Difference-in-Differences settings. This package provides functions to
estimate group-time average treatment effect and event-study type
estimands associated with DDD designs. The setups allowed are:

- Two-periods or Multiple Periods DDD.
- Single Treatment Date and Variation in Treatment Timing (i.e.,
  staggered adoption).
- Conditional and Unconditional DDD parallel trends assumptions.

The `triplediff` package implements the framework proposed in:

- [Ortiz-Villavicencio, Marcelo and Pedro H.C. Sant’Anna. "Better
  Understanding Triple Differences Estimators",
  2025.](https://arxiv.org/abs/2505.09942)

## Installation

You can install the latest *development* version of `triplediff` from
Github with:

``` r
# install.packages("devtools")
devtools::install_github("marcelortizv/triplediff")

library(triplediff)
```

## Quick Start: How to use `triplediff`

We provide a quick start example to illustrate the main features of the
package. The UX is designed to be similar to the `did` package, so users
familiar with it should feel comfortable using `triplediff`.

`ddd` is the main function for computing the DDD estimators proposed in
[Ortiz-Villavicencio and Sant’Anna
(2025)](https://arxiv.org/abs/2505.09942). At its core, `ddd` employs
regression adjustment, inverse probability weighing, and doubly robust
estimators that are valid under conditional DDD parallel trends.

To use `ddd`, the minimal data requirements include:

- `id`: Unit unique identifier (e.g., firm, individual, etc.). Its
  column name is associated with parameter `idname`.

- `period`: Time identifier (e.g., year, month, etc.). Its column name
  is associated with parameter `tname`.

- `outcome`: Outcome variable of interest. Its column name is associated
  with parameter `yname`.

- `state`: First period when treatment is enabled for a particular unit.
  It is a positive number for treated units and defines which group the
  unit belongs to. It takes value `0` or `Inf` for never-enabling units.
  Its column name is associated with parameter `gname`.

- `partition`: This is an indicator variable that is 1 for the units
  eligible for treatment and 0 otherwise. Its column name is associated
  with parameter `pname`.

The following are two simplified examples of how to use the `ddd`
function. The first one is for a two-period DDD setup, while the second
one is for a multiple-period DDD setup with staggered treatment
adoption.

#### Case: Two-Periods DDD with covariates

First, we simulate some data with a built-in function `gen_dgp_2periods`
that generates a two-period DDD setup with a single treatment date. This
function receives the number of units and the type of DGP design to
generate. The `gen_dgp_2periods` function returns a data frame with the
required columns for the `ddd` function with 4 covariates.

``` r
set.seed(1234) # Set seed for reproducibility
# Simulate data for a two-periods DDD setup
df <- gen_dgp_2periods(
  size = 5000, # Number of units
  dgp_type = 1 # Type of DGP design
)$data

head(df)
#> Key: <id>
#>       id state partition  time        y        cov1       cov2      cov3
#>    <int> <num>     <num> <int>    <num>       <num>      <num>     <num>
#> 1:     1     0         0     1 209.9152 -0.97080934 -1.1726958 2.3893945
#> 2:     1     0         0     2 417.5260 -0.97080934 -1.1726958 2.3893945
#> 3:     2     0         0     1 211.4919  0.02591115  0.2763066 0.1063123
#> 4:     2     0         0     2 420.3656  0.02591115  0.2763066 0.1063123
#> 5:     3     0         0     1 221.9431  0.97147321 -0.4292088 0.5012794
#> 6:     3     0         0     2 440.9623  0.97147321 -0.4292088 0.5012794
#>          cov4 cluster
#>         <num>   <int>
#> 1:  0.2174955      39
#> 2:  0.2174955      39
#> 3: -0.1922253      29
#> 4: -0.1922253      29
#> 5:  1.1027248      44
#> 6:  1.1027248      44
```

Now we can estimate the average treatment effect on the treated (ATT)
using the `ddd` function.

``` r
# Estimate the average treatment effect on the treated, ATT(2,2)

att_22 <- ddd(yname = "y", tname = "time", idname = "id", gname = "state",
             pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
             data = df, control_group = "nevertreated", est_method = "dr")

summary(att_22)
#>  Call:
#> ddd(yname = "y", tname = "time", idname = "id", gname = "state", 
#>     pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4, 
#>     data = df, control_group = "nevertreated", est_method = "dr")
#> =========================== DDD Summary ==============================
#>  DR-DDD estimation for the ATT: 
#>      ATT       Std. Error    Pr(>|t|)  [95% Ptwise. Conf. Band]              
#>     -0.0780       0.0828       0.3463      -0.2404       0.0843              
#> 
#>  Note: * indicates that the confidence interval does not contain zero.
#>  --------------------------- Data Info   -----------------------------
#>  Panel Data
#>  Outcome variable: y
#>  Qualification variable: partition
#>  No. of units at each subgroup:
#>    treated-and-eligible: 1232
#>    treated-but-ineligible: 1285
#>    eligible-but-untreated: 1256
#>    untreated-and-ineligible: 1227
#>  --------------------------- Algorithms ------------------------------
#>  Outcome Regression estimated using: OLS
#>  Propensity score estimated using: Maximum Likelihood
#>  --------------------------- Std. Errors  ----------------------------
#>  Level of significance:  0.05
#>  Analytical standard errors.
#>  Type of confidence band:  Pointwise Confidence Interval
#>  =====================================================================
#>  See Ortiz-Villavicencio and Sant'Anna (2025) for details.
```

#### Case: Multiple Periods DDD with staggered treatment adoption

Next, we can simulate some data with a built-in function
`gen_dgp_mult_periods` that generates a multiple-period DDD setup with
staggered treatment adoption. This function receives the number of units
and the type of DGP design to generate. The `gen_dgp_mult_periods`
function returns a data frame with the required columns for the `ddd`
function with 4 covariates.

``` r
set.seed(1234) # Set seed for reproducibility
# Simulate data for a multiple-period DDD setup with staggered treatment adoption
data <- gen_dgp_mult_periods(size = 500, dgp_type = 1)$data

head(data)
#> Key: <id>
#>       id state partition  time         y        cov1       cov2       cov3
#>    <int> <num>     <num> <int>     <num>       <num>      <num>      <num>
#> 1:     1     3         1     1 1371.1923 -0.97080934  1.3995704 1.48834130
#> 2:     1     3         1     2 1690.9157 -0.97080934  1.3995704 1.48834130
#> 3:     1     3         1     3 2034.5661 -0.97080934  1.3995704 1.48834130
#> 4:     2     2         1     1  934.0812  0.02591115 -0.9747527 0.01714001
#> 5:     2     2         1     2 1189.5495  0.02591115 -0.9747527 0.01714001
#> 6:     2     2         1     3 1444.1072  0.02591115 -0.9747527 0.01714001
#>          cov4 cluster
#>         <num>   <int>
#> 1:  0.3853346      27
#> 2:  0.3853346      27
#> 3:  0.3853346      27
#> 4: -0.7822000      45
#> 5: -0.7822000      45
#> 6: -0.7822000      45
```

Now we can estimate the group-time average treatment effect on the
treated using the `ddd` function.

``` r
# Estimate the group-time average treatment effect in DDD

att_gt <- ddd(yname = "y", tname = "time", idname = "id", gname = "state", 
              pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
              data = data, control_group = "nevertreated", 
              base_period = "universal", est_method = "dr")

summary(att_gt)
#>  Call:
#> ddd(yname = "y", tname = "time", idname = "id", gname = "state", 
#>     pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4, 
#>     data = data, control_group = "nevertreated", base_period = "universal", 
#>     est_method = "dr")
#> =========================== DDD Summary ==============================
#>  DR-DDD estimation for the ATT(g,t): 
#> Group Time  ATT(g,t)  Std. Error [95% Pointwise  Conf. Band]  
#>   2    1       0.0000         NA           NA            NA   
#>   2    2      10.4901     0.3613       9.7819       11.1983  *
#>   2    3      20.5986     0.3868      19.8405       21.3568  *
#>   3    1      -0.4593     0.3660      -1.1767        0.2580   
#>   3    2       0.0000         NA           NA            NA   
#>   3    3      24.5983     0.3333      23.9450       25.2515  *
#> 
#>  Note: * indicates that the confidence interval does not contain zero.
#>  --------------------------- Data Info   -----------------------------
#>  Panel Data
#>  Outcome variable: y
#>  Qualification variable: partition
#>  Control group: Never Treated
#>  No. of units per treatment group:
#>   Units enabling treatment at period 3: 198
#>   Units enabling treatment at period 2: 195
#>   Units never enabling treatment: 107
#>  --------------------------- Algorithms ------------------------------
#>  Outcome Regression estimated using: OLS
#>  Propensity score estimated using: Maximum Likelihood
#>  --------------------------- Std. Errors  ----------------------------
#>  Level of significance:  0.05
#>  Analytical standard errors.
#>  Type of confidence band:  Pointwise Confidence Interval
#>  =====================================================================
#>  See Ortiz-Villavicencio and Sant'Anna (2025) for details.
```

Then, we can aggregate the group-time average treatment effect on the
treated to obtain the event-study type estimates. The `agg_ddd` function
allows us to aggregate the results from the `ddd` function.

``` r
es_e <- agg_ddd(att_gt, type = "eventstudy")

summary(es_e)
#>  Call:
#> agg_ddd(ddd_obj = att_gt, type = "eventstudy")
#> ========================= DDD Aggregation ============================
#>  Overall summary of ATT's based on event-study/dynamic aggregation: 
#>      ATT    Std. Error     [ 95%  Conf. Int.]  
#>  19.0983        0.3213    18.4685     19.7281 *
#> 
#>  Event Study:
#>  Event time Estimate Std. Error [95% Ptwise.  Conf. Band]  
#>          -2  -0.4593     0.3660       -1.1767      0.2580  
#>          -1   0.0000         NA            NA          NA  
#>           0  17.5980     0.4095       16.7955     18.4005 *
#>           1  20.5986     0.3868       19.8405     21.3568 *
#> 
#>  Note: * indicates that the confidence interval does not contain zero.
#>  --------------------------- Data Info   -----------------------------
#>  Outcome variable: y
#>  Qualification variable: partition
#>  Control group:  Never Treated
#>  --------------------------- Std. Errors  ----------------------------
#>  Level of significance:  0.05
#>  Analytical standard errors.
#>  =====================================================================
#>  See Ortiz-Villavicencio and Sant'Anna (2025) for details.
```

From these event-study type estimates, we can plot the results using the
`ggplot` library to create crisp visualizations.

``` r
es_df <- with(es_e$aggte_ddd, {
    tibble::tibble(period   = egt,
                   estimate = att.egt,
                   ci_lower = estimate - crit.val.egt * se.egt,
                   ci_upper = estimate + crit.val.egt * se.egt)
  })


all_per <- sort(unique(es_df$period))

p <- ggplot(es_df, aes(period, estimate)) +
    geom_errorbar(aes(ymin = ci_lower, ymax = ci_upper),
                  width = .15,
                  position = position_dodge(width = .3)) +
    geom_point(size = 2.2,
               position = position_dodge(width = .3)) +
    geom_hline(yintercept = 0, linewidth = .3) +
    geom_vline(xintercept = -1, linetype = "dashed",
               linewidth = .3) +
    scale_x_continuous(
      breaks = all_per
    ) +
         labs(title = "",
         x = "Event Time",
         y = "Treatment Effect") +
    theme_bw(base_size = 12) +
    theme(plot.title   = element_text(hjust = .25),
          legend.position = "right")
  
  p
```

![](man/figures/README-unnamed-chunk-10-1.png)<!-- -->

Finally, we can also estimate the group-time average treatment effect
using a GMM-based estimator with not-yet-treated units as comparison
group. This is done by setting the `control_group` parameter to
`"notyettreated"` in the `ddd` function.

``` r
# do the same but using GMM-based notyettreated
att_gt_nyt <- ddd(yname = "y", tname = "time", idname = "id", gname = "state", 
              pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
              data = data, control_group = "notyettreated", 
              base_period = "universal", est_method = "dr")

summary(att_gt_nyt)
#>  Call:
#> ddd(yname = "y", tname = "time", idname = "id", gname = "state", 
#>     pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4, 
#>     data = data, control_group = "notyettreated", base_period = "universal", 
#>     est_method = "dr")
#> =========================== DDD Summary ==============================
#>  DR-DDD estimation for the ATT(g,t): 
#> Group Time  ATT(g,t)  Std. Error [95% Pointwise  Conf. Band]  
#>   2    1       0.0000         NA           NA            NA   
#>   2    2      10.1894     0.2759       9.6486       10.7302  *
#>   2    3      20.5986     0.3868      19.8405       21.3568  *
#>   3    1      -0.4593     0.3660      -1.1767        0.2580   
#>   3    2       0.0000         NA           NA            NA   
#>   3    3      24.5983     0.3333      23.9450       25.2515  *
#> 
#>  Note: * indicates that the confidence interval does not contain zero.
#>  --------------------------- Data Info   -----------------------------
#>  Panel Data
#>  Outcome variable: y
#>  Qualification variable: partition
#>  Control group: Not yet Treated (GMM-based)
#>  No. of units per treatment group:
#>   Units enabling treatment at period 3: 198
#>   Units enabling treatment at period 2: 195
#>   Units never enabling treatment: 107
#>  --------------------------- Algorithms ------------------------------
#>  Outcome Regression estimated using: OLS
#>  Propensity score estimated using: Maximum Likelihood
#>  --------------------------- Std. Errors  ----------------------------
#>  Level of significance:  0.05
#>  Analytical standard errors.
#>  Type of confidence band:  Pointwise Confidence Interval
#>  =====================================================================
#>  See Ortiz-Villavicencio and Sant'Anna (2025) for details.
```

From this result, we can observe the standard errors for $ATT(2,2)$ are
lower than the ones obtained with the `control_group = "nevertreated"`
option since we leverage the additional information from the
not-yet-treated units in the estimation.

- ⚠️ We recommend users to read our paper for further details on the
  GMM-based procedure and its difference with other estimators relying
  in not-yet-treated units as comparison group.

## Release Note 📢

The current version is released in an **alpha** stage: the core features
are implemented and made available so users can try them out and provide
feedback. Please, be aware that it remains under active development, so
long-term stability is not guaranteed—APIs may evolve, adding more
features and/or breaking changes can occur at any time, without a
deprecation cycle.

Use [GitHub Issues](https://github.com/marcelortizv/triplediff/issues)
to report bugs, request features or provide feedback.

### 🚀 Currently Supported

- ️✅ Two-period DDD with single treatment date.
- ️✅ Multiple-period DDD with single treatment date.
- ️✅ DDD with staggered treatment adoption (i.e., more than two periods
  and variation in treatment timing).
- ️✅ Aggregations procedures (e.g., event-study type estimates).
- ️✅ GMM-based estimations with not-yet-treated units as comparison
  group.
- ✅ Repeated cross-sectional data.
- ✅ Unbalanced panel data.

### ⚠️ Not Yet Supported

- 🔲 Built-in plotting capabilities for visualizing results.
  - This can be done easily by users. E.g., event-study type estimates
    can be plotted using `ggplot2`. See the quick start example.
- 🔲 User-specified data-driven models for estimations of nuisance
  parameters.

#' @useDynLib triplediff
#' @importFrom Rcpp sourceCpp
NULL
utils::globalVariables(c("first_treat", "treat", "tail", "period", "id", "subgroup", "partition",
                         "post", "deltaY", "y1", "y0", "y", "..cols_to_keep","asif_never_treated","treated_first_period",
                         "V1", "..tname", "..pname", "..idname", "..gname", "psi1", "psi2", "psi3", ".", ".row_id", "N", "control"))

#' Utility functions to compute aggregation procedures.
#' @importFrom stats update
#' @import data.table
#' @noRd
#--------------------------------------------------

# --------------------------------------
# FUNCTIONS FOR AGGREGATION PROCEDURES
# --------------------------------------

#' @title Compute extra term in influence function due to estimating weights
#'
#' @description A function to compute the extra term that shows up in the
#'  influence function for aggregated treatment effect parameters
#'  due to estimating the weights
#'
#' @param keepers a vector of indices for which group-time average
#'  treatment effects are used to compute a particular aggregated parameter
#' @param pg a vector with same length as total number of group-time average
#'  treatment effects that contains the probability of being in particular group
#' @param weights additional sampling weights (nx1)
#' @param G vector containing which group a unit belongs to (nx1)
#' @param group vector of groups
#'
#' @return nxk influence function matrix
#'
#' @keywords internal
get_weight_influence <- function(keepers, pg, weights, G, group) {
  # note: weights are all of the form P(G=g|cond)/sum_cond(P(G=g|cond))
  # this is equal to P(G=g)/sum_cond(P(G=g)) which simplifies things here

  # effect of estimating weights in the numerator
  if1 <- sapply(keepers, function(k) {
    (weights * BMisc::TorF(G==group[k]) - pg[k]) /
      sum(pg[keepers])
  })
  # effect of estimating weights in the denominator
  if2 <- base::rowSums( sapply( keepers, function(k) {
    weights * BMisc::TorF(G==group[k]) - pg[k]
  })) %*%
    t(pg[keepers]/(sum(pg[keepers])^2))

  # return the influence function for the weights
  return(if1 - if2)
}

#' @title Get an influence function for particular aggregate parameters
#'
#' @title This is a generic internal function for combining influence
#'  functions across ATT(g,t)'s to return an influence function for
#'  various aggregated treatment effect parameters.
#'
#' @param att vector of group-time average treatment effects
#' @param inf_func influence function for all group-time average treatment effects
#'  (matrix)
#' @param whichones which elements of att will be used to compute the aggregated
#'  treatment effect parameter
#' @param weights_agg the weights to apply to each element of att(whichones);
#'  should have the same dimension as att(whichones)
#' @param wif extra influence function term coming from estimating the weights;
#'  should be n x k matrix where k is dimension of whichones
#'
#' @return nx1 influence function
#'
#' @keywords internal
get_agg_inf_func <- function(att, inf_func, whichones, weights_agg, wif=NULL) {
  # enforce weights are in matrix form
  weights_agg <- as.matrix(weights_agg)

  # multiplies influence function times weights and sums to get vector of weighted IF (of length n)
  agg_inf_func <- inf_func[,whichones] %*% weights_agg

  # Incorporate influence function of the weights
  if (!is.null(wif)) {
    agg_inf_func <- agg_inf_func + wif%*%as.matrix(att[whichones])
  }

  # return influence function
  return(agg_inf_func)
}

#' @title Take influence function and compute standard errors
#'
#' @description Function to take an nx1 influence function and return
#'  a standard error
#'
#' @param influence_function An influence function
#' @param boot a boolean indicating whether bootstrapping was performed
#' @param boot_std_errors a vector of bootstrapped standard errors
#'
#' @return scalar standard error
#'
#' @keywords internal
compute_se_agg <- function(influence_function, boot=FALSE, boot_std_errors = NA,
                           cluster_vec = NULL) {

  n <- length(influence_function)

  # if we performed boot, boot_std_errors will be a vector of bootstrapped standard errors no null
  if (boot) {
    return(boot_std_errors)
  }
  # Analytical cluster-robust SE: cluster sums of the aggregated influence function,
  # matching the cluster-sum bootstrap (CS 2021, Remark 10). i.i.d. when no clusters.
  if (!is.null(cluster_vec) && length(cluster_vec) == n) {
    S <- rowsum(influence_function, cluster_vec)
    return(sqrt(sum(S^2)) / n)
  }
  return(sqrt( mean((influence_function)^2)/n ))
}

#' @title Compute Aggregated Treatment Effect Parameters
#'
#' @description Does the heavy lifting on computing aggregated group-time
#'  average treatment effects

#' @param ddd_obj a ddd object (i.e., the results of the [ddd()] function)
#' @param type Which type of aggregated treatment effect parameter to compute.
#'   \code{"simple"} just computes a weighted average of all
#'   group-time average treatment effects with weights proportional to group
#'   size.
#'   \code{"eventstudy"} computes average effects across
#'   different lengths of exposure to the treatment (event times). Here the overall effect averages the effect of the
#'   treatment across the positive lengths of exposure. This is the default option;
#'   \code{"group"} computes average treatment effects across different groups/cohorts; here
#'   the overall effect averages the effect across different groups using group size as weights;
#'   \code{"calendar"} computes average treatment effects across different
#'   time periods, with weights proportional to the group size; here the overall effect averages the effect across each
#'   time period.
#' @param cluster The name of the variable to be used for clustering. The maximum number of cluster variables is 1. Default is \code{NULL}.
#' @param balance_e If set (and if one computes event study), it balances
#'  the sample with respect to event time.  For example, if `balance_e=2`,
#'  `agg_ddd` will drop groups that are not exposed to treatment for
#'  at least three periods, the initial period `e=0` as well as the
#'  next two periods, `e=1` and `e=2`.  This ensures that
#'  the composition of groups does not change when event time changes.
#' @param min_e For event studies, this is the smallest event time to compute
#'  dynamic effects for.  By default, `min_e = -Inf` so that effects at
#'  all lengths of exposure are computed.
#' @param max_e For event studies, this is the largest event time to compute
#'  dynamic effects for.  By default, `max_e = Inf` so that effects at
#'  all lengths of exposure are computed.
#' @param na.rm Logical value if we are to remove missing Values from analyses. Defaults is FALSE.
#' @param boot Boolean for whether or not to compute standard errors using
#'  the multiplier bootstrap.  If standard errors are clustered, then one
#'  must set `boot=TRUE`. Default is value set in the ddd object.  If `boot = FALSE`, then analytical
#'  standard errors are reported.
#' @param nboot The number of bootstrap iterations to use.  The default is the value set in the ddd object,
#'  and this is only applicable if `boot=TRUE`.
#' @param alpha The level of confidence for the confidence intervals.  The default is 0.05. Otherwise, it will
#' use the value set in the ddd object.
#' @param cband Boolean for whether or not to compute a uniform confidence
#'  band that covers all of the group-time average treatment effects
#'  with fixed probability `1 - alpha`.  In order to compute uniform confidence
#'  bands, `boot` must also be set to `TRUE`.  The default is
#'  the value set in the ddd object
#'
#' @return Aggregation object (list) of class [`agg_ddd`]
#'
#' @keywords internal
#'
#' @export
compute_aggregation <- function(ddd_obj,
                                type = "simple",
                                cluster = NULL,
                                balance_e = NULL,
                                min_e = -Inf,
                                max_e = Inf,
                                na.rm = FALSE,
                                boot = FALSE,
                                nboot = NULL,
                                cband = NULL,
                                alpha = 0.05){
  # check if the object is of class `ddd`
  if (!inherits(ddd_obj, "ddd")) {
    stop("Object must be of class `ddd`")
  }

  # check if the object is a multiple period object
  if (!(ddd_obj$argu$multiple_periods)) {
    stop("Object must be a multiple period object")
  }

  # validate types of aggregation
  if (!(type %in% c("simple", "eventstudy", "group", "calendar"))){
    stop("type must be one of 'simple', 'eventstudy', 'group' or 'calendar'")
  }

  # get parameters
  groups <- ddd_obj$groups
  periods <- ddd_obj$periods
  ATT <- ddd_obj$ATT
  inf_func_mat <- ddd_obj$inf_func_mat
  n <- ddd_obj$n
  #data <- ddd_obj$data
  tlist <- ddd_obj$tlist
  glist <- ddd_obj$glist
  dta <- ddd_obj$first_period_dta # data only for first period
  yname <- ddd_obj$argu$yname
  pname <- ddd_obj$argu$pname
  control_group <- ddd_obj$argu$control_group

  # overwriting parameters for multiplier bootstrap if needed:

  # for cluster variable
  if (is.null(cluster)){
    cluster <- ddd_obj$argu$cluster
  }

  # Resolve the per-unit cluster vector for aggregation-level clustering.
  # Reuse the aligned vector stored on the ddd object (built in att_gt()).
  agg_cluster_vec <- NULL
  if (!is.null(cluster) && length(cluster) > 0) {
    cv  <- ddd_obj$cluster_vector
    can_cluster <- !is.null(cv) && length(cv) == n &&
      identical(cluster, ddd_obj$cluster_var)
    if (can_cluster) {
      agg_cluster_vec <- cv
    } else {
      warning(paste0(
        "Clustered standard errors were requested in agg_ddd()/compute_aggregation() ",
        "(cluster = '", cluster, "'), but the needed cluster information is not available ",
        "on this ddd object",
        if (!is.null(ddd_obj$cluster_var)) {
          paste0(" (ddd() clustered on '", ddd_obj$cluster_var,
                 "', and aggregation cannot switch to a different cluster variable)")
        } else {
          " (cluster was not supplied to ddd(), so aggregation cannot introduce clustering)"
        },
        ". Reporting standard errors that do NOT account for clustering; re-run ddd() with ",
        "cluster = '", cluster, "' for clustered inference."))
      cluster <- NULL    # fall back to i.i.d. for BOTH analytic and bootstrap paths
    }
  }

  # for bootstrap Boolean
  if (is.null(boot)){
    boot <- ddd_obj$argu$boot
  }

  # for number of bootstrap iterations
  if (is.null(nboot)){
    nboot <- ddd_obj$argu$nboot
  }

  # for uniform confidence bands Boolean
  if (is.null(cband)){
    cband <- ddd_obj$argu$cband
  }

  # for alpha level of significance
  if (is.null(alpha)){
    alpha <- ddd_obj$argu$alpha
  }

  # this is useful for summary tables only
  new_argu <- list(cluster = cluster,
                   boot = boot,
                   nboot = nboot,
                   cband = cband,
                   alpha = alpha)

  # flag for boot and cband
  if ((!boot) && (cband)){
    stop("cband is only available when boot = TRUE")
  }

  # recreating a `did_preprocessed` object only with needed parameters to compute bootstrapped standard errors
  did_preprocessed <- list()
  did_preprocessed$preprocessed_data <- dta # we only require data from the first period
  did_preprocessed$cluster <- cluster # cluster variable
  did_preprocessed$nboot <- nboot # number of bootstrap iterations
  did_preprocessed$alpha <- alpha # level of significance
  did_preprocessed$panel <- ddd_obj$argu$panel # panel indicator
  did_preprocessed$allow_unbalanced_panel <- ddd_obj$argu$allow_unbalanced_panel # unbalanced panel indicator
  did_preprocessed$cluster_vector <- agg_cluster_vec # aligned per-unit cluster vector for mboot


  if((na.rm == FALSE) && base::anyNA(ATT)) stop("Missing values at att_gt found. If you want to remove these, set `na.rm = TRUE'.")

  # removing NA values from ATT(g,t), groups and periods objects
  if(na.rm){
    notna <- !is.na(ATT)
    groups <- groups[notna]
    periods <- periods[notna]
    ATT <- ATT[notna]
    inf_func_mat <- inf_func_mat[, notna]
    glist <- sort(unique(groups))

    # Ensure we have non-missing post-treatment ATTs for each group if the type is "group"
    if(type == "group"){
      # Get the groups that have some non-missing ATT(g,t) in post-treatmemt periods
      gnotna <- sapply(glist, function(g) {
        # look at post-treatment periods for group g
        whichg <- which((groups == g) & (g <= periods))
        attg <- ATT[whichg]
        group_select <- !is.na(mean(attg))
        return(group_select)
      })
      gnotna <- glist[gnotna]
      # indicator for not all post-treatment ATT(g,t) missing
      not_all_na <- groups %in% gnotna
      # Re-do the na.rm thing to update the groups
      groups <- groups[not_all_na]
      periods <- periods[not_all_na]
      ATT <- ATT[not_all_na]
      inf_func_mat <- inf_func_mat[, not_all_na]
      glist <- sort(unique(groups))
    }
  }

  #-----------------------------------------------------------------------------
  # data manipulation
  #-----------------------------------------------------------------------------

  orig_periods <- periods
  orig_group <- groups
  orig_glist <- glist
  orig_tlist <- tlist
  # In case g's are not part of tlist
  orig_gtlist <- sort(unique(c(orig_tlist, orig_glist)))
  uniquet <- seq(1,length(unique(orig_gtlist)))

  # function to switch from "new" t values to  original t values
  t2orig <- function(t) {
    origt <- unique(c(orig_gtlist,0))[which(c(uniquet,0)==t)]
    return(origt)
  }

  # function to switch between "original" t values and new t values
  orig2t <- function(orig) {
    new_t <- c(uniquet,0)[which(unique(c(orig_gtlist,0))==orig)]
    out <- ifelse(length(new_t) == 0, NA, new_t)
    return(out)
  }

  t <- sapply(orig_periods, orig2t)
  group <- sapply(orig_group, orig2t)
  glist <- sapply(orig_glist, orig2t)
  tlist <- unique(t)
  maxT <- max(t)

  weights = dta[["weights"]]

  # compute probability of belonging of each group in glist
  pg <- sapply(orig_glist, function(g) mean(weights*(dta[["first_treat"]]==g)))
  pgg <- pg

  # making pg have the same length of ATT(g,t)
  pg <- pg[match(group, glist)]

  # which group time average treatment effects are post-treatment
  keepers <- which(group <= t & t <= (group + max_e))

  # n x 1 vector of group variable
  G <-  unlist(lapply(dta[["first_treat"]], orig2t))

  #-----------------------------------------------------------------------------
  # Compute the simple ATT summary
  #-----------------------------------------------------------------------------

  if (type == "simple") {

    # simple att
    # averages all post-treatment ATT(g,t) with weights
    # given by group size
    simple.att <- sum(ATT[keepers]*pg[keepers])/(sum(pg[keepers]))
    if(is.nan(simple.att)) simple.att <- NA

    # get the part of the influence function coming from estimated weights
    simple.wif <- get_weight_influence(keepers, pg, weights, G, group)

    # get the overall influence function
    simple.if <- get_agg_inf_func(att=ATT,
                                  inf_func=inf_func_mat,
                                  whichones=keepers,
                                  weights_agg=pg[keepers]/sum(pg[keepers]),
                                  wif=simple.wif)
    # Make it as vector
    simple.if <- as.numeric(simple.if)
    # RUN MULTIPLIER BOOTSTRAP
    if (boot){
      simple_bres <- mboot(inf_func = simple.if, did_preprocessed = did_preprocessed)
      simple_boot_see <- simple_bres$se
    } else {
      simple_boot_see <- NA
    }

    # get standard errors from overall influence function
    simple.se <- compute_se_agg(simple.if, boot, simple_boot_see, cluster_vec = agg_cluster_vec)
    if(!is.na(simple.se)){
      if(simple.se <= sqrt(.Machine$double.eps)*10) simple.se <- NA
    }


    return(list(overall.att = simple.att,
                overall.se = simple.se,
                type = type,
                yname = yname,
                argu = new_argu,
                pname = pname,
                control_group = control_group,
                inf.function = list(simple.att = simple.if)))
  }

  #-----------------------------------------------------------------------------
  # Compute the group (i.e., selective) treatment timing estimators
  #-----------------------------------------------------------------------------

  if (type == "group") {

    # get group specific ATTs
    # note: there are no estimated weights here
    selective.att.g <- sapply(glist, function(g) {
      # look at post-treatment periods for group g
      whichg <- which( (group == g) & (g <= t) & (t <= (group + max_e))) # added last condition to allow for limit on longest period included in att
      attg <- ATT[whichg]
      mean(attg)
    })
    selective.att.g[is.nan(selective.att.g)] <- NA


    # get standard errors for each group specific ATT
    selective.se.inner <- lapply(glist, function(g) {
      whichg <- which( (group == g) & (g <= t) & (t <= (group + max_e)))  # added last condition to allow for limit on longest period included in att
      inf.func.g <- as.numeric(get_agg_inf_func(att=ATT,
                                                inf_func=inf_func_mat,
                                                whichones=whichg,
                                                weights_agg=pg[whichg]/sum(pg[whichg]),
                                                wif=NULL))
      if (boot){
        g_bres <- mboot(inf_func = inf.func.g, did_preprocessed = did_preprocessed)
        g_boot_se <- g_bres$se
      } else{
        g_boot_se <- NA
      }
      se.g <- compute_se_agg(inf.func.g, boot, g_boot_se, cluster_vec = agg_cluster_vec)

      list(inf.func=inf.func.g, se=se.g)
    })

    # recover standard errors separately by group
    selective.se.g <- unlist(BMisc::get_list_element(selective.se.inner, "se"))
    selective.se.g[selective.se.g <= sqrt(.Machine$double.eps)*10] <- NA

    # recover influence function separately by group
    selective.inf.func.g <- simplify2array(BMisc::get_list_element(selective.se.inner, "inf.func"))

    # use multiplier bootstrap (across groups) to get critical value
    # for constructing uniform confidence bands
    selective.crit.val <- stats::qnorm(1 - alpha/2)

    # GET CRITICAL VALUES FOR UNIFORM CONFIDENCE BANDS
    if (cband){
      # if we enter here, it's because we already perform bootstrap. Cannot allow cband without doing bootstrap
      selective.crit.val <- mboot(inf_func = selective.inf.func.g, did_preprocessed = did_preprocessed)$unif_crit_val

      if(is.na(selective.crit.val) | is.infinite(selective.crit.val)){
        warning('Simultaneous critical value is NA. This probably happened because we cannot compute t-statistic (std errors are NA). We then report pointwise conf. intervals.')
        selective.crit.val <- stats::qnorm(1 - alpha/2)
      }

      if(selective.crit.val < stats::qnorm(1 - alpha/2)){
        warning('Simultaneous conf. band is somehow smaller than pointwise one using normal approximation. Since this is unusual, we are reporting pointwise confidence intervals')
        selective.crit.val <- stats::qnorm(1 - alpha/2)
      }

      if(selective.crit.val >= 7){
        warning("Simultaneous critical value is arguably `too large' to be reliable. This usually happens when number of observations per group is small and/or there is no much variation in outcomes.")
      }
    }

    # get overall att under selective treatment timing
    # (here use pgg instead of pg because we can just look at each group)
    selective.att <- sum(selective.att.g * pgg)/sum(pgg)

    # account for having to estimate pgg in the influence function
    selective.wif <- get_weight_influence(keepers=1:length(glist),
                                          pg=pgg,
                                          weights=weights,
                                          G=G,
                                          group=group)

    # get overall influence function
    selective.inf.func <- get_agg_inf_func(att=selective.att.g,
                                           inf_func=selective.inf.func.g,
                                           whichones=(1:length(glist)),
                                           weights_agg=pgg/sum(pgg),
                                           wif=selective.wif)


    selective.inf.func <- as.numeric(selective.inf.func)
    # RUN MULTIPLIER BOOTSTRAP
    if (boot){
      selective_bres <- mboot(inf_func = selective.inf.func, did_preprocessed = did_preprocessed)
      selective_boot_se <- selective_bres$se
    } else{
      selective_boot_se <- NA
    }
    # get overall standard error
    selective.se <- compute_se_agg(selective.inf.func, boot, selective_boot_se, cluster_vec = agg_cluster_vec)
    if(!is.na(selective.se)){
      if((selective.se <= sqrt(.Machine$double.eps)*10)) selective.se <- NA
    }

    return(list(overall.att=selective.att,
                overall.se=selective.se,
                type=type,
                yname = yname,
                pname = pname,
                control_group = control_group,
                argu = new_argu,
                egt=orig_glist,
                att.egt=selective.att.g,
                se.egt=selective.se.g,
                crit.val.egt=selective.crit.val,
                inf.function = list(selective.inf.func.g = selective.inf.func.g,
                                    selective.inf.func = selective.inf.func)
    ))

  }

  #-----------------------------------------------------------------------------
  # calendar time effects
  #-----------------------------------------------------------------------------

  if (type == "calendar") {

    # drop time periods where no one is treated yet
    # (can't get treatment effects in those periods)
    minG <- min(group)
    calendar.tlist <- tlist[tlist>=minG]

    # calendar time specific atts
    calendar.att.t <- sapply(calendar.tlist, function(t1) {
      # look at post-treatment periods for group g
      whicht <- which( (t == t1) & (group <= t))
      attt <- ATT[whicht]
      pgt <- pg[whicht]/(sum(pg[whicht]))
      sum(pgt * attt)
    })

    # get standard errors and influence functions
    # for each time specific att
    calendar.se.inner <- lapply(calendar.tlist, function(t1) {
      whicht <- which( (t == t1) & (group <= t))
      pgt <- pg[whicht]/(sum(pg[whicht]))
      wif.t <- get_weight_influence(keepers=whicht,
                                    pg=pg,
                                    weights=weights,
                                    G=G,
                                    group=group)
      inf.func.t <- as.numeric(get_agg_inf_func(att=ATT,
                                                inf_func=inf_func_mat,
                                                whichones=whicht,
                                                weights_agg=pgt,
                                                wif=wif.t))
      if (boot){
        t_bres <- mboot(inf_func = inf.func.t, did_preprocessed = did_preprocessed)
        t_boot_se <- t_bres$se
      } else{
        t_boot_se <- NA
      }
      se.t <- compute_se_agg(inf.func.t, boot, t_boot_se, cluster_vec = agg_cluster_vec)
      list(inf.func=inf.func.t, se=se.t)
    })

    # recover standard errors separately by time
    calendar.se.t <- unlist(BMisc::get_list_element(calendar.se.inner, "se"))
    calendar.se.t[calendar.se.t <= sqrt(.Machine$double.eps)*10] <- NA
    # recover influence function separately by time
    calendar.inf.func.t <- simplify2array(BMisc::get_list_element(calendar.se.inner, "inf.func"))

    # use multiplier boostrap (across groups) to get critical value
    # for constructing uniform confidence bands
    calendar.crit.val <-  stats::qnorm(1-alpha/2)

    # GET CRITICAL VALUES FOR UNIFORM CONFIDENCE BANDS
    if (cband){
      # if we enter here, it's because we already perform bootstrap. Cannot allow cband without doing bootstrap
      calendar.crit.val <- mboot(inf_func = calendar.inf.func.t, did_preprocessed = did_preprocessed)$unif_crit_val

      if(is.na(calendar.crit.val) | is.infinite(calendar.crit.val)){
        warning('Simultaneous critical value is NA. This probably happened because we cannot compute t-statistic (std errors are NA). We then report pointwise conf. intervals.')
        calendar.crit.val <- stats::qnorm(1 - alpha/2)
      }

      if(calendar.crit.val < stats::qnorm(1 - alpha/2)){
        warning('Simultaneous conf. band is somehow smaller than pointwise one using normal approximation. Since this is unusual, we are reporting pointwise confidence intervals')
        calendar.crit.val <- stats::qnorm(1 - alpha/2)
      }

      if(calendar.crit.val >= 7){
        warning("Simultaneous critical value is arguably `too large' to be reliable. This usually happens when number of observations per group is small and/or there is no much variation in outcomes.")
      }
    }

    # get overall att under calendar time effects
    # this is just average over all time periods
    calendar.att <- mean(calendar.att.t)

    # get overall influence function
    calendar.inf.func <- get_agg_inf_func(att=calendar.att.t,
                                          inf_func=calendar.inf.func.t,
                                          whichones=(1:length(calendar.tlist)),
                                          weights_agg=rep(1/length(calendar.tlist), length(calendar.tlist)),
                                          wif=NULL)
    calendar.inf.func <- as.numeric(calendar.inf.func)
    # RUN MULTIPLIER BOOTSTRAP
    if (boot){
      calendar_bres <- mboot(inf_func = calendar.inf.func, did_preprocessed = did_preprocessed)
      calendar_boot_se <- calendar_bres$se
    } else {
      calendar_boot_se <- NA
    }

    # get overall standard error
    calendar.se <- compute_se_agg(calendar.inf.func, boot, calendar_boot_se, cluster_vec = agg_cluster_vec)
    if(!is.na(calendar.se)){
      if (calendar.se <= sqrt(.Machine$double.eps)*10) calendar.se <- NA
    }
    return(list(overall.att=calendar.att,
                overall.se=calendar.se,
                type=type,
                yname = yname,
                pname = pname,
                control_group = control_group,
                argu = new_argu,
                egt=sapply(calendar.tlist,t2orig),
                att.egt=calendar.att.t,
                se.egt=calendar.se.t,
                crit.val.egt=calendar.crit.val,
                inf.function = list(calendar.inf.func.t = calendar.inf.func.t,
                                    calendar.inf.func = calendar.inf.func)
    ))
  }

  #-----------------------------------------------------------------------------
  # Compute the event-study estimators
  #-----------------------------------------------------------------------------

  if (type == "eventstudy") {

    # note: event times can be negative here.
    # note: event time = 0 corresponds to "on impact"
    # event times
    eseq <- unique(orig_periods - orig_group)
    eseq <- eseq[order(eseq)]

    # if the user specifies balance_e, then we are going to
    # drop some event times and some groups; if not, we just
    # keep everything (that is what this variable is for)
    include.balanced.gt <- rep(TRUE, length(orig_group))

    # if we balance the sample with respect to event time
    if (!is.null(balance_e)) {
      include.balanced.gt <- (t2orig(maxT) - orig_group >= balance_e)

      eseq <- unique(orig_periods[include.balanced.gt] - orig_group[include.balanced.gt])
      eseq <- eseq[order(eseq)]

      eseq <- eseq[ (eseq <= balance_e) & (eseq >= balance_e - t2orig(maxT) + t2orig(1))]

    }

    # only looks at some event times
    eseq <- eseq[ (eseq >= min_e) & (eseq <= max_e) ]

    # compute atts that are specific to each event time
    dynamic.att.e <- sapply(eseq, function(e) {
      # keep att(g,t) for the right g&t as well as ones that
      # are not trimmed out from balancing the sample
      whiche <- which( (orig_periods - orig_group == e) & (include.balanced.gt) )
      atte <- ATT[whiche]
      pge <- pg[whiche]/(sum(pg[whiche]))
      sum(atte*pge)
    })

    # compute standard errors for dynamic effects
    dynamic.se.inner <- lapply(eseq, function(e) {
      whiche <- which( (orig_periods - orig_group == e) & (include.balanced.gt) )
      pge <- pg[whiche]/(sum(pg[whiche]))
      wif.e <- get_weight_influence(whiche, pg, weights, G, group)
      inf.func.e <- as.numeric(get_agg_inf_func(att=ATT,
                                                inf_func=inf_func_mat,
                                                whichones=whiche,
                                                weights_agg=pge,
                                                wif=wif.e))
      if (boot){
        e_bres <- mboot(inf_func = inf.func.e, did_preprocessed = did_preprocessed)
        e_boot_se <- e_bres$se
      } else{
        e_boot_se <- NA
      }
      se.e <- compute_se_agg(inf.func.e, boot, e_boot_se, cluster_vec = agg_cluster_vec)
      list(inf.func=inf.func.e, se=se.e)
    })

    dynamic.se.e <- unlist(BMisc::get_list_element(dynamic.se.inner, "se"))
    dynamic.se.e[dynamic.se.e <= sqrt(.Machine$double.eps)*10] <- NA

    dynamic.inf.func.e <- simplify2array(BMisc::get_list_element(dynamic.se.inner, "inf.func"))

    dynamic.crit.val <- stats::qnorm(1 - alpha/2)

    # GET CRITICAL VALUES FOR UNIFORM CONFIDENCE BANDS
    if (cband){
      # if we enter here, it's because we already perform bootstrap. Cannot allow cband without doing bootstrap
      dynamic.crit.val <- mboot(inf_func = dynamic.inf.func.e, did_preprocessed = did_preprocessed)$unif_crit_val

      if(is.na(dynamic.crit.val) | is.infinite(dynamic.crit.val)){
        warning('Simultaneous critical value is NA. This probably happened because we cannot compute t-statistic (std errors are NA). We then report pointwise conf. intervals.')
        dynamic.crit.val <- stats::qnorm(1 - alpha/2)
      }

      if(dynamic.crit.val < stats::qnorm(1 - alpha/2)){
        warning('Simultaneous conf. band is somehow smaller than pointwise one using normal approximation. Since this is unusual, we are reporting pointwise confidence intervals')
        dynamic.crit.val <- stats::qnorm(1 - alpha/2)
      }

      if(dynamic.crit.val >= 7){
        warning("Simultaneous critical value is arguably `too large' to be reliable. This usually happens when number of observations per group is small and/or there is no much variation in outcomes.")
      }
    }


    # get overall average treatment effect
    # by averaging over positive dynamics
    epos <- eseq >= 0
    dynamic.att <- mean(dynamic.att.e[epos])
    dynamic.inf.func <- get_agg_inf_func(att=dynamic.att.e[epos],
                                         inf_func=as.matrix(dynamic.inf.func.e[,epos]),
                                         whichones=(1:sum(epos)),
                                         weights_agg=(rep(1/sum(epos), sum(epos))),
                                         wif=NULL)

    dynamic.inf.func <- as.numeric(dynamic.inf.func)
    # RUN MULTIPLIER BOOTSTRAP
    if (boot){
      dynamic_bres <- mboot(inf_func = dynamic.inf.func, did_preprocessed = did_preprocessed)
      dynamic_boot_se <- dynamic_bres$se
    } else {
      dynamic_boot_se <- NA
    }
    dynamic.se <- compute_se_agg(dynamic.inf.func, boot, dynamic_boot_se, cluster_vec = agg_cluster_vec)
    if(!is.na(dynamic.se)){
      if (dynamic.se <= sqrt(.Machine$double.eps)*10) dynamic.se <- NA
    }

    return(list(overall.att=dynamic.att,
                overall.se=dynamic.se,
                type=type,
                yname = yname,
                pname = pname,
                control_group = control_group,
                argu = new_argu,
                egt=eseq,
                att.egt=dynamic.att.e,
                se.egt=dynamic.se.e,
                crit.val.egt=dynamic.crit.val,
                inf.function = list(dynamic.inf.func.e = dynamic.inf.func.e,
                                    dynamic.inf.func = dynamic.inf.func),
                min_e=min_e,
                max_e=max_e,
                balance_e=balance_e
    ))
  }

}

# =============================================================================
# Tests for analytic cluster-robust SEs in agg_ddd()/compute_aggregation() and
# the warn + i.i.d.-fallback behaviour when clustering is requested on an object
# that was NOT clustered on the requested variable.
#
# Note on reachability:
#   An earlier revision of run_preprocess_multPeriods() forced boot = TRUE (and
#   cband = TRUE) for ddd(..., cluster = "cl", boot = FALSE); that override has
#   since been REMOVED. The aggregation-level analytic SEs are exercised by
#   calling agg_ddd(..., boot = FALSE, cband = FALSE), which recomputes SEs from
#   the stored influence function independently of the ddd object's boot flag.
# =============================================================================

make_mult_data_with_clusters <- function(seed = 123, size = 500, n_clusters = 25,
                                         add_other = FALSE) {
  set.seed(seed)
  d <- gen_dgp_mult_periods(size = size, dgp_type = 1)[["data"]]
  ids <- sort(unique(d$id))
  clmap <- data.frame(id = ids, cl = rep(seq_len(n_clusters), length.out = length(ids)))
  if (add_other) clmap$other <- rep(seq_len(5), length.out = length(ids))
  merge(d, clmap, by = "id")
}

fit_ddd <- function(d, cluster = NULL) {
  suppressWarnings(ddd(
    yname = "y", tname = "time", idname = "id", gname = "state",
    pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
    data = d, control_group = "nevertreated", base_period = "varying",
    est_method = "dr", cluster = cluster, boot = FALSE))
}

# -----------------------------------------------------------------------------
# CHECK 3-of-aggregation flavour: analytic clustered aggregation SE equals the
# manual cluster-sum CRVE of the aggregated influence function, for every
# aggregation type. (This is the aggregation-level analogue of (g,t) check 2.)
# -----------------------------------------------------------------------------
test_that("agg_ddd analytic clustered SE == manual cluster-sum CRVE (all types)", {
  d <- make_mult_data_with_clusters(seed = 123, size = 500, n_clusters = 25)
  obj <- fit_ddd(d, cluster = "cl")
  cv  <- obj$cluster_vector
  expect_equal(obj$cluster_var, "cl")
  expect_false(is.null(cv))

  manual_crve <- function(ifu) {
    ifu <- as.numeric(ifu)
    n <- length(ifu)
    sqrt(sum(rowsum(ifu, cv)^2)) / n
  }

  # simple
  a <- agg_ddd(obj, type = "simple", boot = FALSE, cband = FALSE)$aggte_ddd
  expect_equal(a$overall.se, manual_crve(a$inf.function$simple.att), tolerance = 1e-8)

  # eventstudy (overall + per-event)
  es <- agg_ddd(obj, type = "eventstudy", boot = FALSE, cband = FALSE)$aggte_ddd
  expect_equal(es$overall.se, manual_crve(es$inf.function$dynamic.inf.func), tolerance = 1e-8)
  ife <- es$inf.function$dynamic.inf.func.e
  man_e <- apply(as.matrix(ife), 2, manual_crve)
  ok_e <- !is.na(es$se.egt) & !is.na(man_e)
  expect_equal(as.numeric(es$se.egt[ok_e]), as.numeric(man_e[ok_e]), tolerance = 1e-8)

  # group
  gr <- agg_ddd(obj, type = "group", boot = FALSE, cband = FALSE)$aggte_ddd
  expect_equal(gr$overall.se, manual_crve(gr$inf.function$selective.inf.func), tolerance = 1e-8)

  # calendar
  ca <- agg_ddd(obj, type = "calendar", boot = FALSE, cband = FALSE)$aggte_ddd
  expect_equal(ca$overall.se, manual_crve(ca$inf.function$calendar.inf.func), tolerance = 1e-8)
})

# -----------------------------------------------------------------------------
# Companion: analytic i.i.d. aggregation SE (no clustering) equals the i.i.d.
# formula sqrt(mean(IF^2)/n) from the aggregated influence function.
# -----------------------------------------------------------------------------
test_that("agg_ddd analytic i.i.d. SE == sqrt(mean(IF^2)/n) (no clustering)", {
  d <- make_mult_data_with_clusters(seed = 321, size = 500, n_clusters = 25)
  obj <- fit_ddd(d, cluster = NULL)
  expect_true(is.null(obj$cluster_vector))

  manual_iid <- function(ifu) {
    ifu <- as.numeric(ifu); n <- length(ifu)
    sqrt(mean(ifu^2) / n)
  }

  a <- agg_ddd(obj, type = "simple", boot = FALSE, cband = FALSE)$aggte_ddd
  expect_equal(a$overall.se, manual_iid(a$inf.function$simple.att), tolerance = 1e-9)

  es <- agg_ddd(obj, type = "eventstudy", boot = FALSE, cband = FALSE)$aggte_ddd
  expect_equal(es$overall.se, manual_iid(es$inf.function$dynamic.inf.func), tolerance = 1e-9)
})

# -----------------------------------------------------------------------------
# CHECK 6(a): clustering requested on an object clustered on a DIFFERENT variable
# must warn and fall back to i.i.d. SEs (no error, no silent mis-report).
# -----------------------------------------------------------------------------
test_that("agg_ddd warns and falls back to i.i.d. when cluster var differs", {
  d <- make_mult_data_with_clusters(seed = 7, size = 500, n_clusters = 25, add_other = TRUE)
  obj <- fit_ddd(d, cluster = "cl")          # clustered on 'cl'
  expect_equal(obj$cluster_var, "cl")

  # requesting a DIFFERENT cluster variable must warn
  expect_warning(
    agg_ddd(obj, type = "simple", cluster = "other", boot = FALSE, cband = FALSE),
    regexp = "cannot switch|do NOT account|not available"
  )

  ovr <- suppressWarnings(
    agg_ddd(obj, type = "simple", cluster = "other", boot = FALSE, cband = FALSE)$aggte_ddd)

  # fallback SEs must equal the i.i.d. analytic formula (NOT a cluster-robust value)
  ifu <- as.numeric(ovr$inf.function$simple.att); n <- length(ifu)
  expect_equal(ovr$overall.se, sqrt(mean(ifu^2) / n), tolerance = 1e-9)

  # and the returned argu reflects the fallback to NULL clustering
  expect_true(is.null(ovr$argu$cluster))
})

# -----------------------------------------------------------------------------
# CHECK 6(b): clustering requested on an object built with cluster = NULL must
# warn and fall back to i.i.d. SEs; the fallback SEs equal the no-cluster SEs.
# -----------------------------------------------------------------------------
test_that("agg_ddd warns and falls back when ddd object was not clustered", {
  d <- make_mult_data_with_clusters(seed = 11, size = 500, n_clusters = 25)
  obj <- fit_ddd(d, cluster = NULL)          # NOT clustered
  expect_true(is.null(obj$cluster_var))

  expect_warning(
    agg_ddd(obj, type = "eventstudy", cluster = "cl", boot = FALSE, cband = FALSE),
    regexp = "not supplied|do NOT account|not available"
  )

  # i.i.d. baseline: no cluster requested at all
  base <- agg_ddd(obj, type = "eventstudy", boot = FALSE, cband = FALSE)$aggte_ddd
  ovr  <- suppressWarnings(
    agg_ddd(obj, type = "eventstudy", cluster = "cl", boot = FALSE, cband = FALSE)$aggte_ddd)

  expect_equal(ovr$overall.se, base$overall.se, tolerance = 1e-10)
  ok <- !is.na(base$se.egt) & !is.na(ovr$se.egt)
  expect_equal(as.numeric(ovr$se.egt[ok]), as.numeric(base$se.egt[ok]), tolerance = 1e-10)
})

# -----------------------------------------------------------------------------
# CHECK 1 (aggregation flavour): analytic clustered aggregation SE approx the
# clustered bootstrap aggregation SE (large nboot). Monte-Carlo tolerance.
# This path is reachable (aggregation honours its own boot flag), so it is run.
# -----------------------------------------------------------------------------
test_that("agg_ddd analytic clustered SE approx bootstrap clustered SE", {
  skip_on_cran()
  d <- make_mult_data_with_clusters(seed = 2025, size = 800, n_clusters = 40)
  obj <- fit_ddd(d, cluster = "cl")

  for (ty in c("simple", "eventstudy")) {
    a <- agg_ddd(obj, type = ty, boot = FALSE, cband = FALSE)$aggte_ddd
    set.seed(99)
    b <- suppressWarnings(
      agg_ddd(obj, type = ty, boot = TRUE, nboot = 5000, cband = FALSE)$aggte_ddd)
    rel <- abs(a$overall.se - b$overall.se) / a$overall.se
    expect_lt(rel, 0.10)
  }
})

#' Utility functions to compute nuisances parameters.
#' @import fastglm
#' @import data.table
#' @noRd
#--------------------------------------------------

# Helper function to get descriptive comparison name for a subgroup
# Subgroup encoding:
#   4 = Treated + Eligible (focal group)
#   3 = Treated + Ineligible
#   2 = Eligible + Untreated
#   1 = Untreated + Ineligible
get_comparison_description <- function(condition_subgroup) {
  switch(as.character(condition_subgroup),
    "3" = "Treated-Eligible vs Treated-Ineligible",
    "2" = "Treated-Eligible vs Eligible-Untreated",
    "1" = "Treated-Eligible vs Untreated-Ineligible"
  )
}

# Helper function to check overlap and generate descriptive warning messages
# Returns NULL if overlap is acceptable, otherwise returns warning message
check_overlap_and_get_warning <- function(propensity_scores, condition_subgroup,
                                          threshold = 1e-3, max_proportion = 0.05) {
  # Calculate overlap diagnostics
  n_total <- length(propensity_scores)
  n_below_threshold <- sum(propensity_scores < threshold)
  prop_below <- n_below_threshold / n_total

  # Warn if more than max_proportion of units have ps below threshold
  if (prop_below <= max_proportion) {
    return(NULL)
  }

  # Build comparison description
  comparison_desc <- get_comparison_description(condition_subgroup)

  # Build informative warning message
 msg <- paste0(
    "Poor propensity score overlap detected.\n",
    "  Comparison: ", comparison_desc, " units.\n",
    "  Diagnostics: ", round(prop_below * 100, 1), "% of units have propensity score < ", threshold, ".\n",
    "  Consider checking covariate balance or using fewer/different covariates."
  )

  return(msg)
}

# Function to compute propensity scores using fastglm for multiple subgroups
compute_pscore <- function(data, condition_subgroup, xformula, trim_level = 0.995) {
  # Subset data for condition_subgroup and subgroup == 4 or the given condition_subgroup
  condition_data <- data[data$subgroup %in% c(condition_subgroup, 4)]
  # Adding treatment variable P(1{PA4 = 4}|X)
  condition_data[, "PA4" := ifelse(condition_data$subgroup == 4, 1, 0)]
  uid_condition_data <- unique(condition_data, by = "id")

  # Get covariate matrix and treatment indicator
  covX <- stats::model.matrix(xformula, data = uid_condition_data)
  PA4 <- uid_condition_data$PA4
  i_weights <- uid_condition_data$weights
  n <- nrow(uid_condition_data)

  # Fit logistic regression model using fastglm
  model <- suppressWarnings(
    fastglm::fastglm(
      x         = covX,
      y         = PA4,
      family    = stats::binomial(),
      weights   = i_weights,
      intercept = FALSE,
      method    = 3
    )
  )

  # Flag for convergence of glm
  if (!model$converged) {
    warning(paste0(
      "Propensity score model did not converge.\n",
      "  Comparison: ", get_comparison_description(condition_subgroup), " units.\n",
      "  Consider using fewer covariates or checking for separation issues."
    ))
  }

  # Flag for weird coefficients
  if(anyNA(model$coefficients)) {
    stop(paste0(
      "Propensity score model has NA coefficients.\n",
      "  Comparison: ", get_comparison_description(condition_subgroup), " units.\n",
      "  This is likely due to multicollinearity among covariates."
    ))
  }

  # Compute propensity scores using fitted values
  propensity_scores <- fitted(model)

  # Warning for overlap condition (proportion-based check)
  overlap_warning <- check_overlap_and_get_warning(propensity_scores, condition_subgroup)
  if (!is.null(overlap_warning)) {
    warning(overlap_warning)
  }

  # Avoid divide by zero (following DRDID approach)
  propensity_scores <- pmin(propensity_scores, 1 - 1e-6)

  # Trimming indicator (following DRDID approach)
  # - Treated units (PA4=1): always included (trim.ps = TRUE)
  # - Control units (PA4=0): excluded if pscore >= trim_level
  keep_ps <- (propensity_scores < 1.01)  # Always TRUE initially
  keep_ps[PA4 == 0] <- (propensity_scores[PA4 == 0] < trim_level)

  # Compute Hessian matrix manually (following DRDID approach)
  # Hessian = chol2inv(chol(t(X) %*% (W * X))) * n
  # where W = ps * (1 - ps) * weights
  W <- propensity_scores * (1 - propensity_scores) * i_weights
  hessian_matrix <- chol2inv(chol(t(covX) %*% (W * covX))) * n

  return(list(propensity_scores = propensity_scores, hessian_matrix = hessian_matrix, keep_ps = keep_ps))
}

compute_pscore_null <- function(data, condition_subgroup) {
  # This is valid when REG is used
  # Subset data for condition_subgroup and subgroup == 4 or the given condition_subgroup
  condition_data <- data[data$subgroup %in% c(condition_subgroup, 4)]
  uid_condition_data <- unique(condition_data, by = "id")

  n <- nrow(uid_condition_data)
  propensity_scores <- rep(1, n)
  hessian_matrix <- NA
  # No trimming needed for REG method - all units included
  keep_ps <- rep(TRUE, n)

  return(list(propensity_scores = propensity_scores, hessian_matrix = hessian_matrix, keep_ps = keep_ps))
}

# Function to compute outcome regression for multiple subgroups
compute_outcome_regression <- function(data, condition_subgroup, xformula){
  # Subset data for condition_subgroup and subgroup == 4 or the given condition_subgroup
  condition_data <- data[data$subgroup %in% c(condition_subgroup, 4)]
  # Subset data for the control group (subgroup != 4)
  control_data <- condition_data[condition_data$subgroup == condition_subgroup]
  y1_control = control_data[control_data$post == 1, y]
  y0_control = control_data[control_data$post == 0, y]
  # generate deltaY for control group
  deltaY_control = y1_control - y0_control
  # get covariates including the intercept (conditioning in pre-treatment period since covariates are invariant)
  cov_control <- stats::model.matrix(as.formula(xformula), data = control_data[control_data$post == 0])
  # get weights (conditioning in pre-treatment period since weights are invariant)
  i_weights = control_data[control_data$post == 0, weights]

  # Fit outcome regression using fastglm
  reg.coeff <- stats::coef(
    fastglm::fastglm(
      x       = cov_control,
      y       = deltaY_control,
      weights = i_weights,
      family  = stats::gaussian()
    )
  )

  # Flag for NA coefficients
  if(anyNA(reg.coeff)) {
    stop(paste0(
      "Outcome regression model has NA coefficients.\n",
      "  Comparison: ", get_comparison_description(condition_subgroup), " units.\n",
      "  This is likely due to multicollinearity among covariates."
    ))
  }

  # compute regression adjustment
  y1 = condition_data[condition_data$post == 1, y]
  y0 = condition_data[condition_data$post == 0, y]
  deltaY = y1 - y0
  # get covariates including the intercept (conditioning in pre-treatment period since covariates are invariant)
  covX <- stats::model.matrix(xformula, data = condition_data[condition_data$post == 0])
  # compute OR delta
  or_delta = as.vector(tcrossprod(reg.coeff, as.matrix(covX)))
  # compute regression adjustment
  # reg_adjust = deltaY - or_delta
  return(list(deltaY = deltaY, or_delta = or_delta))
}

compute_outcome_regression_null <- function(data, condition_subgroup){
  # This is valid when IPW is used
  # Subset data for condition_subgroup and subgroup == 4 or the given condition_subgroup
  condition_data <- data[data$subgroup %in% c(condition_subgroup, 4)]

  # compute regression adjustment
  y1 = condition_data[condition_data$post == 1, y]
  y0 = condition_data[condition_data$post == 0, y]
  deltaY = y1 - y0

  # Filter the data for the pre-treatment period
  pre_treatment_data <- condition_data[condition_data$post == 0]
  # Create a zero vector with the same length as the number of rows in pre_treatment_data
  or_delta <- numeric(nrow(pre_treatment_data))

  return(list(deltaY = deltaY, or_delta = or_delta))
}

# Function to compute the average treatment effect for multiple subgroups
compute_did <- function(data, condition_subgroup, pscores, reg_adjustment, xformula, est_method){

  data <- unique(data, by = "id")
  condition_data <- data[data$subgroup %in% c(condition_subgroup, 4)]
  PA4 = ifelse(condition_data$subgroup == 4, 1, 0)
  PAa = ifelse(condition_data$subgroup == condition_subgroup, 1, 0)

  # Get propensity scores, hessian, trim indicator, and outcome regression
  if (condition_subgroup == 3) {
    pscore <- pscores[[1]]$propensity_scores
    hessian <- pscores[[1]]$hessian_matrix
    keep_ps <- pscores[[1]]$keep_ps
    deltaY <- reg_adjustment[[1]]$deltaY
    or_delta <- reg_adjustment[[1]]$or_delta

  } else if (condition_subgroup == 2) {
    pscore <- pscores[[2]]$propensity_scores
    hessian <- pscores[[2]]$hessian_matrix
    keep_ps <- pscores[[2]]$keep_ps
    deltaY <- reg_adjustment[[2]]$deltaY
    or_delta <- reg_adjustment[[2]]$or_delta

  } else if (condition_subgroup == 1) {
    pscore <- pscores[[3]]$propensity_scores
    hessian <- pscores[[3]]$hessian_matrix
    keep_ps <- pscores[[3]]$keep_ps
    deltaY <- reg_adjustment[[3]]$deltaY
    or_delta <- reg_adjustment[[3]]$or_delta

  } else {
    stop("Invalid condition_subgroup")
  }

  # get weights
  # i_weights = condition_data[condition_data$post == 0, weights]
  i_weights = condition_data$weights

  ################################
  # Get doubly-robust estimation #
  ################################
  # Apply trimming indicator to weights (following DRDID approach)
  # Control units with pscore >= trim_level get keep_ps = FALSE, so their weight becomes 0
  w_treat = keep_ps * i_weights * PA4
  if (est_method == "reg") {
    # Compute regression-based estimation (no IPW weights)
    w_control = keep_ps * (i_weights * PAa)
  } else {
    # IPW or DR: apply propensity score weighting
    w_control = keep_ps * (i_weights * pscore * PAa) / (1 - pscore)
  }
  riesz_treat = w_treat * (deltaY - or_delta)
  riesz_control = w_control * (deltaY - or_delta)
  att_treat = mean(riesz_treat, na.rm = TRUE) / mean(w_treat, na.rm = TRUE)
  att_control = mean(riesz_control, na.rm = TRUE) / mean(w_control, na.rm = TRUE)

  dr_att = att_treat - att_control

  ##########################
  # Get influence function #
  ##########################

  # Influence function related to the estimation of pscores
  covX = stats::model.matrix(xformula, data = condition_data)
  if (est_method == "reg") {
    inf_control_pscore <- 0
  } else {
    M2 <- base::colMeans(w_control * (deltaY - or_delta - att_control) * covX, na.rm = TRUE) # reg_adjust = deltaY - m_delta(x)
    score_ps <- i_weights * (PA4 - pscore) * covX
    # asymptotic linear representation of logit's beta
    score_ps_no_na <- na.omit(score_ps) # Exclude rows with NA values
    asy_lin_rep_ps <- score_ps_no_na %*% hessian
    inf_control_pscore <- asy_lin_rep_ps %*% as.matrix(M2)
  }

  if (est_method == "ipw") {
    inf_treat_or <- 0
    inf_cont_or <- 0
  } else {
    # Influence function related to the estimation of outcome model
    M1 <- base::colMeans(w_treat * covX, na.rm = TRUE)
    M3 <- base::colMeans(w_control * covX, na.rm = TRUE)

    # Influence function related to the estimation of regression model
    or_x <- i_weights * PAa * covX
    or_ex <- i_weights * PAa * (deltaY - or_delta) * covX
    XpX <- crossprod(or_x, covX)/nrow(condition_data)

    # Check if design matrix is singular (following DRDID approach)
    if (base::rcond(XpX) < .Machine$double.eps) {
      stop("The regression design matrix is singular. Consider removing some covariates.")
    }

    #asymptotic linear representation of the beta
    asy_linear_or <- t(solve(XpX, t(or_ex)))

    #or for treat
    inf_treat_or <- -asy_linear_or %*% M1
    #or for control
    inf_cont_or <- -asy_linear_or %*% M3
  }

  # Influence function from did
  inf_control_did <- riesz_control - w_control*att_control
  inf_treat_did <- riesz_treat - w_treat*att_treat

  # Influence function for the ATT
  inf_control <- (inf_control_did + inf_control_pscore + inf_cont_or)/mean(w_control)
  inf_treat <- (inf_treat_did + inf_treat_or)/mean(w_treat)
  # putting all together
  inf_func <- inf_treat - inf_control

  # fill zeros in influence function for observation outside the subgroup analyzed
  data[, "inf_func_result" := numeric(.N)]
  data[data$subgroup %in% c(condition_subgroup, 4), "inf_func_result" := inf_func]

  return(list(dr_att = dr_att, inf_func = data[["inf_func_result"]]))
}


#--------------------------------------------------
# Repeated Cross-Section Functions
#--------------------------------------------------

compute_pscore_rc <- function(data, condition_subgroup, xformula, trim_level = 0.995) {
  # Similar to compute_pscore but no unique(by="id") - uses all RCS data
  condition_data <- data[data$subgroup %in% c(condition_subgroup, 4)]
  condition_data[, "PA4" := ifelse(condition_data$subgroup == 4, 1, 0)]

  # Get covariate matrix and treatment indicator (all RCS observations)
  covX <- stats::model.matrix(xformula, data = condition_data)
  PA4 <- condition_data$PA4
  i_weights <- condition_data$weights
  n <- nrow(condition_data)

  # Fit logistic regression model using fastglm
  model <- suppressWarnings(
    fastglm::fastglm(
      x         = covX,
      y         = PA4,
      family    = stats::binomial(),
      weights   = i_weights,
      intercept = FALSE,
      method    = 3
    )
  )

  # Flag for convergence of glm
  if (!model$converged) {
    warning(paste0(
      "Propensity score model did not converge.\n",
      "  Comparison: ", get_comparison_description(condition_subgroup), " units.\n",
      "  Consider using fewer covariates or checking for separation issues."
    ))
  }

  # Flag for weird coefficients
  if(anyNA(model$coefficients)) {
    stop(paste0(
      "Propensity score model has NA coefficients.\n",
      "  Comparison: ", get_comparison_description(condition_subgroup), " units.\n",
      "  This is likely due to multicollinearity among covariates."
    ))
  }

  # Compute propensity scores using fitted values
  propensity_scores <- fitted(model)

  # Warning for overlap condition (proportion-based check)
  overlap_warning <- check_overlap_and_get_warning(propensity_scores, condition_subgroup)
  if (!is.null(overlap_warning)) {
    warning(overlap_warning)
  }

  # Avoid divide by zero (following DRDID approach)
  propensity_scores <- pmin(propensity_scores, 1 - 1e-6)

  # Trimming indicator (following DRDID approach)
  # - Treated units (PA4=1): always included (trim.ps = TRUE)
  # - Control units (PA4=0): excluded if pscore >= trim_level
  keep_ps <- (propensity_scores < 1.01)  # Always TRUE initially
  keep_ps[PA4 == 0] <- (propensity_scores[PA4 == 0] < trim_level)

  # Compute Hessian matrix manually (following DRDID approach)
  # Hessian = chol2inv(chol(t(X) %*% (W * X))) * n
  # where W = ps * (1 - ps) * weights
  W <- propensity_scores * (1 - propensity_scores) * i_weights
  hessian_matrix <- chol2inv(chol(t(covX) %*% (W * covX))) * n

  return(list(propensity_scores = propensity_scores, hessian_matrix = hessian_matrix, keep_ps = keep_ps))
}

compute_pscore_null_rc <- function(data, condition_subgroup) {
  condition_data <- data[data$subgroup %in% c(condition_subgroup, 4)]
  n <- nrow(condition_data)
  propensity_scores <- rep(1, n)
  hessian_matrix <- NA
  # No trimming needed for REG method - all units included
  keep_ps <- rep(TRUE, n)
  return(list(propensity_scores = propensity_scores, hessian_matrix = hessian_matrix, keep_ps = keep_ps))
}

compute_outcome_regression_rc <- function(data, condition_subgroup, xformula){
  # Subset data for condition_subgroup and subgroup == 4 or the given condition_subgroup
  condition_data <- data[data$subgroup %in% c(condition_subgroup, 4)]
  # compute outcome regression model for each of the four cells: (subgroup=4, post=1), (subgroup=4, post=0), (subgroup=condition_subgroup, post=1), (subgroup=condition_subgroup, post=0)
  fit_predict_mu <- function(subg, pst) {
     subset_data <- condition_data[condition_data$subgroup == subg & condition_data$post == pst]
     if(nrow(subset_data) == 0) return(rep(NA, nrow(condition_data)))

     # Covariates including intercept for fitting
     cov_fit <- stats::model.matrix(as.formula(xformula), data = subset_data)
     y_fit <- subset_data$y
     w_fit <- subset_data$weights

     # Fit using fastglm
     reg.coeff <- stats::coef(
       fastglm::fastglm(
         x       = cov_fit,
         y       = y_fit,
         weights = w_fit,
         family  = stats::gaussian()
       )
     )

     if(anyNA(reg.coeff)) {
       subg_desc <- if(subg == 4) "Treated-Eligible" else get_comparison_description(condition_subgroup)
       period_desc <- if(pst == 1) "post-treatment" else "pre-treatment"
       stop(paste0(
         "Outcome regression model has NA coefficients.\n",
         "  Cell: ", subg_desc, " units in ", period_desc, " period.\n",
         "  This is likely due to multicollinearity among covariates."
       ))
     }

     # Predict for ALL
     cov_all <- stats::model.matrix(as.formula(xformula), data = condition_data)
     as.vector(tcrossprod(reg.coeff, as.matrix(cov_all)))
  }

  or_trt_post <- fit_predict_mu(4, 1) # OR for: S=g, Q=1, T=post
  or_trt_pre <- fit_predict_mu(4, 0) # OR for: S=g, Q=1, T=pre
  or_ctrl_post <- fit_predict_mu(condition_subgroup, 1) # OR for: S=g', Q=q', T=post
  or_ctrl_pre <- fit_predict_mu(condition_subgroup, 0) # OR for: S=g', Q=q', T=pre


  return(list(or_trt_post=or_trt_post, or_trt_pre=or_trt_pre, or_ctrl_post=or_ctrl_post, or_ctrl_pre=or_ctrl_pre))
}

compute_outcome_regression_null_rc <- function(data, condition_subgroup){
  condition_data <- data[data$subgroup %in% c(condition_subgroup, 4)]
  n <- nrow(condition_data)
  return(list(or_trt_post=rep(0,n), or_trt_pre=rep(0,n), or_ctrl_post=rep(0,n), or_ctrl_pre=rep(0,n)))
}

compute_did_rc <- function(data, condition_subgroup, pscores, reg_adjustment, xformula, est_method){
  condition_data <- data[data$subgroup %in% c(condition_subgroup, 4)]

  if (condition_subgroup == 3) idx <- 1 else if (condition_subgroup == 2) idx <- 2 else idx <- 3

  pscore <- pscores[[idx]]$propensity_scores
  hessian <- pscores[[idx]]$hessian_matrix
  keep_ps <- pscores[[idx]]$keep_ps
  mu <- reg_adjustment[[idx]]

  # Basic variables
  n <- nrow(condition_data)
  post <- condition_data$post
  PA4 <- ifelse(condition_data$subgroup == 4, 1, 0) # PA4=1 for treated (subgroup 4), PA4=0 for control
  PAa <- ifelse(condition_data$subgroup == condition_subgroup, 1, 0) # PAa=1 for control subgroup
  y <- condition_data$y
  i_weights <- condition_data$weights

  # Get covariate matrix
  covX <- stats::model.matrix(xformula, data = condition_data)

  #############################################
  # Method-specific estimation
  #############################################

  if (est_method == "ipw") {
    #-----------------------------------------
    # IPW Estimator
    #-----------------------------------------
    # Riesz representers (with trimming applied following DRDID approach)
    riesz_treat_pre <- keep_ps * i_weights * PA4 * (1 - post)
    riesz_treat_post <- keep_ps * i_weights * PA4 * post
    riesz_control_pre <- keep_ps * i_weights * pscore * PAa * (1 - post) / (1 - pscore)
    riesz_control_post <- keep_ps * i_weights * pscore * PAa * post / (1 - pscore)

    # Elements of the influence function (summands)
    eta_treat_pre <- riesz_treat_pre * y / mean(riesz_treat_pre)
    eta_treat_post <- riesz_treat_post * y / mean(riesz_treat_post)
    eta_control_pre <- riesz_control_pre * y / mean(riesz_control_pre)
    eta_control_post <- riesz_control_post * y / mean(riesz_control_post)

    # Estimator of each component
    att_treat_pre <- mean(eta_treat_pre)
    att_treat_post <- mean(eta_treat_post)
    att_control_pre <- mean(eta_control_pre)
    att_control_post <- mean(eta_control_post)

    # ATT estimator
    att <- (att_treat_post - att_treat_pre) - (att_control_post - att_control_pre)

    # Influence function
    # Asymptotic linear representation of logit's beta's
    W <- pscore * (1 - pscore) * i_weights
    score_ps <- i_weights * (PA4 - pscore) * covX
    asy_lin_rep_ps <- score_ps %*% hessian

    # Leading term of the influence function
    inf_treat_pre <- eta_treat_pre - riesz_treat_pre * att_treat_pre / mean(riesz_treat_pre)
    inf_treat_post <- eta_treat_post - riesz_treat_post * att_treat_post / mean(riesz_treat_post)
    inf_treat <- inf_treat_post - inf_treat_pre

    inf_control_pre <- eta_control_pre - riesz_control_pre * att_control_pre / mean(riesz_control_pre)
    inf_control_post <- eta_control_post - riesz_control_post * att_control_post / mean(riesz_control_post)
    inf_control <- inf_control_post - inf_control_pre

    # Estimation effect from gamma hat (pscore)
    M2_pre <- base::colMeans(riesz_control_pre * (y - att_control_pre) * covX) / mean(riesz_control_pre)
    M2_post <- base::colMeans(riesz_control_post * (y - att_control_post) * covX) / mean(riesz_control_post)
    inf_control_ps <- asy_lin_rep_ps %*% (M2_post - M2_pre)

    # Final influence function
    inf_control <- inf_control + inf_control_ps
    att_inf_func <- inf_treat - inf_control

  } else if (est_method == "reg") {
    #-----------------------------------------
    # REG Estimator
    #-----------------------------------------
    or_ctrl_pre <- mu$or_ctrl_pre
    or_ctrl_post <- mu$or_ctrl_post

    # Riesz representers (with trimming applied following DRDID approach)
    riesz_treat_pre <- keep_ps * i_weights * PA4 * (1 - post)
    riesz_treat_post <- keep_ps * i_weights * PA4 * post
    riesz_control <- keep_ps * i_weights * PA4

    reg_att_treat_pre <- riesz_treat_pre * y
    reg_att_treat_post <- riesz_treat_post * y
    reg_att_control <- riesz_control * (or_ctrl_post - or_ctrl_pre)

    eta_treat_pre <- mean(reg_att_treat_pre) / mean(riesz_treat_pre)
    eta_treat_post <- mean(reg_att_treat_post) / mean(riesz_treat_post)
    eta_control <- mean(reg_att_control) / mean(riesz_control)

    att <- (eta_treat_post - eta_treat_pre) - eta_control

    # Influence function
    # Asymptotic linear representation of OLS parameters
    # Pre-period
    weights_ols_pre <- i_weights * PAa * (1 - post)
    wols_x_pre <- weights_ols_pre * covX
    wols_eX_pre <- weights_ols_pre * (y - or_ctrl_pre) * covX
    XpX_pre <- base::crossprod(wols_x_pre, covX) / n
    if (base::rcond(XpX_pre) < .Machine$double.eps) {
      stop("The regression design matrix for pre-treatment is singular. Consider removing some covariates.")
    }
    XpX_inv_pre <- solve(XpX_pre)
    asy_lin_rep_ols_pre <- wols_eX_pre %*% XpX_inv_pre

    # Post-period
    weights_ols_post <- i_weights * PAa * post
    wols_x_post <- weights_ols_post * covX
    wols_eX_post <- weights_ols_post * (y - or_ctrl_post) * covX
    XpX_post <- base::crossprod(wols_x_post, covX) / n
    if (base::rcond(XpX_post) < .Machine$double.eps) {
      stop("The regression design matrix for post-treatment is singular. Consider removing some covariates.")
    }
    XpX_inv_post <- solve(XpX_post)
    asy_lin_rep_ols_post <- wols_eX_post %*% XpX_inv_post

    # Leading term
    inf_treat_pre <- (reg_att_treat_pre - riesz_treat_pre * eta_treat_pre) / mean(riesz_treat_pre)
    inf_treat_post <- (reg_att_treat_post - riesz_treat_post * eta_treat_post) / mean(riesz_treat_post)
    inf_treat <- inf_treat_post - inf_treat_pre

    # Control component
    inf_control_1 <- (reg_att_control - riesz_control * eta_control)
    M1 <- base::colMeans(riesz_control * covX)
    inf_control_2_post <- asy_lin_rep_ols_post %*% M1
    inf_control_2_pre <- asy_lin_rep_ols_pre %*% M1
    inf_control <- (inf_control_1 + inf_control_2_post - inf_control_2_pre) / mean(riesz_control)

    att_inf_func <- inf_treat - inf_control

  } else if (est_method == "dr") {
    #-----------------------------------------
    # DR Estimator
    #-----------------------------------------
    or_ctrl_pre <- mu$or_ctrl_pre
    or_ctrl_post <- mu$or_ctrl_post
    or_ctrl <- post * or_ctrl_post + (1 - post) * or_ctrl_pre

    or_trt_pre <- mu$or_trt_pre
    or_trt_post <- mu$or_trt_post

    # Riesz representers (with trimming applied following DRDID approach)
    riesz_treat_pre <- keep_ps * i_weights * PA4 * (1 - post)
    riesz_treat_post <- keep_ps * i_weights * PA4 * post
    riesz_control_pre <- keep_ps * i_weights * pscore * PAa * (1 - post) / (1 - pscore)
    riesz_control_post <- keep_ps * i_weights * pscore * PAa * post / (1 - pscore)

    riesz_d <- keep_ps * i_weights * PA4
    riesz_dt1 <- keep_ps * i_weights * PA4 * post
    riesz_dt0 <- keep_ps * i_weights * PA4 * (1 - post)

    # Elements of the influence function (summands)
    eta_treat_pre <- riesz_treat_pre * (y - or_ctrl) / mean(riesz_treat_pre)
    eta_treat_post <- riesz_treat_post * (y - or_ctrl) / mean(riesz_treat_post)
    eta_control_pre <- riesz_control_pre * (y - or_ctrl) / mean(riesz_control_pre)
    eta_control_post <- riesz_control_post * (y - or_ctrl) / mean(riesz_control_post)

    # Extra elements for the locally efficient DRDID
    eta_d_post <- riesz_d * (or_trt_post - or_ctrl_post) / mean(riesz_d)
    eta_dt1_post <- riesz_dt1 * (or_trt_post - or_ctrl_post) / mean(riesz_dt1)
    eta_d_pre <- riesz_d * (or_trt_pre - or_ctrl_pre) / mean(riesz_d)
    eta_dt0_pre <- riesz_dt0 * (or_trt_pre - or_ctrl_pre) / mean(riesz_dt0)

    # Estimator of each component
    att_treat_pre <- mean(eta_treat_pre)
    att_treat_post <- mean(eta_treat_post)
    att_control_pre <- mean(eta_control_pre)
    att_control_post <- mean(eta_control_post)

    att_d_post <- mean(eta_d_post)
    att_dt1_post <- mean(eta_dt1_post)
    att_d_pre <- mean(eta_d_pre)
    att_dt0_pre <- mean(eta_dt0_pre)

    # ATT estimator
    att <- (att_treat_post - att_treat_pre) - (att_control_post - att_control_pre) +
      (att_d_post - att_dt1_post) - (att_d_pre - att_dt0_pre)

    # Influence function
    # Asymptotic linear representation of OLS parameters
    # Control pre-period
    weights_ols_pre <- i_weights * PAa * (1 - post)
    wols_x_pre <- weights_ols_pre * covX
    wols_eX_pre <- weights_ols_pre * (y - or_ctrl_pre) * covX
    XpX_pre <- base::crossprod(wols_x_pre, covX) / n
    if (base::rcond(XpX_pre) < .Machine$double.eps) {
      stop("The regression design matrix for control pre-treatment is singular. Consider removing some covariates.")
    }
    XpX_inv_pre <- solve(XpX_pre)
    asy_lin_rep_ols_pre <- wols_eX_pre %*% XpX_inv_pre

    # Control post-period
    weights_ols_post <- i_weights * PAa * post
    wols_x_post <- weights_ols_post * covX
    wols_eX_post <- weights_ols_post * (y - or_ctrl_post) * covX
    XpX_post <- base::crossprod(wols_x_post, covX) / n
    if (base::rcond(XpX_post) < .Machine$double.eps) {
      stop("The regression design matrix for control post-treatment is singular. Consider removing some covariates.")
    }
    XpX_inv_post <- solve(XpX_post)
    asy_lin_rep_ols_post <- wols_eX_post %*% XpX_inv_post

    # Treated pre-period
    weights_ols_pre_treat <- i_weights * PA4 * (1 - post)
    wols_x_pre_treat <- weights_ols_pre_treat * covX
    wols_eX_pre_treat <- weights_ols_pre_treat * (y - or_trt_pre) * covX
    XpX_pre_treat <- base::crossprod(wols_x_pre_treat, covX) / n
    if (base::rcond(XpX_pre_treat) < .Machine$double.eps) {
      stop("The regression design matrix for treated pre-treatment is singular. Consider removing some covariates.")
    }
    XpX_inv_pre_treat <- solve(XpX_pre_treat)
    asy_lin_rep_ols_pre_treat <- wols_eX_pre_treat %*% XpX_inv_pre_treat

    # Treated post-period
    weights_ols_post_treat <- i_weights * PA4 * post
    wols_x_post_treat <- weights_ols_post_treat * covX
    wols_eX_post_treat <- weights_ols_post_treat * (y - or_trt_post) * covX
    XpX_post_treat <- base::crossprod(wols_x_post_treat, covX) / n
    if (base::rcond(XpX_post_treat) < .Machine$double.eps) {
      stop("The regression design matrix for treated post-treatment is singular. Consider removing some covariates.")
    }
    XpX_inv_post_treat <- solve(XpX_post_treat)
    asy_lin_rep_ols_post_treat <- wols_eX_post_treat %*% XpX_inv_post_treat

    # Asymptotic linear representation of logit's beta's
    W <- pscore * (1 - pscore) * i_weights
    score_ps <- i_weights * (PA4 - pscore) * covX
    asy_lin_rep_ps <- score_ps %*% hessian

    # Influence function of the "treat" component
    inf_treat_pre <- eta_treat_pre - riesz_treat_pre * att_treat_pre / mean(riesz_treat_pre)
    inf_treat_post <- eta_treat_post - riesz_treat_post * att_treat_post / mean(riesz_treat_post)

    # Estimation effect from beta hat
    M1_post <- -base::colMeans(riesz_treat_post * post * covX) / mean(riesz_treat_post)
    M1_pre <- -base::colMeans(riesz_treat_pre * (1 - post) * covX) / mean(riesz_treat_pre)

    inf_treat_or_post <- asy_lin_rep_ols_post %*% M1_post
    inf_treat_or_pre <- asy_lin_rep_ols_pre %*% M1_pre

    # Influence function of control component
    inf_control_pre <- eta_control_pre - riesz_control_pre * att_control_pre / mean(riesz_control_pre)
    inf_control_post <- eta_control_post - riesz_control_post * att_control_post / mean(riesz_control_post)

    # Estimation effect from gamma hat (pscore)
    M2_pre <- base::colMeans(riesz_control_pre * (y - or_ctrl - att_control_pre) * covX) / mean(riesz_control_pre)
    M2_post <- base::colMeans(riesz_control_post * (y - or_ctrl - att_control_post) * covX) / mean(riesz_control_post)
    inf_control_ps <- asy_lin_rep_ps %*% (M2_post - M2_pre)

    # Estimation effect from beta hat
    M3_post <- -base::colMeans(riesz_control_post * post * covX) / mean(riesz_control_post)
    M3_pre <- -base::colMeans(riesz_control_pre * (1 - post) * covX) / mean(riesz_control_pre)

    inf_control_or_post <- asy_lin_rep_ols_post %*% M3_post
    inf_control_or_pre <- asy_lin_rep_ols_pre %*% M3_pre

    # Adjustment terms
    inf_eff1 <- eta_d_post - riesz_d * att_d_post / mean(riesz_d)
    inf_eff2 <- eta_dt1_post - riesz_dt1 * att_dt1_post / mean(riesz_dt1)
    inf_eff3 <- eta_d_pre - riesz_d * att_d_pre / mean(riesz_d)
    inf_eff4 <- eta_dt0_pre - riesz_dt0 * att_dt0_pre / mean(riesz_dt0)
    inf_eff <- (inf_eff1 - inf_eff2) - (inf_eff3 - inf_eff4)

    # Estimation effect of OR coefficients
    mom_post <- base::colMeans((riesz_d / mean(riesz_d) - riesz_dt1 / mean(riesz_dt1)) * covX)
    mom_pre <- base::colMeans((riesz_d / mean(riesz_d) - riesz_dt0 / mean(riesz_dt0)) * covX)
    inf_or_post <- (asy_lin_rep_ols_post_treat - asy_lin_rep_ols_post) %*% mom_post
    inf_or_pre <- (asy_lin_rep_ols_pre_treat - asy_lin_rep_ols_pre) %*% mom_pre

    # Put all pieces together
    inf_treat_or <- inf_treat_or_post + inf_treat_or_pre
    inf_control_or <- inf_control_or_post + inf_control_or_pre
    inf_or <- inf_or_post - inf_or_pre

    inf_treat <- inf_treat_post - inf_treat_pre + inf_treat_or
    inf_control <- inf_control_post - inf_control_pre + inf_control_ps + inf_control_or

    # Final influence function
    att_inf_func <- inf_treat - inf_control + inf_eff + inf_or

  } else {
    stop("Invalid est_method. Must be 'ipw', 'reg', or 'dr'")
  }

  # Fill zeros in influence function for observations outside the subgroup analyzed
  data[, "inf_func_result" := numeric(.N)]
  data[data$subgroup %in% c(condition_subgroup, 4), "inf_func_result" := att_inf_func]

  return(list(dr_att = att, inf_func = data[["inf_func_result"]]))
}

# Testing error handling in run_preprocess_2periods() function
library(data.table)

test_that("Testing error handling in run_preprocess_2periods() function", {

  # generating dataset without errors
  two_periods_no_errors_df = generate_test_panel()

  # ------------------------------
  # Performing tests
  # ------------------------------

  # Introducing discrepancy in the dataset
  y_not_numeric = copy(two_periods_no_errors_df)
  y_not_numeric$outcome = as.character(y_not_numeric$outcome) # Introducing error: converting numeric to character

  # Introducing missing values in the outcome
  missing_values_outcome_df = copy(two_periods_no_errors_df)
  missing_values_outcome_df[1:2, "outcome"] = NA # Introducing missing values in the outcome variable

  # Introducing missing values in the treatment
  missing_values_treat_df = copy(two_periods_no_errors_df)
  missing_values_treat_df[1:2, "treat"] = NA # Introducing missing values in the treatment variable

  # Dataset only with 1 treated unit (inference is no feasible)
  one_treated_unit_df = copy(two_periods_no_errors_df)
  one_treated_unit_df$treat = ifelse(one_treated_unit_df$treat == 1, 0, 0)
  one_treated_unit_df[1:2, "treat"] = 1

  # partition is no unique by id
  partition_not_unique_df = copy(two_periods_no_errors_df)
  partition_not_unique_df$partition[1] <- ifelse(partition_not_unique_df$partition[1] == 1, 0, 1)

  # treatment variables "dname" is not unique by id
  treat_not_unique_df = copy(two_periods_no_errors_df)
  treat_not_unique_df$treat[1] <- ifelse(treat_not_unique_df$treat[1] == 1, 0, 1)

  # covariates are varying over time
  covariates_invariant_df = copy(two_periods_no_errors_df)
  covariates_invariant_df$x1 = rnorm(nrow(covariates_invariant_df))

  # more than 2 time periods
  more_than_two_periods_df = generate_test_panel(time = 3)

  # more than 2 groups
  more_than_two_groups_df = copy(two_periods_no_errors_df)
  more_than_two_groups_df$treat[1] = 2

  # partition is not numeric
  partition_not_numeric_df = copy(two_periods_no_errors_df)
  partition_not_numeric_df$partition = as.character(partition_not_numeric_df$partition)

  # partition is not binary
  partition_not_binary_df = copy(two_periods_no_errors_df)
  partition_not_binary_df$partition = rnorm(nrow(partition_not_binary_df))

  # Providing weights columns and one of them is null
  weights_null_df = copy(two_periods_no_errors_df)
  weights_null_df$weights = rep(1, nrow(weights_null_df))
  weights_null_df[weights_null_df$id == 1]$weights = NA

  # ------------------------------
  # Warnings
  # ------------------------------

  # Test 2: Warning for est_method
  expect_warning(ddd(yname = "outcome", tname = "year", idname = "id", gname = "treat",
                     pname = "partition", xformla = ~x1 + x2,
                     data = two_periods_no_errors_df, control_group = NULL, base_period = NULL, est_method = "whatever",
                     weightsname = NULL, boot = FALSE, nboot = NULL,
                     inffunc = FALSE, skip_data_checks = FALSE))

  # Test 3: Warning for missing values in outcome variable "yname"
  expect_warning(ddd(yname = "outcome", tname = "year", idname = "id", gname = "treat",
                     pname = "partition", xformla = ~x1 + x2,
                     data = missing_values_outcome_df, control_group = NULL, base_period = NULL, est_method = "dr",
                     weightsname = NULL, boot = FALSE, nboot = NULL,
                     inffunc = FALSE, skip_data_checks = FALSE))


  # ------------------------------
  # Errors
  # ------------------------------
  # Test 4: handling of non-numeric "yname" column
  expect_error(ddd(yname = "outcome", tname = "year", idname = "id", gname = "treat",
                   pname = "partition", xformla = ~x1 + x2,
                   data = y_not_numeric, control_group = NULL, base_period = NULL, est_method = "dr",
                   weightsname = NULL, boot = FALSE, nboot = NULL,
                   inffunc = FALSE, skip_data_checks = FALSE))

  # Test 5: error for missing values in treatment variable "gname"
  expect_error(suppressWarnings(ddd(yname = "outcome", tname = "year", idname = "id", gname = "treat",
                   pname = "partition", xformla = ~x1 + x2,
                   data = missing_values_treat_df, control_group = "nevertreated", base_period = NULL, est_method = "dr",
                   weightsname = NULL, boot = FALSE, nboot = NULL,
                   inffunc = FALSE, skip_data_checks = FALSE)))

  # Test 6: error for small groups for inference (e.g. only one treated unit)
  expect_error(ddd(yname = "outcome", tname = "year", idname = "id", gname = "treat",
                   pname = "partition", xformla = ~x1 + x2,
                   data = one_treated_unit_df, control_group = NULL, base_period = NULL, est_method = "dr",
                   weightsname = NULL, boot = FALSE,  nboot = NULL,
                   inffunc = TRUE, skip_data_checks = FALSE))

  # Test 7: error when "yname" is not in data
  expect_error(ddd(yname = "whatever", tname = "year", idname = "id", gname = "treat",
                   pname = "partition", xformla = ~x1 + x2,
                   data = two_periods_no_errors_df, control_group = NULL, base_period = NULL, est_method = "dr",
                   weightsname = NULL, boot = FALSE, nboot = NULL,
                   inffunc = FALSE, skip_data_checks = FALSE))

  # Test 8: error when "tname" is not in data
  expect_error(ddd(yname = "outcome", tname = "whatever", idname = "id", gname = "treat",
                   pname = "partition", xformla = ~x1 + x2,
                   data = two_periods_no_errors_df, control_group = NULL, base_period = NULL, est_method = "dr",
                   weightsname = NULL, boot = FALSE, nboot = NULL,
                   inffunc = FALSE, skip_data_checks = FALSE))

  # Test 9: error when "dname" is not in data
  expect_error(ddd(yname = "outcome", tname = "year", idname = "id", gname = "whatever",
                   pname = "partition", xformla = ~x1 + x2,
                   data = two_periods_no_errors_df, control_group = NULL, base_period = NULL, est_method = "dr",
                   weightsname = NULL, boot = FALSE, nboot = NULL,
                   inffunc = FALSE, skip_data_checks = FALSE))

  # Test 10: error when "partition" is not in data
  expect_error(ddd(yname = "outcome", tname = "year", idname = "id", gname = "treat",
                   pname = "whatever", xformla = ~x1 + x2,
                   data = two_periods_no_errors_df, control_group = NULL, base_period = NULL, est_method = "dr",
                   weightsname = NULL, boot = FALSE, nboot = NULL,
                   inffunc = FALSE, skip_data_checks = FALSE))

  # Test 11: error when "idname" is not in data
  expect_error(ddd(yname = "outcome", tname = "year", idname = "whatever", gname = "treat",
                   pname = "whatever", xformla = ~x1 + x2,
                   data = two_periods_no_errors_df, control_group = NULL, base_period = NULL, est_method = "dr",
                   weightsname = NULL, boot = FALSE, nboot = NULL,
                   inffunc = FALSE, skip_data_checks = FALSE))

  # Test 12: error when "xformla" is not a formula
  expect_error(ddd(yname = "outcome", tname = "year", idname = "id", gname = "treat",
                   pname = "partition", xformla = "x1 - x2",
                   data = two_periods_no_errors_df, control_group = NULL, base_period = NULL, est_method = "dr",
                   weightsname = NULL, boot = FALSE, nboot = NULL,
                   inffunc = FALSE, skip_data_checks = FALSE))

  # Test 13: partition variable "partition" is not unique by id
  expect_error(ddd(yname = "outcome", tname = "year", idname = "id", gname = "treat",
                   pname = "partition", xformla = ~x1 + x2,
                   data = partition_not_unique_df, control_group = NULL, base_period = NULL, est_method = "dr",
                   weightsname = NULL, boot = FALSE, nboot = NULL,
                   inffunc = FALSE, skip_data_checks = FALSE))

  # Test 14: treatment variable "dname" is not unique by id
  expect_error(ddd(yname = "outcome", tname = "year", idname = "id", gname = "treat",
                   pname = "partition", xformla = ~x1 + x2,
                   data = treat_not_unique_df, control_group = NULL, base_period = NULL, est_method = "dr",
                   weightsname = NULL, boot = FALSE, nboot = NULL,
                   inffunc = FALSE, skip_data_checks = FALSE))

  # Test 15: error when covariates are not invariant
  expect_error(ddd(yname = "outcome", tname = "year", idname = "id", gname = "treat",
                   pname = "partition", xformla = ~x1 + x2,
                   data = covariates.not.invariant.df, control_group = NULL, base_period = NULL, est_method = "dr",
                   weightsname = NULL, boot = FALSE, nboot = NULL,
                   inffunc = FALSE, skip_data_checks = FALSE))

  # Test 16: More that 2 time periods
  expect_error(suppressWarnings(ddd(yname = "outcome", tname = "year", idname = "id", gname = "treat",
                   pname = "partition", xformla = ~x1 + x2,
                   data = more_than_two_periods_df, control_group = "nevertreated", base_period = NULL, est_method = "dr",
                   weightsname = NULL, boot = FALSE, nboot = NULL,
                   inffunc = FALSE, skip_data_checks = FALSE)))

  # Test 17: Partition is not numeric
  expect_error(ddd(yname = "outcome", tname = "year", idname = "id", gname = "treat",
                   pname = "partition", xformla = ~x1 + x2,
                   data = partition_not_numeric_df, control_group = NULL, base_period = NULL, est_method = "dr",
                   weightsname = NULL, boot = FALSE, nboot = NULL,
                   inffunc = FALSE, skip_data_checks = FALSE))

  # Test 18: Partition is not numeric
  expect_error(ddd(yname = "outcome", tname = "year", idname = "id", gname = "treat",
                   pname = "partition", xformla = ~x1 + x2,
                   data = partition_not_numeric_df, control_group = NULL, base_period = NULL, est_method = "dr",
                   weightsname = NULL, boot = FALSE, nboot = NULL,
                   inffunc = FALSE, skip_data_checks = FALSE))

  # Test 19: Weights columns contains Null values
  expect_error(ddd(yname = "outcome", tname = "year", idname = "id", gname = "treat",
                   pname = "partition", xformla = ~x1 + x2,
                   data = weights_null_df, control_group = NULL, base_period = NULL, est_method = "dr",
                   weightsname = "weights", boot = FALSE, nboot = NULL,
                   inffunc = FALSE, skip_data_checks = FALSE))

})

# Tests for collinearity handling

test_that("check_partition_collinearity helper function detects global collinearity", {
  # Test the helper function directly with globally collinear variables
  library(data.table)

  # Create test data with known collinearity (x3 = x1 exactly)
  set.seed(789)
  n <- 100
  test_data <- data.table(
    subgroup = c(rep(4, 25), rep(3, 25), rep(2, 25), rep(1, 25)),
    x1 = rnorm(n),
    x2 = rnorm(n)
  )

  # Add a variable that's collinear with x1 (globally collinear)
  test_data$x3 <- test_data$x1

  # Test the function
  result <- triplediff:::check_partition_collinearity(
    test_data,
    "subgroup",
    c("x1", "x2", "x3")
  )

  # Should detect collinearity in all partitions (since x3 = x1 everywhere)
  expect_true(length(result$all_collinear) > 0)
  expect_true("x3" %in% result$all_collinear)
})

test_that("check_partition_collinearity detects subset-specific collinearity", {
  # Test the helper function with subset-specific collinearity
  # x3 is a linear combination of x1 only within subgroups 4 and 3
  library(data.table)

  set.seed(999)
  n <- 200

  # Create test data
  test_data <- data.table(
    subgroup = c(rep(4, 50), rep(3, 50), rep(2, 50), rep(1, 50)),
    x1 = rnorm(n),
    x2 = rnorm(n)
  )

  # Add x3 that's collinear with x1 only in subgroups 4 and 3
  test_data$x3 <- rnorm(n)
  # Make x3 = 2*x1 only in subgroups 4 and 3 (perfect collinearity)
  test_data[subgroup %in% c(4, 3), x3 := 2 * x1]

  # Test the function
  result <- triplediff:::check_partition_collinearity(
    test_data,
    "subgroup",
    c("x1", "x2", "x3")
  )

  # Should detect collinearity in "subgroup 4 vs 3"
  expect_true("x3" %in% result$all_collinear)
  expect_true(any(grepl("subgroup 4 vs 3", result$collinear_vars$x3)))
})

test_that("check_partition_collinearity returns empty when no collinearity", {
  # Test the helper function with independent variables
  library(data.table)

  set.seed(111)
  n <- 100
  test_data <- data.table(
    subgroup = c(rep(4, 25), rep(3, 25), rep(2, 25), rep(1, 25)),
    x1 = rnorm(n),
    x2 = rnorm(n),
    x3 = rnorm(n)
  )

  # Test the function - should find no collinearity
  result <- triplediff:::check_partition_collinearity(
    test_data,
    "subgroup",
    c("x1", "x2", "x3")
  )

  # Should NOT detect any collinearity
  expect_equal(length(result$all_collinear), 0)
})

test_that("check_partition_collinearity identifies correct partition", {
  # Test that the function correctly identifies which partition has collinearity
  library(data.table)

  set.seed(222)
  n <- 200

  # Create test data
  test_data <- data.table(
    subgroup = c(rep(4, 50), rep(3, 50), rep(2, 50), rep(1, 50)),
    x1 = rnorm(n),
    x2 = rnorm(n)
  )

  # Add x3 that's collinear with x1 ONLY in subgroups 4 and 2 (not 3 or 1)
  test_data$x3 <- rnorm(n)
  test_data[subgroup %in% c(4, 2), x3 := 3 * x1]

  # Test the function
  result <- triplediff:::check_partition_collinearity(
    test_data,
    "subgroup",
    c("x1", "x2", "x3")
  )

  # Should detect collinearity in "subgroup 4 vs 2"
  expect_true("x3" %in% result$all_collinear)
  expect_true(any(grepl("subgroup 4 vs 2", result$collinear_vars$x3)))
})

test_that("check_partition_collinearity detects constant variables (collinear with intercept)", {
  # IMPORTANT: A constant variable is collinear with the intercept
  # This is a key case that must be detected
  library(data.table)

  set.seed(333)
  n <- 200

  # Create test data with an intercept column
  test_data <- data.table(
    subgroup = c(rep(4, 50), rep(3, 50), rep(2, 50), rep(1, 50)),
    intercept = rep(1, n),  # Intercept column (all 1s)
    x1 = rnorm(n),
    x2 = rnorm(n)
  )

  # Add x3 that's constant only in subgroups 4 and 3 (collinear with intercept in that partition)
  test_data$x3 <- rnorm(n)
  test_data[subgroup %in% c(4, 3), x3 := 5]  # Constant value in 4 vs 3 comparison

  # Test the function with intercept included
  result <- triplediff:::check_partition_collinearity(
    test_data,
    "subgroup",
    c("intercept", "x1", "x2", "x3")
  )

  # Should detect that x3 is collinear with the intercept in "subgroup 4 vs 3"
  expect_true("x3" %in% result$all_collinear)
  expect_true(any(grepl("subgroup 4 vs 3", result$collinear_vars$x3)))
})

test_that("check_partition_collinearity detects globally constant variables", {
  # A globally constant variable should be detected in ALL partitions
  library(data.table)

  set.seed(444)
  n <- 200

  # Create test data with an intercept column
  test_data <- data.table(
    subgroup = c(rep(4, 50), rep(3, 50), rep(2, 50), rep(1, 50)),
    intercept = rep(1, n),  # Intercept column (all 1s)
    x1 = rnorm(n),
    x2 = rnorm(n),
    x3 = rep(5, n)  # Globally constant - should be collinear with intercept everywhere
  )

  # Test the function
  result <- triplediff:::check_partition_collinearity(
    test_data,
    "subgroup",
    c("intercept", "x1", "x2", "x3")
  )

  # Should detect that x3 is collinear with the intercept in ALL partitions
  expect_true("x3" %in% result$all_collinear)
  # Should be detected in all three comparisons (4 vs 3, 4 vs 2, 4 vs 1)
  expect_true(length(result$collinear_vars$x3) == 3)
})

# =============================================================================
# Propensity Score Trimming Tests
# =============================================================================

test_that("compute_pscore returns keep_ps indicator for trimming", {
  # Test that compute_pscore returns keep_ps and that it follows DRDID approach
  library(data.table)

  # Generate test data with panel structure
  set.seed(555)
  n <- 200

  # Create panel data with 2 time periods
  test_data <- data.table(
    id = rep(1:n, each = 2),
    post = rep(c(0, 1), n),
    subgroup = rep(c(rep(4, n/2), rep(3, n/2)), each = 2),
    y = rnorm(n * 2),
    weights = rep(1, n * 2),
    x1 = rep(rnorm(n), each = 2),
    x2 = rep(rnorm(n), each = 2)
  )

  # Call compute_pscore
  result <- triplediff:::compute_pscore(
    data = test_data,
    condition_subgroup = 3,
    xformula = ~ x1 + x2,
    trim_level = 0.995
  )

  # Should return propensity_scores, hessian_matrix, and keep_ps

  expect_true("propensity_scores" %in% names(result))
  expect_true("hessian_matrix" %in% names(result))
  expect_true("keep_ps" %in% names(result))

  # keep_ps should be a logical vector of same length as propensity_scores
  expect_equal(length(result$keep_ps), length(result$propensity_scores))
  expect_true(is.logical(result$keep_ps))
})

test_that("compute_pscore_null returns keep_ps = TRUE for all units (REG method)", {
  # REG method doesn't use IPW, so no trimming is needed
  library(data.table)

  set.seed(666)
  n <- 100

  test_data <- data.table(
    id = rep(1:n, each = 2),
    post = rep(c(0, 1), n),
    subgroup = rep(c(rep(4, n/2), rep(3, n/2)), each = 2),
    y = rnorm(n * 2),
    weights = rep(1, n * 2)
  )

  result <- triplediff:::compute_pscore_null(
    data = test_data,
    condition_subgroup = 3
  )

  # Should return keep_ps = TRUE for all units
  expect_true("keep_ps" %in% names(result))
  expect_true(all(result$keep_ps == TRUE))
})

test_that("propensity score trimming excludes controls with ps >= 0.995", {
  # Test that control units with high propensity scores are trimmed
  library(data.table)

  # Create data where some controls will have very high propensity scores
  # by making their covariates similar to treated units
  set.seed(777)
  n_treated <- 100
  n_control <- 100

  # Treated units (subgroup 4) have x1 ~ N(3, 0.5) - strongly positive
  treated_x1 <- rnorm(n_treated, mean = 3, sd = 0.5)

  # Controls: most have x1 ~ N(-1, 1), but some have x1 ~ N(3, 0.1) (almost like treated)
  n_high_ps_controls <- 10
  control_x1 <- c(
    rnorm(n_control - n_high_ps_controls, mean = -1, sd = 1),  # Normal controls
    rnorm(n_high_ps_controls, mean = 3, sd = 0.1)  # "Almost treated" controls
  )

  test_data <- data.table(
    id = rep(1:(n_treated + n_control), each = 2),
    post = rep(c(0, 1), n_treated + n_control),
    subgroup = rep(c(rep(4, n_treated), rep(3, n_control)), each = 2),
    y = rnorm((n_treated + n_control) * 2),
    weights = rep(1, (n_treated + n_control) * 2),
    x1 = rep(c(treated_x1, control_x1), each = 2)
  )

  result <- suppressWarnings(triplediff:::compute_pscore(
    data = test_data,
    condition_subgroup = 3,
    xformula = ~ x1,
    trim_level = 0.995
  ))

  # Get unique data to match propensity scores
  uid_data <- unique(test_data, by = "id")
  PA4 <- ifelse(uid_data$subgroup == 4, 1, 0)

  # Treated units (PA4 = 1) should all have keep_ps = TRUE
  expect_true(all(result$keep_ps[PA4 == 1] == TRUE))

  # Some control units should be trimmed (have keep_ps = FALSE)
  # due to their high propensity scores
  control_keep_ps <- result$keep_ps[PA4 == 0]
  control_ps <- result$propensity_scores[PA4 == 0]

  # Verify trimming logic: keep_ps = FALSE when ps >= 0.995
  expect_true(all(control_keep_ps[control_ps >= 0.995] == FALSE))
  expect_true(all(control_keep_ps[control_ps < 0.995] == TRUE))
})

test_that("propensity scores are capped at 1 - 1e-6 (following DRDID)", {
  # Test that propensity scores are bounded away from 1
  library(data.table)

  # Create extreme separation case
  set.seed(888)
  n_treated <- 50
  n_control <- 50

  # Perfect separation: treated have x1 > 0, controls have x1 < 0
  treated_x1 <- runif(n_treated, min = 5, max = 10)
  control_x1 <- runif(n_control, min = -10, max = -5)

  test_data <- data.table(
    id = rep(1:(n_treated + n_control), each = 2),
    post = rep(c(0, 1), n_treated + n_control),
    subgroup = rep(c(rep(4, n_treated), rep(3, n_control)), each = 2),
    y = rnorm((n_treated + n_control) * 2),
    weights = rep(1, (n_treated + n_control) * 2),
    x1 = rep(c(treated_x1, control_x1), each = 2)
  )

  result <- suppressWarnings(
    triplediff:::compute_pscore(
      data = test_data,
      condition_subgroup = 3,
      xformula = ~ x1,
      trim_level = 0.995
    )
  )

  # All propensity scores should be <= 1 - 1e-6
  expect_true(all(result$propensity_scores <= 1 - 1e-6))

  # And > 0 (lower bound)
  expect_true(all(result$propensity_scores > 0))
})

test_that("trimming is applied to ATT weights in compute_did", {
  # Integration test: verify that trimmed units get zero weight
  library(data.table)

  # Generate data where we know some controls will be trimmed
  set.seed(999)
  n_treated <- 80
  n_control <- 80

  # Create strong separation for some controls
  treated_x1 <- rnorm(n_treated, mean = 2, sd = 0.5)
  n_high_ps <- 5
  control_x1 <- c(
    rnorm(n_control - n_high_ps, mean = -2, sd = 1),
    rnorm(n_high_ps, mean = 2, sd = 0.1)  # These should have high ps
  )

  test_data <- data.table(
    id = rep(1:(n_treated + n_control), each = 2),
    post = rep(c(0, 1), n_treated + n_control),
    subgroup = rep(c(rep(4, n_treated), rep(3, n_control)), each = 2),
    y = rnorm((n_treated + n_control) * 2),
    weights = rep(1, (n_treated + n_control) * 2),
    x1 = rep(c(treated_x1, control_x1), each = 2)
  )

  # Compute propensity scores
  pscores_result <- suppressWarnings(triplediff:::compute_pscore(
    data = test_data,
    condition_subgroup = 3,
    xformula = ~ x1,
    trim_level = 0.995
  ))

  # Create pscores list structure as expected by compute_did
  pscores <- list(pscores_result)  # Index 1 for subgroup 3

  # Compute outcome regression (simple version)
  reg_result <- triplediff:::compute_outcome_regression(
    data = test_data,
    condition_subgroup = 3,
    xformula = ~ x1
  )
  reg_adjustment <- list(reg_result)

  # Add deltaY to unique data
  uid_data <- unique(test_data, by = "id")
  uid_data[, subgroup := subgroup]

  # The compute_did function should use keep_ps to zero out trimmed weights
  # This is an indirect test - we verify the propensity score list has trimming info
  expect_true("keep_ps" %in% names(pscores[[1]]))

  # Verify some controls are trimmed
  uid_controls <- uid_data[subgroup == 3]
  control_ps <- pscores_result$propensity_scores[uid_data$subgroup == 3]
  control_keep <- pscores_result$keep_ps[uid_data$subgroup == 3]

  # If any controls have high ps, they should be trimmed
  high_ps_controls <- control_ps >= 0.995
  if (any(high_ps_controls)) {
    expect_true(all(control_keep[high_ps_controls] == FALSE))
  }
})

test_that("confidence interval coverage is valid after trimming", {
  # Monte Carlo test: verify that CI coverage is within 2% of nominal 95% level
  # This tests that the trimming doesn't break the inference
  # DGP with moderate covariate imbalance that may trigger some trimming
  skip_on_cran()  # Skip on CRAN due to computation time

  library(data.table)

  set.seed(1234)
  n_sims <- 1000  # Number of Monte Carlo replications
  true_att <- 2   # True treatment effect
  alpha <- 0.05
  nominal_coverage <- 1 - alpha  # 0.95

  # Track coverage
  covers <- logical(n_sims)

  for (sim in 1:n_sims) {
    # Generate DDD panel data with known treatment effect
    # Need units in all 4 cells: treated/control x partition=0/1
    n_per_cell <- 500  # 500 units per cell = 2000 total (larger sample for better coverage)
    n_units <- 4 * n_per_cell

    # Create balanced 2x2 design
    treat_group <- c(rep(1, 2*n_per_cell), rep(0, 2*n_per_cell))
    partition <- c(rep(1, n_per_cell), rep(0, n_per_cell),
                   rep(1, n_per_cell), rep(0, n_per_cell))

    # Covariates with moderate difference between treated and control
    # This creates propensity score variation but with good overlap
    # Treated units: x1 ~ N(0.5, 1)
    # Control units: most ~ N(-0.5, 1), but a few "almost treated" with extreme x1
    x1_treated <- rnorm(2 * n_per_cell, mean = 0.5, sd = 1)

    # Add ~5 "almost treated" controls per partition (10 total) with x1 ~ N(10, 0.1)
    # These will have propensity scores >= 0.995 and trigger trimming
    n_high_ps <- 5
    n_normal_control <- n_per_cell - n_high_ps
    x1_control <- c(
      rnorm(n_normal_control, mean = -0.5, sd = 1),  # Normal controls, partition=1
      rnorm(n_high_ps, mean = 10, sd = 0.1),         # "Almost treated" controls, partition=1
      rnorm(n_normal_control, mean = -0.5, sd = 1),  # Normal controls, partition=0
      rnorm(n_high_ps, mean = 10, sd = 0.1)          # "Almost treated" controls, partition=0
    )
    x1 <- c(x1_treated, x1_control)
    x2 <- rnorm(n_units)

    # Create panel data (2 time periods)
    panel_data <- data.table(
      id = rep(1:n_units, each = 2),
      year = rep(c(0, 1), n_units),
      partition = rep(partition, each = 2),
      treat = rep(treat_group, each = 2),
      x1 = rep(x1, each = 2),
      x2 = rep(x2, each = 2)
    )

    # Generate outcome with DDD structure:
    # Y = base + partition_effect + time_effect + treat_effect + x_effects + DDD_effect + noise
    panel_data[, outcome := (
      1 +                                      # base
        0.5 * partition +                        # partition effect
        0.3 * year +                             # time effect
        0.4 * treat +                            # treatment group effect
        0.5 * x1 + 0.3 * x2 +                    # covariate effects
        0.2 * partition * year +                 # parallel trends for partition
        0.2 * treat * year +                     # parallel trends for treatment
        0.1 * partition * treat +                # partition-treatment interaction
        # DDD effect: only for treated, post-period, partition=1
        true_att * treat * year * partition +
        rnorm(.N, sd = 1)                        # noise
    )]

    # Run the DDD estimator
    result <- tryCatch({
      suppressWarnings(
        ddd(
          yname = "outcome",
          tname = "year",
          idname = "id",
          gname = "treat",
          pname = "partition",
          xformla = ~ x1 + x2,
          data = panel_data,
          est_method = "dr",
          boot = FALSE,
          inffunc = TRUE
        )
      )
    }, error = function(e) NULL)

    if (!is.null(result) && !is.null(result$ATT) && !is.null(result$se) &&
        !is.na(result$ATT) && !is.na(result$se)) {
      # Check if true ATT is within 95% CI
      # ddd returns ATT, se, lci, uci directly (not nested)
      est <- result$ATT
      se <- result$se
      ci_lower <- est - qnorm(1 - alpha/2) * se
      ci_upper <- est + qnorm(1 - alpha/2) * se
      covers[sim] <- (true_att >= ci_lower) & (true_att <= ci_upper)
    } else {
      covers[sim] <- NA
    }
  }

  # Calculate coverage rate (excluding failed simulations)
  valid_covers <- covers[!is.na(covers)]
  coverage_rate <- mean(valid_covers)

  # Coverage should be within 2% of nominal 95% level (i.e., between 93% and 97%)
  expect_true(coverage_rate >= nominal_coverage - 0.02,
              info = paste("Coverage rate:", round(coverage_rate, 4),
                           "- expected at least", nominal_coverage - 0.02))
  expect_true(coverage_rate <= nominal_coverage + 0.02,
              info = paste("Coverage rate:", round(coverage_rate, 4),
                           "- expected at most", nominal_coverage + 0.02))
})

# =============================================================================
# Global Collinearity Check Tests for Multiple Periods
# =============================================================================

test_that("run_preprocess_multPeriods detects and drops globally collinear covariates", {
  # Test that globally collinear variables are detected and dropped in multi-period setting
  library(data.table)

  set.seed(12345)
  n_units <- 200
  n_periods <- 4

  # Create staggered adoption panel data
  # Groups: 0 (never treated), 2 (treated in period 2), 3 (treated in period 3)
  group_assignment <- sample(c(0, 2, 3), n_units, replace = TRUE, prob = c(0.4, 0.3, 0.3))

  # Create panel structure
  panel_data <- data.table(
    id = rep(1:n_units, each = n_periods),
    year = rep(1:n_periods, n_units),
    first_treat = rep(group_assignment, each = n_periods),
    partition = rep(sample(0:1, n_units, replace = TRUE), each = n_periods)
  )

  # Add covariates
  panel_data[, x1 := rnorm(.N)]
  panel_data[, x2 := rnorm(.N)]
  # x3 is perfectly collinear with x1 (x3 = 2*x1)
  panel_data[, x3 := 2 * x1]

  # Generate outcome
  panel_data[, outcome := 1 + 0.5 * x1 + 0.3 * x2 + rnorm(.N)]

  # Run ddd with collinear covariates - should warn about dropping x3

  expect_warning(
    result <- ddd(
      yname = "outcome",
      tname = "year",
      idname = "id",
      gname = "first_treat",
      pname = "partition",
      xformla = ~ x1 + x2 + x3,
      data = panel_data,
      control_group = "nevertreated",
      base_period = "varying",
      est_method = "dr",
      boot = FALSE
    ),
    regexp = "collinearity"
  )

  # Result should still be valid (estimation should complete)
  expect_true(!is.null(result))
  expect_true(inherits(result, "ddd"))
})

test_that("run_preprocess_multPeriods handles no collinearity case correctly",
{
  # Test that when there's no collinearity, all covariates are kept
  library(data.table)

  set.seed(54321)
  n_units <- 200
  n_periods <- 4

  # Create staggered adoption panel data
  group_assignment <- sample(c(0, 2, 3), n_units, replace = TRUE, prob = c(0.4, 0.3, 0.3))

  panel_data <- data.table(
    id = rep(1:n_units, each = n_periods),
    year = rep(1:n_periods, n_units),
    first_treat = rep(group_assignment, each = n_periods),
    partition = rep(sample(0:1, n_units, replace = TRUE), each = n_periods)
  )

  # Add independent covariates (no collinearity)
  panel_data[, x1 := rnorm(.N)]
  panel_data[, x2 := rnorm(.N)]
  panel_data[, x3 := rnorm(.N)]

  # Generate outcome
  panel_data[, outcome := 1 + 0.5 * x1 + 0.3 * x2 + 0.2 * x3 + rnorm(.N)]

  # Run ddd - should NOT warn about collinearity
  expect_no_warning(
    result <- suppressWarnings(ddd(
      yname = "outcome",
      tname = "year",
      idname = "id",
      gname = "first_treat",
      pname = "partition",
      xformla = ~ x1 + x2 + x3,
      data = panel_data,
      control_group = "nevertreated",
      base_period = "varying",
      est_method = "dr",
      boot = FALSE
    )),
    message = "collinearity"
  )

  # Result should be valid
  expect_true(!is.null(result))
  expect_true(inherits(result, "ddd"))
})

test_that("run_preprocess_multPeriods detects constant covariate (collinear with intercept)", {

  # Test that a constant variable (collinear with intercept) is detected
  library(data.table)

  set.seed(99999)
  n_units <- 200
  n_periods <- 4

  group_assignment <- sample(c(0, 2, 3), n_units, replace = TRUE, prob = c(0.4, 0.3, 0.3))

  panel_data <- data.table(
    id = rep(1:n_units, each = n_periods),
    year = rep(1:n_periods, n_units),
    first_treat = rep(group_assignment, each = n_periods),
    partition = rep(sample(0:1, n_units, replace = TRUE), each = n_periods)
  )

  # Add covariates - x3 is constant (collinear with intercept)
  panel_data[, x1 := rnorm(.N)]
  panel_data[, x2 := rnorm(.N)]
  panel_data[, x3 := 5]  # Constant value

  # Generate outcome
  panel_data[, outcome := 1 + 0.5 * x1 + 0.3 * x2 + rnorm(.N)]

  # Run ddd - should warn about dropping x3
  expect_warning(
    result <- ddd(
      yname = "outcome",
      tname = "year",
      idname = "id",
      gname = "first_treat",
      pname = "partition",
      xformla = ~ x1 + x2 + x3,
      data = panel_data,
      control_group = "nevertreated",
      base_period = "varying",
      est_method = "dr",
      boot = FALSE
    ),
    regexp = "collinearity"
  )

  expect_true(!is.null(result))
  expect_true(inherits(result, "ddd"))
})

# =============================================================================
# Repeated Cross-Section (RCS) Propensity Score Trimming Tests
# =============================================================================

test_that("compute_pscore_rc returns keep_ps indicator for trimming in RCS data", {

  # Test that compute_pscore_rc returns keep_ps for repeated cross-section data
  # Key difference from panel: each observation is independent (no repeated IDs)
  library(data.table)

  set.seed(1111)
  n_pre <- 100   # observations in pre-period
  n_post <- 100  # observations in post-period

  # Create RCS data structure: different units in pre and post periods
  # In RCS, each row is a unique observation with unique ID
  test_data <- data.table(
    id = 1:(n_pre + n_post),  # Each observation has unique ID
    post = c(rep(0, n_pre), rep(1, n_post)),
    subgroup = c(
      rep(4, n_pre/2), rep(3, n_pre/2),   # Pre-period: half treated, half control
      rep(4, n_post/2), rep(3, n_post/2)  # Post-period: half treated, half control
    ),
    y = rnorm(n_pre + n_post),
    weights = rep(1, n_pre + n_post),
    x1 = rnorm(n_pre + n_post),
    x2 = rnorm(n_pre + n_post)
  )

  # Call compute_pscore_rc
  result <- triplediff:::compute_pscore_rc(
    data = test_data,
    condition_subgroup = 3,
    xformula = ~ x1 + x2,
    trim_level = 0.995
  )

  # Should return propensity_scores, hessian_matrix, and keep_ps
  expect_true("propensity_scores" %in% names(result))
  expect_true("hessian_matrix" %in% names(result))
  expect_true("keep_ps" %in% names(result))

  # For RCS, length should match all observations in condition_data (not unique by id)
  condition_data <- test_data[test_data$subgroup %in% c(3, 4)]
  expect_equal(length(result$keep_ps), nrow(condition_data))
  expect_true(is.logical(result$keep_ps))
})

test_that("compute_pscore_null_rc returns keep_ps = TRUE for all units (REG method in RCS)", {
  # REG method doesn't use IPW, so no trimming is needed in RCS
  library(data.table)

  set.seed(2222)
  n <- 200

  # Create RCS data
  test_data <- data.table(
    id = 1:n,
    post = c(rep(0, n/2), rep(1, n/2)),
    subgroup = rep(c(4, 3), n/2),
    y = rnorm(n),
    weights = rep(1, n)
  )

  result <- triplediff:::compute_pscore_null_rc(
    data = test_data,
    condition_subgroup = 3
  )

  # Should return keep_ps = TRUE for all observations
  expect_true("keep_ps" %in% names(result))
  expect_true(all(result$keep_ps == TRUE))

  # Length should match all observations in subgroups 3 and 4
  condition_data <- test_data[test_data$subgroup %in% c(3, 4)]
  expect_equal(length(result$keep_ps), nrow(condition_data))
})

test_that("RCS propensity score trimming excludes controls with ps >= 0.995", {
  # Test that control units with high propensity scores are trimmed in RCS
  library(data.table)

  set.seed(3333)
  n_treated <- 100
  n_control <- 100

  # Treated units (subgroup 4) have x1 ~ N(3, 0.5) - strongly positive
  treated_x1 <- rnorm(n_treated, mean = 3, sd = 0.5)

  # Controls: most have x1 ~ N(-1, 1), but some have x1 ~ N(3, 0.1) (almost like treated)
  n_high_ps_controls <- 10
  control_x1 <- c(
    rnorm(n_control - n_high_ps_controls, mean = -1, sd = 1),
    rnorm(n_high_ps_controls, mean = 3, sd = 0.1)  # "Almost treated" controls
  )

  # Create RCS data: each observation is independent
  test_data <- data.table(
    id = 1:(n_treated + n_control),
    post = sample(0:1, n_treated + n_control, replace = TRUE),  # Random pre/post assignment
    subgroup = c(rep(4, n_treated), rep(3, n_control)),
    y = rnorm(n_treated + n_control),
    weights = rep(1, n_treated + n_control),
    x1 = c(treated_x1, control_x1)
  )

  result <- suppressWarnings(triplediff:::compute_pscore_rc(
    data = test_data,
    condition_subgroup = 3,
    xformula = ~ x1,
    trim_level = 0.995
  ))

  # Get PA4 indicator for the condition_data
  condition_data <- test_data[test_data$subgroup %in% c(3, 4)]
  PA4 <- ifelse(condition_data$subgroup == 4, 1, 0)

  # Treated units (PA4 = 1) should all have keep_ps = TRUE
  expect_true(all(result$keep_ps[PA4 == 1] == TRUE))

  # Verify trimming logic: keep_ps = FALSE when ps >= 0.995 for controls
  control_keep_ps <- result$keep_ps[PA4 == 0]
  control_ps <- result$propensity_scores[PA4 == 0]

  expect_true(all(control_keep_ps[control_ps >= 0.995] == FALSE))
  expect_true(all(control_keep_ps[control_ps < 0.995] == TRUE))
})

test_that("RCS propensity scores are capped at 1 - 1e-6 (following DRDID)", {
  # Test that propensity scores are bounded away from 1 in RCS
  library(data.table)

  set.seed(4444)
  n_treated <- 50
  n_control <- 50

  # Perfect separation: treated have x1 > 0, controls have x1 < 0
  treated_x1 <- runif(n_treated, min = 5, max = 10)
  control_x1 <- runif(n_control, min = -10, max = -5)

  # Create RCS data
  test_data <- data.table(
    id = 1:(n_treated + n_control),
    post = sample(0:1, n_treated + n_control, replace = TRUE),
    subgroup = c(rep(4, n_treated), rep(3, n_control)),
    y = rnorm(n_treated + n_control),
    weights = rep(1, n_treated + n_control),
    x1 = c(treated_x1, control_x1)
  )

  result <- suppressWarnings(
    triplediff:::compute_pscore_rc(
      data = test_data,
      condition_subgroup = 3,
      xformula = ~ x1,
      trim_level = 0.995
    )
  )

  # All propensity scores should be <= 1 - 1e-6
  expect_true(all(result$propensity_scores <= 1 - 1e-6))

  # And > 0 (lower bound)
  expect_true(all(result$propensity_scores > 0))
})

test_that("RCS trimming works correctly with compute_did_rc for IPW method", {
  # Integration test: verify trimming is applied in compute_did_rc
  library(data.table)

  set.seed(5555)
  n_treated <- 80
  n_control <- 80

  # Create separation for some controls
  treated_x1 <- rnorm(n_treated, mean = 2, sd = 0.5)
  n_high_ps <- 5
  control_x1 <- c(
    rnorm(n_control - n_high_ps, mean = -2, sd = 1),
    rnorm(n_high_ps, mean = 2, sd = 0.1)  # High propensity score controls
  )

  # Create RCS data with pre and post periods
  n_total <- n_treated + n_control
  test_data <- data.table(
    id = 1:n_total,
    post = c(rep(0, n_total/2), rep(1, n_total/2)),  # Half pre, half post
    subgroup = rep(c(rep(4, n_treated/2), rep(3, n_control/2)), 2),
    y = rnorm(n_total),
    weights = rep(1, n_total),
    x1 = rep(c(treated_x1[1:(n_treated/2)], control_x1[1:(n_control/2)]), 2)
  )

  # Compute propensity scores for RCS
  pscores_result <- suppressWarnings(triplediff:::compute_pscore_rc(
    data = test_data,
    condition_subgroup = 3,
    xformula = ~ x1,
    trim_level = 0.995
  ))

  # Verify the propensity score result has trimming info
  expect_true("keep_ps" %in% names(pscores_result))

  # Check that some controls are trimmed if they have high propensity scores
  condition_data <- test_data[test_data$subgroup %in% c(3, 4)]
  PA4 <- ifelse(condition_data$subgroup == 4, 1, 0)
  control_ps <- pscores_result$propensity_scores[PA4 == 0]
  control_keep <- pscores_result$keep_ps[PA4 == 0]

  # If any controls have high ps, they should be trimmed
  high_ps_controls <- control_ps >= 0.995
  if (any(high_ps_controls)) {
    expect_true(all(control_keep[high_ps_controls] == FALSE))
  }
})

test_that("End-to-end RCS DDD estimation with trimming works correctly", {
  # Full integration test: run ddd() with panel=FALSE (RCS) and verify trimming
  skip_on_cran()  # Skip on CRAN due to computation time
  library(data.table)

  set.seed(6666)
  n_obs <- 400  # Total observations

  # Create RCS data with known structure
  # In RCS, we need: treated/control x partition x pre/post
  # Each observation is independent (different units)

  # Create balanced design across cells
  n_per_cell <- n_obs / 8  # 8 cells: 2 treat x 2 partition x 2 time

  test_data <- data.table(
    id = 1:n_obs,  # Each row is unique observation
    year = rep(c(1, 2), each = n_obs/2),  # Pre and post periods
    treat = rep(rep(c(0, 1), each = n_per_cell * 2), 2),
    partition = rep(rep(c(0, 1), each = n_per_cell), 4)
  )

  # Add covariates with moderate separation
  # Treated units have higher x1 on average
  test_data[, x1 := ifelse(treat == 1, rnorm(.N, mean = 1, sd = 1), rnorm(.N, mean = -1, sd = 1))]
  test_data[, x2 := rnorm(.N)]

  # Add a few "almost treated" controls to trigger trimming
  n_high_ps <- 5
  high_ps_indices <- sample(which(test_data$treat == 0), n_high_ps)
  test_data[high_ps_indices, x1 := rnorm(n_high_ps, mean = 5, sd = 0.1)]

  # Generate outcome with DDD structure
  true_att <- 2
  test_data[, outcome := (
    1 +
    0.5 * partition +
    0.3 * year +
    0.4 * treat +
    0.3 * x1 + 0.2 * x2 +
    true_att * treat * year * partition +
    rnorm(.N, sd = 1)
  )]

  # Run DDD with panel = FALSE (RCS mode)
  result <- suppressWarnings(
    ddd(
      yname = "outcome",
      tname = "year",
      idname = "id",
      gname = "treat",
      pname = "partition",
      xformla = ~ x1 + x2,
      data = test_data,
      panel = FALSE,  # This triggers RCS mode
      est_method = "dr",
      boot = FALSE
    )
  )

  # Result should be valid
  expect_true(!is.null(result))
  expect_true(inherits(result, "ddd"))
  expect_true(!is.na(result$ATT))
  expect_true(!is.na(result$se))
})

test_that("RCS confidence interval coverage is valid after trimming", {
  # Monte Carlo test: verify that CI coverage is within 2% of nominal 95% level

  # This tests that the trimming doesn't break the inference for RCS data
  # DGP with moderate covariate imbalance that may trigger some trimming
  skip_on_cran()  # Skip on CRAN due to computation time

  library(data.table)

  set.seed(5678)
  n_sims <- 1000  # Number of Monte Carlo replications
  true_att <- 2   # True treatment effect
  alpha <- 0.05
  nominal_coverage <- 1 - alpha  # 0.95

  # Track coverage
  covers <- logical(n_sims)

  for (sim in 1:n_sims) {
    # Generate DDD RCS data with known treatment effect
    # In RCS, each observation is independent (different units in pre vs post)
    # Need observations in all 8 cells: treated/control x partition x pre/post
    n_per_cell <- 250  # 250 obs per cell = 2000 total
    n_obs <- 8 * n_per_cell

    # Create balanced 2x2x2 design for RCS
    # Each row is a unique observation (unique id)
    rcs_data <- data.table(
      id = 1:n_obs,
      year = rep(c(0, 1), each = n_obs/2),
      treat = rep(rep(c(0, 1), each = n_per_cell * 2), 2),
      partition = rep(rep(c(0, 1), each = n_per_cell), 4)
    )

    # Covariates with moderate difference between treated and control
    # Treated units: x1 ~ N(0.5, 1)
    # Control units: most ~ N(-0.5, 1), but a few "almost treated" with extreme x1
    rcs_data[, x1 := ifelse(treat == 1,
                            rnorm(.N, mean = 0.5, sd = 1),
                            rnorm(.N, mean = -0.5, sd = 1))]
    rcs_data[, x2 := rnorm(.N)]

    # Add ~10 "almost treated" controls with x1 ~ N(10, 0.1)
    # These will have propensity scores >= 0.995 and trigger trimming
    n_high_ps <- 10
    high_ps_indices <- sample(which(rcs_data$treat == 0), n_high_ps)
    rcs_data[high_ps_indices, x1 := rnorm(n_high_ps, mean = 10, sd = 0.1)]

    # Generate outcome with DDD structure:
    # Y = base + partition_effect + time_effect + treat_effect + x_effects + DDD_effect + noise
    rcs_data[, outcome := (
      1 +                                      # base
        0.5 * partition +                        # partition effect
        0.3 * year +                             # time effect
        0.4 * treat +                            # treatment group effect
        0.5 * x1 + 0.3 * x2 +                    # covariate effects
        0.2 * partition * year +                 # parallel trends for partition
        0.2 * treat * year +                     # parallel trends for treatment
        0.1 * partition * treat +                # partition-treatment interaction
        # DDD effect: only for treated, post-period, partition=1
        true_att * treat * year * partition +
        rnorm(.N, sd = 1)                        # noise
    )]

    # Run the DDD estimator with panel = FALSE (RCS mode)
    result <- tryCatch({
      suppressWarnings(
        ddd(
          yname = "outcome",
          tname = "year",
          idname = "id",
          gname = "treat",
          pname = "partition",
          xformla = ~ x1 + x2,
          data = rcs_data,
          panel = FALSE,  # RCS mode
          est_method = "dr",
          boot = FALSE,
          inffunc = TRUE
        )
      )
    }, error = function(e) NULL)

    if (!is.null(result) && !is.null(result$ATT) && !is.null(result$se) &&
        !is.na(result$ATT) && !is.na(result$se)) {
      # Check if true ATT is within 95% CI
      est <- result$ATT
      se <- result$se
      ci_lower <- est - qnorm(1 - alpha/2) * se
      ci_upper <- est + qnorm(1 - alpha/2) * se
      covers[sim] <- (true_att >= ci_lower) & (true_att <= ci_upper)
    } else {
      covers[sim] <- NA
    }
  }

  # Calculate coverage rate (excluding failed simulations)
  valid_covers <- covers[!is.na(covers)]
  coverage_rate <- mean(valid_covers)

  # Coverage should be within 2% of nominal 95% level (i.e., between 93% and 97%)
  expect_true(coverage_rate >= nominal_coverage - 0.02,
              info = paste("RCS Coverage rate:", round(coverage_rate, 4),
                           "- expected at least", nominal_coverage - 0.02))
  expect_true(coverage_rate <= nominal_coverage + 0.02,
              info = paste("RCS Coverage rate:", round(coverage_rate, 4),
                           "- expected at most", nominal_coverage + 0.02))
})

# =============================================================================
# End-to-End Panel DDD Estimation with Trimming Tests
# =============================================================================

test_that("End-to-end Panel DDD estimation with trimming works correctly", {
  # Full integration test: run ddd() with panel=TRUE (default) and verify trimming
  skip_on_cran()  # Skip on CRAN due to computation time
  library(data.table)

  set.seed(7777)
  n_units <- 200  # Number of unique units

  # Create panel data with known structure
  # In panel data, same units are observed in pre and post periods
  # Need: treated/control x partition balanced design

  n_per_cell <- n_units / 4  # 4 cells: 2 treat x 2 partition

  # Create unit-level attributes (constant across time)
  unit_data <- data.table(
    id = 1:n_units,
    treat = rep(c(0, 1), each = n_units/2),
    partition = rep(rep(c(0, 1), each = n_per_cell), 2)
  )

  # Add covariates with moderate separation (unit-level)
  # Treated units have higher x1 on average
  unit_data[, x1 := ifelse(treat == 1, rnorm(.N, mean = 1, sd = 1), rnorm(.N, mean = -1, sd = 1))]
  unit_data[, x2 := rnorm(.N)]

  # Add a few "almost treated" controls to trigger trimming
  n_high_ps <- 5
  high_ps_indices <- sample(which(unit_data$treat == 0), n_high_ps)
  unit_data[high_ps_indices, x1 := rnorm(n_high_ps, mean = 5, sd = 0.1)]

  # Expand to panel data (2 time periods per unit)
  test_data <- rbindlist(lapply(c(1, 2), function(t) {
    dt <- copy(unit_data)
    dt[, year := t]
    dt
  }))
  setorder(test_data, id, year)

  # Generate outcome with DDD structure
  true_att <- 2
  test_data[, outcome := (
    1 +
    0.5 * partition +
    0.3 * (year - 1) +  # year effect (year 1 = pre, year 2 = post)
    0.4 * treat +
    0.3 * x1 + 0.2 * x2 +
    true_att * treat * (year - 1) * partition +  # DDD effect only in post period
    rnorm(.N, sd = 1)
  )]

  # Run DDD with panel = TRUE (default, panel mode)
  result <- suppressWarnings(
    ddd(
      yname = "outcome",
      tname = "year",
      idname = "id",
      gname = "treat",
      pname = "partition",
      xformla = ~ x1 + x2,
      data = test_data,
      panel = TRUE,  # Explicitly set panel mode
      est_method = "dr",
      boot = FALSE
    )
  )

  # Result should be valid
  expect_true(!is.null(result))
  expect_true(inherits(result, "ddd"))
  expect_true(!is.na(result$ATT))
  expect_true(!is.na(result$se))

  # ATT should be reasonably close to true value (within 3 SE)
  # This is a sanity check, not a strict test
  expect_true(abs(result$ATT - true_att) < 3 * result$se,
              info = paste("ATT:", round(result$ATT, 3), "SE:", round(result$se, 3)))
})

test_that("End-to-end Panel DDD estimation with IPW method and trimming", {
  # Test IPW method specifically to ensure trimming is applied correctly
  skip_on_cran()
  library(data.table)

  set.seed(8888)
  n_units <- 200

  n_per_cell <- n_units / 4

  # Create unit-level data
  unit_data <- data.table(
    id = 1:n_units,
    treat = rep(c(0, 1), each = n_units/2),
    partition = rep(rep(c(0, 1), each = n_per_cell), 2)
  )

  # Add covariates - treated units have higher x1
  unit_data[, x1 := ifelse(treat == 1, rnorm(.N, mean = 1.5, sd = 1), rnorm(.N, mean = -1.5, sd = 1))]
  unit_data[, x2 := rnorm(.N)]

  # Add "almost treated" controls to trigger trimming
  n_high_ps <- 8
  high_ps_indices <- sample(which(unit_data$treat == 0), n_high_ps)
  unit_data[high_ps_indices, x1 := rnorm(n_high_ps, mean = 6, sd = 0.1)]

  # Expand to panel
  test_data <- rbindlist(lapply(c(1, 2), function(t) {
    dt <- copy(unit_data)
    dt[, year := t]
    dt
  }))
  setorder(test_data, id, year)

  # Generate outcome
  true_att <- 1.5
  test_data[, outcome := (
    1 +
    0.5 * partition +
    0.3 * (year - 1) +
    0.4 * treat +
    0.3 * x1 + 0.2 * x2 +
    true_att * treat * (year - 1) * partition +
    rnorm(.N, sd = 1)
  )]

  # Run DDD with IPW method
  result <- suppressWarnings(
    ddd(
      yname = "outcome",
      tname = "year",
      idname = "id",
      gname = "treat",
      pname = "partition",
      xformla = ~ x1 + x2,
      data = test_data,
      panel = TRUE,
      est_method = "ipw",
      boot = FALSE
    )
  )

  # Result should be valid
  expect_true(!is.null(result))
  expect_true(inherits(result, "ddd"))
  expect_true(!is.na(result$ATT))
  expect_true(!is.na(result$se))
})

test_that("End-to-end Panel DDD estimation with REG method (no trimming needed)", {
  # REG method doesn't use propensity scores, so trimming shouldn't affect it
  skip_on_cran()
  library(data.table)

  set.seed(9999)
  n_units <- 200

  n_per_cell <- n_units / 4

  # Create unit-level data
  unit_data <- data.table(
    id = 1:n_units,
    treat = rep(c(0, 1), each = n_units/2),
    partition = rep(rep(c(0, 1), each = n_per_cell), 2)
  )

  # Add covariates
  unit_data[, x1 := ifelse(treat == 1, rnorm(.N, mean = 1, sd = 1), rnorm(.N, mean = -1, sd = 1))]
  unit_data[, x2 := rnorm(.N)]

  # Add "almost treated" controls (shouldn't matter for REG)
  n_high_ps <- 5
  high_ps_indices <- sample(which(unit_data$treat == 0), n_high_ps)
  unit_data[high_ps_indices, x1 := rnorm(n_high_ps, mean = 5, sd = 0.1)]

  # Expand to panel
  test_data <- rbindlist(lapply(c(1, 2), function(t) {
    dt <- copy(unit_data)
    dt[, year := t]
    dt
  }))
  setorder(test_data, id, year)

  # Generate outcome
  true_att <- 2.5
  test_data[, outcome := (
    1 +
    0.5 * partition +
    0.3 * (year - 1) +
    0.4 * treat +
    0.3 * x1 + 0.2 * x2 +
    true_att * treat * (year - 1) * partition +
    rnorm(.N, sd = 1)
  )]

  # Run DDD with REG method
  result <- suppressWarnings(
    ddd(
      yname = "outcome",
      tname = "year",
      idname = "id",
      gname = "treat",
      pname = "partition",
      xformla = ~ x1 + x2,
      data = test_data,
      panel = TRUE,
      est_method = "reg",
      boot = FALSE
    )
  )

  # Result should be valid
  expect_true(!is.null(result))
  expect_true(inherits(result, "ddd"))
  expect_true(!is.na(result$ATT))
  expect_true(!is.na(result$se))
})

# Main function for aggregations steps in triplediff
NULL
#' Aggregate Group-Time Average Treatment Effects in Staggered Triple-Differences Designs.
#' @description
#' \code{agg_ddd} is a function that take group-time average treatment effects
#'  and aggregate them into a smaller number of summary parameters in staggered triple differences designs.
#'  There are several possible aggregations including \code{"simple"}, \code{"eventstudy"}, \code{"group"},
#'  and \code{"calendar"}.  Default is \code{"eventstudy"}.
#'
#' @param ddd_obj a \code{ddd} object (i.e., the results of the [ddd()] function)
#' @param type Which type of aggregated treatment effect parameter to compute.
#'   \code{"simple"} just computes a weighted average of all
#'   group-time average treatment effects with weights proportional to group
#'   size.
#'   \code{"eventstudy"} computes average effects across
#'   different lengths of exposure to the treatment (event times). Here the overall effect averages the effect of the
#'   treatment across the positive lengths of exposure. This is the default option;
#'   \code{"group"} computes average treatment effects across different groups/cohorts; here
#'   the overall effect averages the effect across different groups using group size as weights;
#'   \code{"calendar"} computes average treatment effects across different
#'   time periods, with weights proportional to the group size; here the overall effect averages the effect across each
#'   time period.
#' @param cluster The name of the variable to be used for clustering. The maximum number of cluster variables is 1. Default is \code{NULL}.
#' @param balance_e If set (and if one computes event study), it balances
#'  the sample with respect to event time.  For example, if `balance_e=2`,
#'  `agg_ddd` will drop groups that are not exposed to treatment for
#'  at least three periods, the initial period `e=0` as well as the
#'  next two periods, `e=1` and `e=2`.  This ensures that
#'  the composition of groups does not change when event time changes.
#' @param min_e For event studies, this is the smallest event time to compute
#'  dynamic effects for.  By default, `min_e = -Inf` so that effects at
#'  all lengths of exposure are computed.
#' @param max_e For event studies, this is the largest event time to compute
#'  dynamic effects for.  By default, `max_e = Inf` so that effects at
#'  all lengths of exposure are computed.
#' @param na.rm Logical value if we are to remove missing Values from analyses. Defaults is FALSE.
#' @param boot Boolean for whether or not to compute standard errors using
#'  the multiplier bootstrap.  If standard errors are clustered, then one
#'  must set `boot=TRUE`. Default is value set in the ddd object.  If `boot = FALSE`, then analytical
#'  standard errors are reported.
#' @param nboot The number of bootstrap iterations to use.  The default is the value set in the ddd object,
#'  and this is only applicable if `boot=TRUE`.
#' @param alpha The level of confidence for the confidence intervals.  The default is 0.05. Otherwise, it will
#' use the value set in the ddd object.
#' @param cband Boolean for whether or not to compute a uniform confidence
#'  band that covers all of the group-time average treatment effects
#'  with fixed probability `0.95`.  In order to compute uniform confidence
#'  bands, `boot` must also be set to `TRUE`.  The default is
#'  the value set in the ddd object
#'
#' @return A object (list) of class [`agg_ddd`] that holds the results from the
#'  aggregation step.
#'
#' @examples
#' #----------------------------------------------------------
#' # Triple Diff with multiple time periods
#' #----------------------------------------------------------
#'
#' data <- gen_dgp_mult_periods(size = 500, dgp_type = 1)[["data"]]
#'
#' out <- ddd(yname = "y", tname = "time", idname = "id",
#'             gname = "state", pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
#'             data = data, control_group = "nevertreated", base_period = "varying",
#'             est_method = "dr")
#' # Simple aggregation
#' agg_ddd(out, type = "simple", alpha = 0.10)
#'
#' # Event study aggregation
#' agg_ddd(out, type = "eventstudy", alpha = 0.10)
#'
#' # Group aggregation
#' agg_ddd(out, type = "group", alpha = 0.10)
#'
#' # Calendar aggregation
#' agg_ddd(out, type = "calendar", alpha = 0.10)
#'
#'
#' @export
agg_ddd <- function(ddd_obj,
                   type = "eventstudy",
                   cluster = NULL,
                   balance_e = NULL,
                   min_e = -Inf,
                   max_e = Inf,
                   na.rm = FALSE,
                   boot = NULL,
                   nboot = NULL,
                   cband = NULL,
                   alpha = 0.05) {

 call <- match.call()

 aggte_ddd <- compute_aggregation(ddd_obj = ddd_obj,
                                  type = type,
                                  cluster = cluster,
                                  balance_e = balance_e,
                                  min_e = min_e,
                                  max_e = max_e,
                                  na.rm = na.rm,
                                  boot = boot,
                                  nboot = nboot,
                                  cband = cband,
                                  alpha = alpha
                                )


 ret <- list(aggte_ddd = aggte_ddd,
             call.params = call)

 # define a class
 class(ret) <- "agg_ddd"

 return(ret)

}

# =============================================================================
# Tests for ANALYTICAL cluster-robust standard errors in the MULTI-PERIOD path.
#
# Feature under test (working tree, multi-period only):
#   ddd(..., cluster = "cl", boot = FALSE) should return ANALYTICAL
#   cluster-robust SEs = cluster-sum CRVE on the influence function with 1/n
#   scaling. The ddd object exposes $cluster_vector (per-unit cluster ids aligned
#   to inf_func_mat rows), $cluster_var, $inf_func_mat, $se, $n.
#
# NOTE on reachability:
#   An earlier revision of run_preprocess_multPeriods() FORCED boot = TRUE (and
#   cband = TRUE) whenever a `cluster` was supplied with boot = FALSE, which made
#   the analytic cluster-robust branch in att_gt() (gated on `!boot`) unreachable
#   in the multi-period path. That override has since been REMOVED, so the
#   analytic path is now reached directly by ddd(..., cluster = "cl",
#   boot = FALSE). The analytic_cluster_reachable() guard below is kept as a
#   REGRESSION CHECK: it asserts boot stayed FALSE and no bootstrap ran, and will
#   fail if that override is ever re-introduced.
# =============================================================================

# ---- shared helper: attach a time-invariant cluster column to mult-period data
make_mult_data_with_clusters <- function(seed = 123, size = 600, n_clusters = 30,
                                         dgp_type = 1, balanced = TRUE) {
  set.seed(seed)
  d <- gen_dgp_mult_periods(size = size, dgp_type = dgp_type)[["data"]]
  ids <- sort(unique(d$id))
  if (balanced) {
    # equal-sized clusters: ids spread round-robin so each cluster has ~same size
    cl <- rep(seq_len(n_clusters), length.out = length(ids))
  } else {
    set.seed(seed + 1)
    cl <- sample(seq_len(n_clusters), size = length(ids), replace = TRUE)
  }
  clmap <- data.frame(id = ids, cl = cl)
  merge(d, clmap, by = "id")
}

# Returns TRUE if ddd() actually reached the analytic cluster path (i.e. it did
# NOT silently flip to bootstrap). Asserted below as a regression guard against
# the old preprocessing override that forced boot = TRUE under clustering.
analytic_cluster_reachable <- function(ddd_obj) {
  isFALSE(ddd_obj$argu$boot) && is.null(ddd_obj$bT)
}

# -----------------------------------------------------------------------------
# CHECK 2: Analytic clustered == manual cluster-sum CRVE (tight tolerance).
# -----------------------------------------------------------------------------
test_that("analytic clustered (g,t) SEs equal the manual cluster-sum CRVE", {
  d <- make_mult_data_with_clusters(seed = 123, size = 600, n_clusters = 30)

  o <- suppressWarnings(ddd(
    yname = "y", tname = "time", idname = "id", gname = "state",
    pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
    data = d, control_group = "nevertreated", base_period = "varying",
    est_method = "dr", cluster = "cl", boot = FALSE))

  # Sanity on exposed fields / alignment
  expect_equal(o$cluster_var, "cl")
  expect_false(is.null(o$cluster_vector))
  expect_equal(length(o$cluster_vector), o$n)

  # Analytic cluster path must be reachable: boot stays FALSE, no bootstrap artifacts.
  expect_true(analytic_cluster_reachable(o))

  IF <- as.matrix(o$inf_func_mat)
  cv <- o$cluster_vector
  n  <- o$n
  # cluster-sum CRVE, exactly as specified in the porting guide check 2
  manual <- sqrt(diag(crossprod(rowsum(IF, cv))) / n) / sqrt(n)

  expect_equal(as.numeric(o$se), as.numeric(manual), tolerance = 1e-8)
})

# -----------------------------------------------------------------------------
# CHECK 4: No clustering -> byte-identical to the i.i.d. analytic formula.
# This path is NOT affected by the bug and must pass.
# -----------------------------------------------------------------------------
test_that("no-cluster (g,t) SEs equal the i.i.d. analytic formula from IF", {
  d <- make_mult_data_with_clusters(seed = 321, size = 500, n_clusters = 25)

  o <- ddd(
    yname = "y", tname = "time", idname = "id", gname = "state",
    pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
    data = d, control_group = "nevertreated", base_period = "varying",
    est_method = "dr", cluster = NULL, boot = FALSE)

  # the new cluster branch must be gated OFF when cluster is NULL
  expect_true(is.null(o$cluster_vector))
  expect_false(o$argu$boot)   # analytic path, no bootstrap
  expect_true(is.null(o$bT))

  IF <- as.matrix(o$inf_func_mat)
  n  <- o$n
  iid_a <- sqrt(colSums(IF^2)) / n
  iid_b <- sqrt(diag(crossprod(IF)) / n) / sqrt(n)  # algebraically identical

  expect_equal(as.numeric(o$se), as.numeric(iid_a), tolerance = 1e-10)
  expect_equal(as.numeric(o$se), as.numeric(iid_b), tolerance = 1e-10)
})

# -----------------------------------------------------------------------------
# CHECK 5: NYT / GMM cells.
#   (a) cluster = NULL : single-control cells use the analytic i.i.d. SE
#       sqrt(colSums(IF^2))/n EXACTLY; multi-control GMM cells instead report the
#       per-cell gmm_se = sqrt(1/(n*sum(inv_OMEGA))) override (se_gt_ddd_nyt).
#       We regression-test BOTH facts and confirm NO bootstrap is triggered
#       (this cluster = NULL NYT path is NOT affected by the preprocessing bug).
#   (b) WITH clustering : GMM cells should report the cluster-sum CRVE (matching
#       check 2). Depends on the analytic cluster path -> SKIPPED if unreachable.
#
# Note: the source comment in att_gt() asserts sum(IF_gmm^2)/n == 1/sum(inv_OMEGA),
# i.e. that the i.i.d. formula reproduces gmm_se exactly. In practice the two
# differ slightly for GMM cells (the GMM weights make IF_gmm's raw second moment
# depart from 1/sum(inv_OMEGA)); the reported SE is the gmm_se override, so we
# test against gmm_se on GMM cells and against the i.i.d. formula elsewhere.
# -----------------------------------------------------------------------------
test_that("NYT/GMM cells without clustering: i.i.d. on single-control, gmm_se on GMM", {
  d <- make_mult_data_with_clusters(seed = 99, size = 700, n_clusters = 35)

  o <- ddd(
    yname = "y", tname = "time", idname = "id", gname = "state",
    pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
    data = d, control_group = "notyettreated", base_period = "varying",
    est_method = "dr", cluster = NULL, boot = FALSE)

  expect_false(o$argu$boot)        # analytic, no bootstrap (bug does not touch this path)
  expect_true(is.null(o$bT))
  expect_true(is.null(o$cluster_vector))

  IF <- as.matrix(o$inf_func_mat)
  n  <- o$n
  iid <- sqrt(colSums(IF^2)) / n
  iid[iid <= sqrt(.Machine$double.eps) * 10] <- NA

  ok <- !is.na(o$se) & !is.na(iid)
  # GMM cells = cells whose reported SE differs from the plain i.i.d. formula
  # (these are the not-yet-treated cells with >1 control, carrying the gmm_se
  # override). All other reported SEs must equal the i.i.d. formula EXACTLY.
  is_gmm <- ok & (abs(o$se - iid) > 1e-7)
  is_iid_cell <- ok & !is_gmm

  # single-control cells reproduce the i.i.d. analytic formula exactly
  expect_equal(as.numeric(o$se[is_iid_cell]),
               as.numeric(iid[is_iid_cell]), tolerance = 1e-8)

  # GMM cells carry a distinct, finite, positive override SE that is close to
  # (but not identical to) the i.i.d. value (regression-test of the gmm_se path)
  if (any(is_gmm)) {
    expect_true(all(is.finite(o$se[is_gmm])))
    expect_true(all(o$se[is_gmm] > 0))
    rel <- abs(o$se[is_gmm] - iid[is_gmm]) / iid[is_gmm]
    expect_true(all(rel < 0.05))   # gmm_se override stays in the same ballpark
  } else {
    skip("No multi-control GMM cells present in this design; nothing to assert for the gmm_se override.")
  }
})

test_that("NYT/GMM cells WITH clustering match the manual cluster-sum CRVE", {
  d <- make_mult_data_with_clusters(seed = 99, size = 700, n_clusters = 35)

  o <- suppressWarnings(ddd(
    yname = "y", tname = "time", idname = "id", gname = "state",
    pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
    data = d, control_group = "notyettreated", base_period = "varying",
    est_method = "dr", cluster = "cl", boot = FALSE))

  expect_equal(o$cluster_var, "cl")
  expect_false(is.null(o$cluster_vector))

  # Analytic cluster path must be reachable for GMM cells too.
  expect_true(analytic_cluster_reachable(o))

  IF <- as.matrix(o$inf_func_mat)
  cv <- o$cluster_vector
  n  <- o$n
  manual <- sqrt(diag(crossprod(rowsum(IF, cv))) / n) / sqrt(n)

  ok <- !is.na(o$se) & !is.na(manual)
  expect_equal(as.numeric(o$se[ok]), as.numeric(manual[ok]), tolerance = 1e-8)
})

# -----------------------------------------------------------------------------
# CHECK 1: Analytic == bootstrap (large nboot), at (g,t) level and through
# agg_ddd() (eventstudy + simple). Bootstrap is Monte-Carlo noisy, so a loose
# relative tolerance is used. Depends on the analytic cluster path -> SKIPPED if
# unreachable.
# -----------------------------------------------------------------------------
test_that("analytic clustered SEs approx bootstrap clustered SEs (g,t and agg)", {
  skip_on_cran()
  d <- make_mult_data_with_clusters(seed = 2024, size = 800, n_clusters = 40)

  o_analytic <- suppressWarnings(ddd(
    yname = "y", tname = "time", idname = "id", gname = "state",
    pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
    data = d, control_group = "nevertreated", base_period = "varying",
    est_method = "dr", cluster = "cl", boot = FALSE))

  # Analytic cluster path must be reachable so we can compare against bootstrap.
  expect_true(analytic_cluster_reachable(o_analytic))

  o_boot <- suppressWarnings(ddd(
    yname = "y", tname = "time", idname = "id", gname = "state",
    pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
    data = d, control_group = "nevertreated", base_period = "varying",
    est_method = "dr", cluster = "cl", boot = TRUE, nboot = 5000, cband = TRUE))

  # (g,t)-level comparison on non-missing cells
  ok <- !is.na(o_analytic$se) & !is.na(o_boot$se)
  rel <- abs(o_analytic$se[ok] - o_boot$se[ok]) / o_analytic$se[ok]
  expect_lt(max(rel), 0.10)

  # through agg_ddd(): eventstudy overall + simple overall
  es_a <- agg_ddd(o_analytic, type = "eventstudy")$aggte_ddd
  es_b <- agg_ddd(o_boot,     type = "eventstudy")$aggte_ddd
  expect_lt(abs(es_a$overall.se - es_b$overall.se) / es_a$overall.se, 0.10)

  s_a <- agg_ddd(o_analytic, type = "simple")$aggte_ddd
  s_b <- agg_ddd(o_boot,     type = "simple")$aggte_ddd
  expect_lt(abs(s_a$overall.se - s_b$overall.se) / s_a$overall.se, 0.10)
})

# -----------------------------------------------------------------------------
# CHECK 7: Alignment. Shuffling input rows must not change clustered (g,t) SEs
# (guards the match(id_index, ...) alignment of the cluster vector).
#
# The i.i.d. (cluster = NULL) variant of this check is order-invariant via the
# same alignment code and is NOT affected by the bug, so it is asserted directly.
# The clustered variant is also asserted: even though the reported SEs currently
# come from the bootstrap, the cluster_vector alignment + bootstrap (with a fixed
# seed) must still be invariant to input row order.
# -----------------------------------------------------------------------------
test_that("clustered (g,t) SEs are invariant to input row permutation", {
  d <- make_mult_data_with_clusters(seed = 55, size = 500, n_clusters = 25)

  run <- function(dd) {
    set.seed(1)  # fix seed so any internal bootstrap is reproducible across runs
    suppressWarnings(ddd(
      yname = "y", tname = "time", idname = "id", gname = "state",
      pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
      data = dd, control_group = "nevertreated", base_period = "varying",
      est_method = "dr", cluster = "cl", boot = FALSE))
  }

  o_ordered <- run(d)

  set.seed(777)
  d_shuf <- d[sample(seq_len(nrow(d))), ]
  o_shuffled <- run(d_shuf)

  ok <- !is.na(o_ordered$se) & !is.na(o_shuffled$se)
  expect_equal(as.numeric(o_ordered$se[ok]),
               as.numeric(o_shuffled$se[ok]),
               tolerance = 1e-6)
})

test_that("no-cluster (g,t) SEs are invariant to input row permutation (analytic)", {
  d <- make_mult_data_with_clusters(seed = 56, size = 500, n_clusters = 25)

  run <- function(dd) {
    ddd(yname = "y", tname = "time", idname = "id", gname = "state",
        pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
        data = dd, control_group = "nevertreated", base_period = "varying",
        est_method = "dr", cluster = NULL, boot = FALSE)
  }

  o_ordered <- run(d)
  set.seed(888)
  d_shuf <- d[sample(seq_len(nrow(d))), ]
  o_shuffled <- run(d_shuf)

  ok <- !is.na(o_ordered$se) & !is.na(o_shuffled$se)
  expect_equal(as.numeric(o_ordered$se[ok]),
               as.numeric(o_shuffled$se[ok]),
               tolerance = 1e-9)
})

# Testing if agg_ddd in generating output
test_that("Testing generation of output in aggregation function", {

  data <- gen_dgp_mult_periods(size = 1000, dgp_type = 1)[["data"]]

  # Performing simple tests
  out <- ddd(yname = "y", tname = "time", idname = "id",
              gname = "state", pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
              data = data, control_group = "nevertreated", base_period = "varying",
              est_method = "dr")
  # Simple aggregation
  simple.agg <- agg_ddd(out, type = "simple", alpha = 0.10)

  # Event study aggregation
  es.agg <- agg_ddd(out, type = "eventstudy", alpha = 0.10)

  # Group aggregation
  group.agg <- agg_ddd(out, type = "group", alpha = 0.10)

  # Calendar aggregation
  calendar.agg <- agg_ddd(out, type = "calendar", alpha = 0.10)



  # Performing tests with bootstrap + clustered standard errors
  ddd_gt_boot_cluster <- ddd(yname = "y", tname = "time", idname = "id",
                             gname = "state", pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
                             data = data, control_group = "nevertreated", base_period = "varying",
                             est_method = "dr", boot = TRUE, nboot = 999, cluster = "cluster", cband = TRUE)

  # Simple aggregation boot
  simple.agg.bott <- agg_ddd(ddd_gt_boot_cluster, type = "simple", alpha = 0.10)

  # Event study aggregation boot
  es.agg.boot <- agg_ddd(ddd_gt_boot_cluster, type = "eventstudy", alpha = 0.10)

  # Group aggregation boot
  group.agg.boot <- agg_ddd(ddd_gt_boot_cluster, type = "group", alpha = 0.10)

  # Calendar aggregation boot
  calendar.agg.boot <- agg_ddd(ddd_gt_boot_cluster, type = "calendar", alpha = 0.10)


  # expecting results
  expect_output(summary(simple.agg))
  expect_output(summary(es.agg))
  expect_output(summary(group.agg))
  expect_output(summary(calendar.agg))
  expect_output(summary(simple.agg.bott))
  expect_output(summary(es.agg.boot))
  expect_output(summary(group.agg.boot))
  expect_output(summary(calendar.agg.boot))

})

# This process run all the error check before enter to the data processing

#' @importFrom Rcpp sourceCpp

validate_args_2Periods <- function(args, dta){
  # get args
  yname <- args$yname
  tname <- args$tname
  idname <- args$idname
  gname <- args$gname
  pname <- args$pname
  xformla <- args$xformla
  est_method <- args$est_method
  base_period <- args$base_period
  learners <- args$learners
  n_folds <- args$n_folds
  weightsname <- args$weightsname
  boot <- args$boot
  nboot <- args$nboot
  inffunc <- args$inffunc
  cband <- args$cband
  cluster <- args$cluster

  # flag for boot and cband
  # if ((!boot) && (cband) && (!is.null(cluster))){
  #   stop("Clustered SEs are only available when boot=TRUE. Please, double check your arguments.")
  # }

  # Flag for yname
  if (!is.element(yname, base::colnames(dta))) {
    stop("yname = ",yname,  " could not be found in the data provided.")
  }

  # Flag for tname
  if (!is.element(tname, base::colnames(dta))) {
    stop("tname = ",tname,  " could not be found in the data provided.")
  }

  # check if times periods are numeric
  if (!all(sapply(dta[, ..tname], is.numeric))) {
    stop("tname = ",tname,  " is not numeric. Please convert it")
  }

  # Check if there is only 2 time periods
  tlist <- unique(dta[[tname]])[base::order(unique(dta[[tname]]))]
  if (length(tlist) != 2) {
    stop("The type of ddd specified only allow for two time periods (pre and post). Change type of ddd for multiple time periods")
  }

  # Flag for gname
  if ( !is.element(gname, base::colnames(dta))) {
    stop("gname = ",gname,  " could not be found in the data provided.")
  }

  # Check if there is only 2 groups
  if (dta[, uniqueN(get(gname))] != 2) {
    stop("The type of ddd specified only allow for two groups (treated and untreated). Change type of ddd for multiple groups")
  }



  # Flag for pname
  if ( !is.element(pname, base::colnames(dta))) {
    stop("pname = ",pname,  " could not be found in the data provided.")
  }


  # check if pname is numeric
  if (!all(sapply(dta[, ..pname], is.numeric))) {
    stop("pname = ",pname,  " is not numeric. Please convert it")
  }

  # Check if partition values are binary
  plist <- unique(dta[[pname]])[base::order(unique(dta[[pname]]))]
  if (length(plist) != 2) {
    stop("pname =", pname, " must have only two values (0 and 1). Please check pname")
  }

  # Check if idname is in the data (skip for RCS placeholder)
  if (idname != ".row_id") {
    if ( !is.element(idname, base::colnames(dta))) {
      stop("idname = ",idname,  " could not be found in the data provided.")
    }

    #  check if idname is numeric
    if (!all(sapply(dta[, ..idname], is.numeric))) {
      stop("data[, idname] must be numeric. Please convert it.")
    }
  }

  # Check if any combination of idname and tname is duplicated
  n_id_year = anyDuplicated(dta[, .(idname, tname)])
  # If any combination is duplicated, stop execution and throw an error
  if (n_id_year > 0) {
    stop("The value of idname must be unique (by tname)")
  }

  # Check if partition is unique by idname
  checkPartitionUniqueness(dta, idname, pname)

  # Check if gname is unique by idname
  checkTreatmentUniqueness(dta, idname, gname)

  # Flag for weightsname
  if(!is.null(weightsname)){
    if ( !is.element(weightsname, base::colnames(dta))) {
      stop("weightsname = ",weightsname,  " could not be found in the data provided.")
    }
  }

  # FLAGS FOR DML ESTIMATION
  # if (est_method=="dml") {
  #   if (is.null(learners)) {
  #     stop("learners should be provided when est_method = 'dml'. Please, check documentation.")
  #   }
  #
  #   if (is.null(n_folds)) {
  #     stop("n_folds should be provided when est_method = 'dml'. Please, check documentation.")
  #   }
  #
  #   # check if there 2 learners provided in learners list
  #   if (length(learners) != 2) {
  #     stop("learners must be a list of two learners: ml_pa and ml_md. Please, check documentation.")
  #   }
  #
  #   # check if learner is a classifier type
  #   if (!inherits(learners$ml_pa, "LearnerClassif")) {
  #     stop("The learner must be a classification learner.")
  #   }
  #
  #   # check if learner  is a regression type
  #   if (!inherits(learners$ml_md, "LearnerRegr")) {
  #     stop("The learner must be a regression learner.")
  #   }
  # }

}


validate_args_multPeriods <- function(args, dta){

  # get args
  yname <- args$yname
  tname <- args$tname
  idname <- args$idname
  gname <- args$gname
  control_group <- args$control_group
  pname <- args$pname
  xformla <- args$xformla
  est_method <- args$est_method
  learners <- args$learners
  n_folds <- args$n_folds
  weightsname <- args$weightsname
  boot <- args$boot
  nboot <- args$nboot
  base_period <- args$base_period
  panel <- args$panel

  # For RCS data (panel=FALSE), idname is not required
  if (is.null(panel)) panel <- TRUE  # Default value
  if (!panel && is.null(idname)) {
    return(invisible(NULL))  # Skip validation for RCS without idname
  }

  # Flag for based period: not in c("universal", "varying"), stop
  if (!base_period %in% c("universal", "varying")) {
    stop("base_period must be either 'universal' or 'varying'.")
  }

  # Flag for control group types
  if(!(control_group %in% c("nevertreated","notyettreated"))){
    stop("control_group must be either 'nevertreated' or 'notyettreated'")
  }

  # Flag for yname
  if (!is.element(yname, base::colnames(dta))) {
    stop("yname = ",yname,  " could not be found in the data provided.")
  }

  # Flag for tname
  if (!is.element(tname, base::colnames(dta))) {
    stop("tname = ",tname,  " could not be found in the data provided.")
  }

  # check if times periods are numeric
  if (!all(sapply(dta[, ..tname], is.numeric))) {
    stop("tname = ",tname,  " is not numeric. Please convert it")
  }

  # Flag for gname
  if ( !is.element(gname, base::colnames(dta))) {
    stop("gname = ",gname,  " could not be found in the data provided.")
  }

  # check if gname is numeric
  if (!all(sapply(dta[, ..gname], is.numeric))) {
    stop("gname = ",gname,  " is not numeric. Please convert it")
  }

  # Flag for pname
  if ( !is.element(pname, base::colnames(dta))) {
    stop("pname = ",pname,  " could not be found in the data provided.")
  }

  # check if pname is numeric
  if (!all(sapply(dta[, ..pname], is.numeric))) {
    stop("pname = ",pname,  " is not numeric. Please convert it")
  }

  # Check if partition values are binary
  plist <- unique(dta[[pname]])[base::order(unique(dta[[pname]]))]
  if (length(plist) != 2) {
    stop("pname =", pname, " must have only two values (0 and 1). Please check pname")
  }

  # Check if idname is in the data (skip for RCS placeholder)
  if (idname != ".row_id") {
    if ( !is.element(idname, base::colnames(dta))) {
      stop("idname = ",idname,  " could not be found in the data provided.")
    }

    #  check if idname is numeric
    if (!all(sapply(dta[, ..idname], is.numeric))) {
      stop("data[, idname] must be numeric. Please convert it.")
    }
  }

  # Check if any combination of idname and tname is duplicated
  n_id_year = anyDuplicated(dta[, .(idname, tname)])
  # If any combination is duplicated, stop execution and throw an error
  if (n_id_year > 0) {
    stop("The value of idname must be unique (by tname)")
  }

  # Flag for weightsname
  if(!is.null(weightsname)){
    if ( !is.element(weightsname, base::colnames(dta))) {
      stop("weightsname = ", weightsname, " could not be found in the data provided.")
    }
  }

  # Faster and useful checks to make sure we have a well-balanced panel.
  # Skip these checks for RCS placeholder (column doesn't exist yet)
  if (idname != ".row_id") {
    # Check if partition is unique by idname
    checkPartitionUniqueness(dta, idname, pname)

    # Check if gname is unique by idname
    checkTreatmentUniqueness(dta, idname, gname)
  }

  #TODO: ADD MORE CHECKS IF NEEDED FOR DML

}

#' @title Print
#'
#' @description Prints a agg_ddd Object
#'
#' @param x A agg_ddd object
#' @param ... Other params (required as generic function, but not used)
#' @importFrom utils write.table
#' @export
#' @noRd

print.agg_ddd <- function(x, ...) {

  pointwise_cval <- qnorm(1-x$aggte_ddd$argu$alpha/2)
  overall_cband_upper <- x$aggte_ddd$overall.att + pointwise_cval * x$aggte_ddd$overall.se
  overall_cband_lower <- x$aggte_ddd$overall.att - pointwise_cval * x$aggte_ddd$overall.se
  out1 <- cbind.data.frame(x$aggte_ddd$overall.att, x$aggte_ddd$overall.se, overall_cband_lower, overall_cband_upper)
  out1 <- round(out1, 4)
  overall_sig <- (overall_cband_upper < 0) | (overall_cband_lower > 0)
  overall_sig[is.na(overall_sig)] <- FALSE
  overall_sig_text <- ifelse(overall_sig, "*", "")
  out1 <- cbind.data.frame(out1, overall_sig_text)

  # Printing results in console
  cat(" Call:\n")
  print(x$call.params)
  cat("========================= DDD Aggregation ============================")
  if (x$aggte$type == "simple") cat("\n", "Overall ATT:")
  if (x$aggte$type == "eventstudy") cat("\n", "Overall summary of ATT\'s based on event-study/dynamic aggregation: ")
  if (x$aggte$type == "group") cat("\n", "Overall summary of ATT\'s based on group/cohort aggregation: ")
  if (x$aggte$type == "calendar") cat("\n", "Overall summary of ATT\'s based on calendar time aggregation: ")

  # Front-end Summary Table
  cat("\n")
  header <- c("ATT","   Std. Error", paste0("    [ ",100*(1-x$aggte_ddd$argu$alpha),"% "), "Conf. Int.]","")
  colnames(out1) <- header

  # utils::write.table(format(rbind(header, out1), justify= "centre", digits=2, nsmall=2),
  #                    row.names=FALSE, col.names=FALSE, quote=FALSE, sep=" ")
  print(out1, row.names=FALSE)

  # handle cases depending on type
  if (x$aggte_ddd$type %in% c("group","eventstudy","calendar")) {

    # header
    if (x$aggte_ddd$type=="eventstudy") { c1name <- "Event time"; cat("\n", "Event Study:") }
    if (x$aggte_ddd$type=="group") { c1name <- "Group"; cat("\n", "Group Effects:") }
    if (x$aggte_ddd$type=="calendar") { c1name <- "Time"; cat("\n", "Calendar Effects:") }

    cat("\n")

    # Front-end Summary Table
    lev_conf <- paste0(round(100 * (1 - x$aggte_ddd$argu$alpha), digits = 2), "% ")
    band_method <- ifelse(x$aggte_ddd$argu$cband, "Simult. ", "Ptwise. ")
    interval_text <- paste0("[", lev_conf, band_method)


    # cband_text1a <- paste0(100*(1-x$aggte_ddd$argu$alpha),"% ")
    # # cband_text1b <- ifelse(x$argu$boot,
    # #                        ifelse(x$argu$cband, "Simult. ", "Pointwise "),
    # #                        "Pointwise ")
    # cband_text1b <- "Pointwise "
    # cband_text1 <- paste0("[", cband_text1a, cband_text1b)

    cband_lower <- x$aggte_ddd$att.egt - x$aggte_ddd$crit.val.egt * x$aggte_ddd$se.egt
    cband_upper <- x$aggte_ddd$att.egt + x$aggte_ddd$crit.val.egt * x$aggte_ddd$se.egt

    sig <- (cband_upper < 0) | (cband_lower > 0)
    sig[is.na(sig)] <- FALSE
    sig_text <- ifelse(sig, "*", "")

    out2 <- cbind.data.frame(x$aggte_ddd$egt, x$aggte_ddd$att.egt, x$aggte_ddd$se.egt, cband_lower, cband_upper)
    out2 <- round(out2, 4)
    out2 <- cbind.data.frame(out2, sig_text)

    header2 <- c(c1name, "Estimate","Std. Error", interval_text, "Conf. Band]", "")

    colnames(out2) <- header2
    # utils::write.table(format(rbind(header2, out2), justify= "centre", digits=2, nsmall=2),
    #                    row.names=FALSE, col.names=FALSE, quote=FALSE, sep=" ")
    print(out2, row.names=FALSE, justify = "centre")
  }
  cat("\n")
  cat(" Note: * indicates that the confidence interval does not contain zero.")

  cat("\n --------------------------- Data Info   -----------------------------")
  cat("\n", paste("Outcome variable:", x$aggte_ddd$yname))
  # add partition variable name
  cat("\n", paste("Qualification variable:", x$aggte_ddd$pname))
  ifelse(x$aggte_ddd$control_group == "nevertreated", control_type <- "Never Treated", control_type <- "Not yet Treated (GMM-based)")
  cat("\n", paste("Control group: ", control_type))


  # Analytical vs bootstrapped standard errors
  cat("\n --------------------------- Std. Errors  ----------------------------")
  cat("\n", paste("Level of significance: ", x$aggte_ddd$argu$alpha))
  if (x$aggte_ddd$argu$boot == T) {
    boot1 <-
      cat(
        "\n Boostrapped standard error based on",
        x$aggte_ddd$argu$nboot, "reps.",
        "\n Bootstrap Method: Multiplier bootstrap."
      )
  } else {
    boot1 <- cat("\n Analytical standard errors.")
  }
  if (!is.null(x$aggte_ddd$argu$cluster)){
    cat("\n", paste0("Clustering Std. Errors by: ", x$aggte_ddd$argu$cluster))
  }
  cat("\n =====================================================================")
  cat("\n See Ortiz-Villavicencio and Sant'Anna (2025) for details.")
}

#' @title Print
#'
#' @description Prints a ddd Object
#'
#' @param x A ddd object
#' @param alpha The significance level for confidence intervals (optional)
#' @param ... Other params (required as generic function, but not used)
#' @importFrom utils write.table
#' @export
#' @noRd

print.ddd <- function(x, alpha = NULL, ...) {
  # Update confidence interval if alpha is provided and different from the one used in estimation
  if (!is.null(alpha) && alpha != x$argu$alpha) {
    x$argu$alpha <- alpha
    if (x$argu$cband){
      new_cv <- quantile(x$bT, 1 - alpha, type=1, na.rm = T)
    } else {
      new_cv <- qnorm(1 - alpha / 2)
    }

    x$lci <- x$ATT - new_cv * x$se
    x$uci <- x$ATT + new_cv * x$se
  }

  # Methods used
  if (x$argu$multiple_periods == FALSE){
    if (x$argu$est_method[1] == 'dr') {
      est_method1 <- "DR-DDD estimation for the ATT: \n"
      est_method2 <- "Outcome Regression estimated using: OLS"
      est_method3 <- "Propensity score estimated using: Maximum Likelihood"
    } else if (x$argu$est_method[1] == 'ipw') {
      est_method1 <- "IPW-DDD estimation for the ATT: \n"
      est_method2 <- "Outcome Regression estimated using: N/A"
      est_method3 <- "Propensity score estimated using: Maximum Likelihood"
    } else if (x$argu$est_method[1] == 'reg') {
      est_method1 <- "REG-DDD estimation for the ATT: \n"
      est_method2 <- "Outcome Regression estimated using: OLS"
      est_method3 <- "Propensity score estimated using: N/A"
    } #else if (x$argu$est_method[1] == 'dml') {
    #   est_method1 <- "DML-DDD estimation for the ATT: \n"
    #   est_method2 <- paste0("Outcome Regression estimated using: ", x$argu$learners$ml_md$label)
    #   est_method3 <- paste0("Propensity score estimated using: ", x$argu$learners$ml_pa$label)
    # }

    # Front-end Summary Table
    lev_conf <- paste0(round(100 * (1 - x$argu$alpha), digits = 2), "% ")
    band_method <- ifelse(x$argu$cband, "Simult.", "Ptwise.")
    interval_text <- paste0("[", lev_conf, band_method)
    #header <- c("ATT", "Std. Error", "t value", "Pr(>|t|)", interval_text, "Conf. Band]", "")
    header <- c("  ATT", "  Std. Error", "   Pr(>|t|)", interval_text, "Conf. Band]", "")
    sig <- (x$uci < 0) | (x$lci > 0)
    sig[is.na(sig)] <- FALSE
    sig_text <- ifelse(sig, "*", "")

    body <- cbind(
      sprintf("%10.4f", x$ATT),
      sprintf("%10.4f", x$se),
      #sprintf("%10.4f", x$ATT / x$se),
      sprintf("%10.4f", 2 * (stats::pnorm(-abs(x$ATT / x$se)))),
      sprintf("%10.4f", x$lci),
      sprintf("%10.4f", x$uci),
      sig_text
    )
    colnames(body) <- header
  }

  if (x$argu$multiple_periods == TRUE){
    if (x$argu$est_method[1] == 'dr') {
      est_method1 <- "DR-DDD estimation for the ATT(g,t): \n"
      est_method2 <- "Outcome Regression estimated using: OLS"
      est_method3 <- "Propensity score estimated using: Maximum Likelihood"
    } else if (x$argu$est_method[1] == 'ipw') {
      est_method1 <- "IPW-DDD estimation for the ATT(g,t): \n"
      est_method2 <- "Outcome Regression estimated using: N/A"
      est_method3 <- "Propensity score estimated using: Maximum Likelihood"
    } else if (x$argu$est_method[1] == 'reg') {
      est_method1 <- "REG-DDD estimation for the ATT(g,t): \n"
      est_method2 <- "Outcome Regression estimated using: OLS"
      est_method3 <- "Propensity score estimated using: N/A"
    } #else if (x$argu$est_method[1] == 'dml') {
    #   est_method1 <- "DML-DDD estimation for the ATT(g,t): \n"
    #   est_method2 <- paste("Outcome Regression estimated using:", x$argu$learners$ml_md$label)
    #   est_method3 <- paste("Propensity score estimated using:", x$argu$learners$ml_pa$label)
    # }

    # Front-end Summary Table
    lev_conf <- paste0(round(100 * (1 - x$argu$alpha), digits = 2), "% ")
    band_method <- ifelse(x$argu$cband, "Simult. ", "Pointwise ")
    interval_text <- paste0("[", lev_conf, band_method)
    header <- c("Group", "Time", "ATT(g,t)", "Std. Error", interval_text, "Conf. Band]", "")

    sig <- (x$uci < 0) | (x$lci > 0)
    sig[is.na(sig)] <- FALSE
    sig_text <- ifelse(sig, "*", "")

    body <- cbind.data.frame(
      x$groups,
      x$periods,
      sprintf("%10.4f", x$ATT),
      sprintf("%10.4f", x$se),
      sprintf("%10.4f", x$lci),
      sprintf("%10.4f", x$uci),
      sig_text
    )
    colnames(body) <- header
  }

  # Printing results in console
  cat(" Call:\n")
  print(x$call.params)
  cat("=========================== DDD Summary ==============================")
  cat("\n", est_method1)
  utils::write.table(format(rbind(header, body), justify = "centre", digits = 4, nsmall = 4),
                     row.names=FALSE, col.names=FALSE, quote=FALSE, sep=" ")

  cat("\n")
  cat(" Note: * indicates that the confidence interval does not contain zero.")

  cat("\n --------------------------- Data Info   -----------------------------")
  # Print type of data: if balanced panel, unbalanced panel or RCS
  data_type <- ifelse(x$argu$panel == TRUE, "Panel Data", "Repeated Cross-Sections")
  cat("\n", data_type)
  cat("\n", paste0("Outcome variable: ", x$argu$yname))
  # add partition variable name
  cat("\n", paste0("Qualification variable: ", x$argu$pname))
  # add control group for multiple periods
  if(x$argu$multiple_periods == TRUE){
    ifelse(x$argu$control_group == "nevertreated", control_type <- "Never Treated", control_type <- "Not yet Treated (GMM-based)")
    cat("\n", paste0("Control group: ", control_type))
  }
  if(x$argu$multiple_periods == FALSE){
    cat("\n", "No. of units at each subgroup:")
    cat("\n", paste0("  treated-and-eligible: ", x$subgroup_counts$count[1])) # subgroup 4
    cat("\n", paste0("  treated-but-ineligible: ", x$subgroup_counts$count[2])) # subgroup 3
    cat("\n", paste0("  eligible-but-untreated: ", x$subgroup_counts$count[3])) # subgroup 2
    cat("\n", paste0("  untreated-and-ineligible: ", x$subgroup_counts$count[4])) # subgroup 1
  } else {
    cat("\n", "No. of units per treatment group:")

    n_cohorts <- nrow(x$cohort_size)
    max_display <- 7  # Maximum cohorts to display before truncating

    if (n_cohorts <= max_display) {
      # Display all cohorts when count is manageable
      for (i in 1:n_cohorts) {
        if (x$cohort_size$first_treat[i] == 0) {
          cat("\n", paste0("  Units never enabling treatment: ", x$cohort_size$N[i]), sep = "")
        } else {
          cat("\n", paste0("  Units enabling treatment at period ", x$cohort_size$first_treat[i], ": ", x$cohort_size$N[i]), sep = "")
        }
      }
    } else {
      # Display first 3 cohorts
      for (i in 1:3) {
        if (x$cohort_size$first_treat[i] == 0) {
          cat("\n", paste0("  Units never enabling treatment: ", x$cohort_size$N[i]), sep = "")
        } else {
          cat("\n", paste0("  Units enabling treatment at period ", x$cohort_size$first_treat[i], ": ", x$cohort_size$N[i]), sep = "")
        }
      }

      # Omission message
      n_omitted <- n_cohorts - 4  # 3 shown at top + 1 shown at bottom
      cat("\n", paste0("  ... (", n_omitted, " cohorts omitted for brevity) ..."), sep = "")

      # Display last cohort (typically never-treated)
      if (x$cohort_size$first_treat[n_cohorts] == 0) {
        cat("\n", paste0("  Units never enabling treatment: ", x$cohort_size$N[n_cohorts]), sep = "")
      } else {
        cat("\n", paste0("  Units enabling treatment at period ", x$cohort_size$first_treat[n_cohorts], ": ", x$cohort_size$N[n_cohorts]), sep = "")
      }
    }
  }

  # TODO: ADD number of covariates and some examples

  cat("\n --------------------------- Algorithms ------------------------------")
  # Estimation Method
  cat("\n", est_method2)
  cat("\n", est_method3)

  # if (x$argu$est_method[1] == 'dml'){
  #   cat("\n -------------------------- Cross-fitting  ---------------------------")
  #   cat("\n No. of folds: ", x$argu$n_folds)
  #   cat("\n Apply cross-fitting: TRUE")
  # }
  # Analytical vs bootstrapped standard errors
  cat("\n --------------------------- Std. Errors  ----------------------------")
  cat("\n", paste("Level of significance: ", x$argu$alpha))
  if (x$argu$boot == T) {
    boot1 <-
      cat(
        "\n Boostrapped standard error based on",
        x$argu$nboot, "reps.",
        "\n Method: Multiplier Bootstrap."
      )
  } else {
    boot1 <- cat("\n Analytical standard errors.")
  }
  cat("\n", "Type of confidence band: ", ifelse(x$argu$cband, "Uniform Confidence Band ", "Pointwise Confidence Interval"))
  if (!is.null(x$argu$cluster)){
    cat("\n", paste0("Clustering Std. Errors by: ", x$argu$cluster))
  }
  cat("\n =====================================================================")
  cat("\n See Ortiz-Villavicencio and Sant'Anna (2025) for details.")

}

#' @title Multiplier Bootstrap
#'
#' @description This function take an influence function and use the
#'  multiplier bootstrap to compute standard errors and critical values for
#'  uniform confidence bands.
#'
#' @param inf_func an influence function
#' @param did_preprocessed A `dp` object obtained after preprocess
#' @param use_parallel Boolean of whether or not to use parallel processing in the multiplier
#'  bootstrap, default is \code{use_parallel=FALSE}
#' @param cores the number of cores to use with parallel processing, default is \code{cores=1}
#'
#' @return A list with following elements
#' \item{bres}{results from each bootstrap iteration.}
#' \item{V}{variance matrix.}
#' \item{se}{standard errors.}
#' \item{crit_val}{a critical value for computing uniform confidence bands.}
#'
#' @keywords internal
mboot <- function(inf_func, did_preprocessed, use_parallel = FALSE, cores = 1) {

  # setup needed variables
  data <- did_preprocessed$preprocessed_data
  cluster <- did_preprocessed$cluster
  # Reuse the per-unit cluster vector that att_gt()/compute_aggregation() aligned to the
  # influence-function rows, when available. Falls back to deriving it from the data.
  cluster_vector_in  <- did_preprocessed$cluster_vector
  use_cluster_vector <- !is.null(cluster_vector_in) &&
    length(cluster_vector_in) == NROW(inf_func)
  biters <- did_preprocessed$nboot
  alpha <- did_preprocessed$alpha
  panel <- did_preprocessed$panel
  allow_unbalanced_panel <- did_preprocessed$allow_unbalanced_panel
  tlist <- sort(unique(data$period))

  # Get data for cluster extraction
  # For balanced panels: use first period (all IDs present)
  # For unbalanced panels: get one row per unique ID (not all IDs in first period)
  if (panel && !allow_unbalanced_panel) {
    # Balanced panel: filter to first period works since all IDs are present
    dta <- data[period == tlist[1]]
  } else {
    # Unbalanced panel: need one row per unique ID to match influence function dimensions
    dta <- data[, .SD[1], by = id]
  }

  # Make sure inf_func is matrix because we need this for computing n below
  inf_func <- as.matrix(inf_func)

  # set correct number of units
  n <- nrow(inf_func)

  # multiplier bootstrap
  n_clusters <- n
  if (length(cluster)==0) {
    bres <- sqrt(n) * run_multiplier_bootstrap(inf_func, biters, use_parallel, cores)
  } else {
    # Cluster-robust multiplier bootstrap, Callaway & Sant'Anna (2021, Remark 10):
    # cluster-specific multipliers on the influence function aggregated to cluster SUMS.
    if (use_cluster_vector) {
      cluster_vec <- cluster_vector_in
    } else {
      unique_clusters_df <- unique(dta[, c("id", "cluster")])
      cluster_vec <- unique_clusters_df$cluster
    }
    n_clusters      <- length(unique(cluster_vec))
    cluster_sum_if  <- rowsum(inf_func, cluster_vec, reorder = TRUE)   # SUMS, not means
    bres <- sqrt(n_clusters) * run_multiplier_bootstrap(cluster_sum_if, biters, use_parallel, cores)
  }


  if (isTRUE(class(bres) == "numeric")) bres <- as.matrix(bres)

  # Non-degenerate dimensions
  ndg.dim <- (!is.na(colSums(bres))) & (base::colSums(bres^2) > sqrt(.Machine$double.eps)*10)
  bres <- as.matrix(bres[ , ndg.dim])

  # bootstrap variance matrix (this matrix can be defective because of degenerate cases)
  V <- cov(bres)
  # bootstrap standard error
  bSigma <- apply(bres, 2,
                  function(b) (quantile(b, .75, type=1, na.rm = T) -
                                 quantile(b, .25, type=1, na.rm = T))/(qnorm(.75) - qnorm(.25)))

  bSigma[bSigma <= sqrt(.Machine$double.eps)*10] <- NA

  # critical value for uniform confidence band
  bT <- base::suppressWarnings(apply(bres, 1, function(b) max(abs(b / bSigma), na.rm = T)))
  bT <- bT[is.finite(bT)]
  crit_val <- quantile(bT, 1-alpha, type=1, na.rm = T) # uniform critical value
  se <- rep(NA, length(ndg.dim))
  # With cluster SUMS this is the cluster-robust SE (1/n)*sqrt(sum_c S_c^2); without
  # clustering n_clusters == n and it reduces to the i.i.d. bSigma/sqrt(n). (Equivalent
  # to the old bSigma/sqrt(n_clusters) only when clusters are equal-sized.)
  se[ndg.dim] <- as.numeric(bSigma) * sqrt(n_clusters) / n

  return(list(bres = bres, V = V, se = se, bT= bT, unif_crit_val = crit_val))
}

run_multiplier_bootstrap <- function(inf_func, biters, use_parallel = FALSE, cores = 1) {
  ngroups = ceiling(biters/cores)
  chunks = rep(ngroups, cores)
  # Round down so you end up with the right number of biters
  chunks[1] = chunks[1] + biters - sum(chunks)

  n <- nrow(inf_func)
  parallel.function <- function(biters) {
    BMisc::multiplier_bootstrap(inf_func, biters)
  }
  # From tests, this is about where it becomes worth it to parallelize
  if(n > 2500 & use_parallel == TRUE & cores > 1) {
    results = parallel::mclapply(
      chunks,
      FUN = parallel.function,
      mc.cores = cores
    )
    results = do.call(rbind, results)
  } else {
    results = BMisc::multiplier_bootstrap(inf_func, biters)
  }
  return(results)
}

# =============================================================================
# Tests that the clustered multiplier bootstrap actually implements the
# Callaway & Sant'Anna (2021, Remark 10) cluster-SUM aggregation on the regime
# where it MATTERS: UNBALANCED clusters.
#
# Why this file exists (adversarial gap found in the original suite):
#   test-mboot-cluster.R only exercises EQUAL-SIZED clusters, where cluster
#   means and cluster sums coincide. On equal-sized clusters the new (sum-based)
#   and old (mean-based) implementations are numerically identical, so reverting
#   Feature 2 entirely is a SURVIVING MUTANT against that file. These tests fail
#   if mboot() is reverted to cluster MEANS or to the old bSigma/sqrt(n_clusters)
#   scaling, because on unbalanced clusters sums != means.
# =============================================================================

# self-contained helpers (do NOT depend on test-mboot-cluster.R internals) -----
.make_dp_unbal <- function(cl, nboot) {
  n  <- length(cl)
  dt <- data.table::data.table(id = seq_len(n), cluster = cl)
  dt2 <- rbind(data.table::copy(dt)[, period := 1L],
               data.table::copy(dt)[, period := 2L])
  list(preprocessed_data = dt2, cluster = "cluster", cluster_vector = cl,
       nboot = nboot, alpha = 0.05, panel = TRUE, allow_unbalanced_panel = FALSE)
}
.iqr_sd <- function(bres) apply(bres, 2, function(b)
  (stats::quantile(b, .75, type = 1, na.rm = TRUE) -
     stats::quantile(b, .25, type = 1, na.rm = TRUE)) /
    (stats::qnorm(.75) - stats::qnorm(.25)))

# unbalanced cluster structure shared by the tests below
.unbalanced_clusters <- function() {
  nc <- 20L
  sizes <- c(rep(2L, 10L), rep(22L, 10L))   # 10 tiny + 10 large clusters
  rep(seq_len(nc), times = sizes)           # n = 240, deliberately lumpy
}

# -----------------------------------------------------------------------------
# PRIMARY GUARD: on UNBALANCED clusters mboot() must equal the cluster-SUM
# definition AND differ materially from the old cluster-MEAN definition.
# (Kills: revert of cluster sums -> means, and/or SE scaling -> /sqrt(nc).)
# -----------------------------------------------------------------------------
test_that("UNBALANCED clusters: mboot uses cluster SUMS (Remark 10), not means", {
  skip_on_cran()
  set.seed(101)
  cl <- .unbalanced_clusters()
  n  <- length(cl); k <- 4L; nc <- length(unique(cl))
  IF <- matrix(stats::rnorm(n * k), n, k)
  dp <- .make_dp_unbal(cl, nboot = 6000)

  set.seed(2024)
  r <- mboot(IF, dp)

  # (1) implemented SE == cluster-SUM definition  bSigma_sum * sqrt(nc) / n
  csum <- rowsum(IF, cl, reorder = TRUE)
  set.seed(2024)
  bres_sum <- sqrt(nc) * BMisc::multiplier_bootstrap(csum, 6000)
  se_sum   <- as.numeric(.iqr_sd(bres_sum)) * sqrt(nc) / n
  expect_equal(as.numeric(r$se), se_sum, tolerance = 1e-10)

  # (2) implemented SE is MATERIALLY DIFFERENT from the old cluster-MEAN form
  #     bSigma_mean / sqrt(nc)  -- this is what makes the test informative.
  cmean <- rowsum(IF, cl, reorder = TRUE) / as.vector(table(cl))
  set.seed(2024)
  bres_mean <- sqrt(nc) * BMisc::multiplier_bootstrap(cmean, 6000)
  se_mean   <- as.numeric(.iqr_sd(bres_mean)) / sqrt(nc)
  rel <- abs(as.numeric(r$se) - se_mean) / se_mean
  expect_gt(max(rel), 0.05)   # sums vs means must diverge on unbalanced clusters
})

# -----------------------------------------------------------------------------
# INTEGRATION: end-to-end ddd() on unbalanced clusters -- analytic cluster-sum
# CRVE must approximate the clustered bootstrap (which derives clusters from the
# data independently). This is the unbalanced analogue of test-cluster-analytic
# CHECK 1, run on the regime the port actually changes.
# -----------------------------------------------------------------------------
test_that("UNBALANCED clusters: analytic clustered SEs approx bootstrap clustered SEs", {
  skip_on_cran()
  set.seed(123)
  d   <- gen_dgp_mult_periods(size = 800, dgp_type = 1)[["data"]]
  ids <- sort(unique(d$id))
  set.seed(124)
  cl  <- sample(seq_len(40), length(ids), replace = TRUE,
                prob = c(rep(8, 4), rep(1, 36)))   # unbalanced
  d   <- merge(d, data.frame(id = ids, cl = cl), by = "id")

  o_a <- suppressWarnings(ddd(
    yname = "y", tname = "time", idname = "id", gname = "state",
    pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
    data = d, control_group = "nevertreated", base_period = "varying",
    est_method = "dr", cluster = "cl", boot = FALSE))
  expect_false(isTRUE(o_a$argu$boot))   # analytic path reached

  o_b <- suppressWarnings(ddd(
    yname = "y", tname = "time", idname = "id", gname = "state",
    pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
    data = d, control_group = "nevertreated", base_period = "varying",
    est_method = "dr", cluster = "cl", boot = TRUE, nboot = 6000, cband = TRUE))

  ok  <- !is.na(o_a$se) & !is.na(o_b$se)
  rel <- abs(o_a$se[ok] - o_b$se[ok]) / o_a$se[ok]
  expect_lt(stats::median(rel), 0.10)

  # and the analytic clustered SEs must NOT collapse to the i.i.d. SEs:
  o_iid <- ddd(
    yname = "y", tname = "time", idname = "id", gname = "state",
    pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
    data = d, control_group = "nevertreated", base_period = "varying",
    est_method = "dr", cluster = NULL, boot = FALSE)
  ok2 <- !is.na(o_a$se) & !is.na(o_iid$se)
  expect_gt(max(abs(o_a$se[ok2] - o_iid$se[ok2]) / o_iid$se[ok2]), 0.02)
})

# =============================================================================
# Tests for the clustered multiplier bootstrap mboot() (CS 2021, Remark 10).
#
# Feature under test:
#   mboot() now aggregates the influence function to cluster SUMS and scales the
#   SE by bSigma * sqrt(n_clusters) / n. For EQUAL-SIZED clusters of size m this
#   must coincide with the pre-change cluster-MEANS result (bSigma_mean / sqrt(nc)),
#   because cluster_sum = m * cluster_mean and the multiplier bootstrap is linear
#   in the influence-function columns:
#
#       se_sum  = bSigma_sum  * sqrt(nc) / n
#               = (m * bSigma_mean) * sqrt(nc) / (m * nc)
#               = bSigma_mean / sqrt(nc)
#               = se_mean.
#
#   These tests exercise mboot() directly on a synthetic influence function so
#   the identity is checked deterministically (fixed multiplier seed), free of
#   the multi-period preprocessing path.
# =============================================================================

# small helper to build the minimal did_preprocessed list mboot() consumes
make_dp_for_mboot <- function(cl, nboot) {
  n  <- length(cl)
  dt <- data.table::data.table(id = seq_len(n), cluster = cl)
  dt2 <- rbind(data.table::copy(dt)[, period := 1L],
               data.table::copy(dt)[, period := 2L])
  list(preprocessed_data = dt2,
       cluster = "cluster",
       cluster_vector = cl,
       nboot = nboot,
       alpha = 0.05,
       panel = TRUE,
       allow_unbalanced_panel = FALSE)
}

# IQR-based robust sd, identical to the one used inside mboot()
iqr_sd <- function(bres) {
  apply(bres, 2, function(b)
    (stats::quantile(b, .75, type = 1, na.rm = TRUE) -
       stats::quantile(b, .25, type = 1, na.rm = TRUE)) /
      (stats::qnorm(.75) - stats::qnorm(.25)))
}

# -----------------------------------------------------------------------------
# CHECK 3 (scaling identity): for equal-sized clusters of size m with nc clusters
# and n = m * nc units, the SE scaling sqrt(nc)/n equals 1/(m*sqrt(nc)). This is
# the algebraic backbone of "sum-based == mean-based" for balanced clusters.
# -----------------------------------------------------------------------------
test_that("equal-sized clusters: SE scaling sqrt(nc)/n == 1/(m*sqrt(nc))", {
  for (m in c(2L, 5L, 10L)) {
    nc <- 30L
    n  <- m * nc
    expect_equal(sqrt(nc) / n, 1 / (m * sqrt(nc)), tolerance = 1e-12)
  }
})

# -----------------------------------------------------------------------------
# CHECK 3 (main): the Remark-10 cluster-SUM mboot SE equals the cluster-MEANS SE
# for equal-sized clusters, to machine precision, using the same multiplier seed.
# -----------------------------------------------------------------------------
test_that("equal-sized clusters: cluster-SUM mboot SE == cluster-MEANS SE", {
  skip_on_cran()
  set.seed(1)
  n  <- 240L; k <- 4L; m <- 6L
  nc <- n / m
  IF <- matrix(stats::rnorm(n * k), n, k)
  cl <- rep(seq_len(nc), each = m)          # exactly equal-sized clusters

  dp <- make_dp_for_mboot(cl, nboot = 4000)

  # mboot() performs the cluster-SUM (Remark 10) computation
  set.seed(123)
  r_sum <- mboot(IF, dp)

  # manual cluster-MEANS computation with the SAME multiplier draws
  cmean <- rowsum(IF, cl) / m
  set.seed(123)
  bres_mean <- sqrt(nc) * BMisc::multiplier_bootstrap(cmean, 4000)
  se_mean   <- as.numeric(iqr_sd(bres_mean)) / sqrt(nc)

  expect_equal(as.numeric(r_sum$se), se_mean, tolerance = 1e-10)
})

# -----------------------------------------------------------------------------
# CHECK 3 (cross-check): the cluster-SUM mboot SE also equals the manual
# cluster-SUM SE bSigma_sum * sqrt(nc) / n (definition of the implemented path).
# -----------------------------------------------------------------------------
test_that("cluster-SUM mboot SE matches its own definition bSigma*sqrt(nc)/n", {
  skip_on_cran()
  set.seed(2)
  n  <- 300L; k <- 3L; m <- 5L
  nc <- n / m
  IF <- matrix(stats::rnorm(n * k), n, k)
  cl <- rep(seq_len(nc), each = m)

  dp <- make_dp_for_mboot(cl, nboot = 3000)

  set.seed(321)
  r_sum <- mboot(IF, dp)

  csum <- rowsum(IF, cl)
  set.seed(321)
  bres_sum <- sqrt(nc) * BMisc::multiplier_bootstrap(csum, 3000)
  se_sum   <- as.numeric(iqr_sd(bres_sum)) * sqrt(nc) / n

  expect_equal(as.numeric(r_sum$se), se_sum, tolerance = 1e-10)
})

# -----------------------------------------------------------------------------
# Sanity: with one unit per cluster (n_clusters == n), the cluster bootstrap
# reduces to the i.i.d. multiplier bootstrap (sqrt(n) * MB(IF)), so the SE equals
# bSigma / sqrt(n).
# -----------------------------------------------------------------------------
test_that("one-unit-per-cluster reduces to the i.i.d. multiplier bootstrap SE", {
  skip_on_cran()
  set.seed(3)
  n  <- 250L; k <- 3L
  IF <- matrix(stats::rnorm(n * k), n, k)
  cl <- seq_len(n)                          # each unit is its own cluster

  dp <- make_dp_for_mboot(cl, nboot = 3000)

  set.seed(555)
  r_cluster <- mboot(IF, dp)

  # i.i.d. bootstrap with same seed: bres = sqrt(n) * MB(IF), se = bSigma / sqrt(n)
  set.seed(555)
  bres_iid <- sqrt(n) * BMisc::multiplier_bootstrap(IF, 3000)
  se_iid   <- as.numeric(iqr_sd(bres_iid)) / sqrt(n)

  expect_equal(as.numeric(r_cluster$se), se_iid, tolerance = 1e-10)
})

#' @title Summary
#' @description Summary of a ddd object
#' @param object A ddd object
#' @param alpha The significance level for confidence intervals (optional)
#' @param ... Other params (required as generic function, but not used)
#' @export
#' @noRd
# Define new summary function
summary.ddd <- function(object, alpha = NULL, ...) {
  ddd.obj <- object
  print(ddd.obj, alpha = alpha)
}

#' Doubly robust DDD estimator for ATT, with repeated cross-section data and 2 periods
#'
#' This function implements a doubly robust estimator for assessing the average
#' treatment effect on the treated (ATT) using a triple differences (DDD) approach
#' in repeated cross-section data settings across two time periods. The function takes preprocessed
#' data structured specifically for this analysis.
#'
#' @import stats
#' @param did_preprocessed A list containing preprocessed data and specifications for the DDD estimation.
#'        Expected elements include:
#'        - `preprocessed_data`: A data table containing the data with variables needed for the analysis.
#'        - `est_method`: The estimation method to be used. Default is \code{est_method = "dr"}.
#'        - `xformula`: The formula for the covariates to be included in the model. It should be of the form \code{~ x1 + x2}.
#'        Default is \code{xformla = ~1} (no covariates).
#'        - `boot`: Logical. If \code{TRUE}, the function use the multiplier bootstrap to compute standard errors. Default is \code{FALSE}.
#'        - `nboot`: The number of bootstrap samples to be used. Default is \code{NULL}. If \code{boot = TRUE}, the default is \code{nboot = 999}.
#'        - `subgroup_counts`: A matrix containing the number of observations in each subgroup.
#'        - `alpha` The level of significance for the confidence intervals.  Default is \code{0.05}.
#'        - `inffunc`: Logical. If \code{TRUE}, the function returns the influence function. Default is \code{FALSE}.
#'        - `use_parallel`: Boolean of whether or not to use parallel processing in the multiplier bootstrap, default is \code{use_parallel=FALSE}
#'        - `cores`: the number of cores to use with parallel processing, default is \code{cores=1}
#'        - `cband`: Boolean of whether or not to compute simultaneous confidence bands, default is \code{cband=FALSE}
#'
#' @keywords internal
#' @return A list with the estimated ATT, standard error, upper and lower confidence intervals, and influence function.
#' @export
att_dr_rc <- function(did_preprocessed) {

  data <- did_preprocessed$preprocessed_data
  est_method <- did_preprocessed$est_method
  xformula <- did_preprocessed$xformula
  boot <- did_preprocessed$boot
  nboot <- did_preprocessed$nboot
  alpha <- did_preprocessed$alpha
  use_parallel <- did_preprocessed$use_parallel # to perform bootstrap
  cores <- did_preprocessed$cores # to perform bootstrap
  cband <- did_preprocessed$cband # to perform bootstrap + simult. conf. band
  inffunc <- did_preprocessed$inffunc # flag to return influence function
  subgroup_counts <- did_preprocessed$subgroup_counts
  #n <- did_preprocessed$n # total number of unique IDs

  # --------------------------------------------------------------------
  # Compute ATT
  # --------------------------------------------------------------------

  # Pre-compute propensity scores for each subgroup
  if (est_method == "reg"){
    # set pscores properly such that weights for control are zero
    pscores <- lapply(c(3, 2, 1), function(condition_subgroup) {
      compute_pscore_null_rc(data, condition_subgroup)
    })
  } else {
    pscores <- lapply(c(3, 2, 1), function(condition_subgroup) {
      compute_pscore_rc(data, condition_subgroup, xformula)
    })
  }

  if (est_method == "ipw"){
    # set or_delta equal to zero
    reg_adjust <- lapply(c(3, 2, 1), function(condition_subgroup) {
      compute_outcome_regression_null_rc(data, condition_subgroup)
    })
  } else {
    # Pre-compute the regression adjustment for each subgroup
    reg_adjust <- lapply(c(3, 2, 1), function(condition_subgroup) {
      compute_outcome_regression_rc(data, condition_subgroup, xformula)
    })
  }

  # Compute Doubly Robust Triple Difference Estimator
  dr_att_inf_func_3 <- compute_did_rc(data, condition_subgroup = 3, pscores = pscores, reg_adjustment = reg_adjust, xformula = xformula, est_method = est_method) # S=g, Q=1 vs. S=g, Q=0
  dr_att_inf_func_2 <- compute_did_rc(data, condition_subgroup = 2, pscores = pscores, reg_adjustment = reg_adjust, xformula = xformula, est_method = est_method) # S=g, Q=1 vs. S=\infty, Q=1
  dr_att_inf_func_1 <- compute_did_rc(data, condition_subgroup = 1, pscores = pscores, reg_adjustment = reg_adjust, xformula = xformula, est_method = est_method) # S=g, Q=1 vs. S=\infty, Q=0

  dr_ddd <- dr_att_inf_func_3$dr_att + dr_att_inf_func_2$dr_att - dr_att_inf_func_1$dr_att

  # For influence function weighting, we need to use the appropriate counts depending on data structure
  # For RCS: use observation counts (each observation is independent)
  # For unbalanced panel: use observation counts too (IFs are at observation level, aggregation by ID happens later)
  # Note: subgroup_counts contains either observation counts (RCS) or unique ID counts (unbalanced panel)
  # depending on how it was constructed in preprocessing

  # Get the appropriate sample sizes for weighting
  # These are used to weight the influence functions from the three DiD components
  n <- nrow(data)  # Total observations in this (g,t) cell
  n3 <- data[subgroup %in% c(3, 4), .N]  # Observations in DiD3 comparison
  n2 <- data[subgroup %in% c(2, 4), .N]  # Observations in DiD2 comparison
  n1 <- data[subgroup %in% c(1, 4), .N]  # Observations in DiD1 comparison

  w3 <- n/n3
  w2 <- n/n2
  w1 <- n/n1
  # rescaling influence function (still at observation level)
  inf_func = w3*dr_att_inf_func_3$inf_func + w2*dr_att_inf_func_2$inf_func - w1*dr_att_inf_func_1$inf_func

  #-----------------------------------------------------------------------------
  # compute confidence intervals / bands
  #-----------------------------------------------------------------------------

  if (boot){
    # perform multiplier bootstrap
    boot_result <- mboot(inf_func, did_preprocessed=did_preprocessed, use_parallel=use_parallel, cores=cores)
    se_ddd <- boot_result$se # save bootstrap standard error
    bT <- boot_result$bT # save sup-t confidence band
    if (cband){

      # get critical value to compute uniform confidence bands
      cv <- boot_result$unif_crit_val
      if(cv >= 7){
        warning("Simultaneous critical value is arguably `too large' to be reliable. This usually happens when number of observations per group is small and/or there is no much variation in outcomes.")
      }

    } else {
      # use regular critical value
      cv <- qnorm(1-alpha/2)
    }
    # Computing uniform confidence bands
    # Estimate of upper boundary of 1-alpha% confidence band
    ci_upper <- dr_ddd + cv * se_ddd
    # Estimate of lower boundary of 1-alpha% confidence band
    ci_lower <- dr_ddd - cv * se_ddd
  } else {
    # compute point-wise confidence intervals
    se_ddd <- stats::sd(inf_func)/sqrt(n)
    bT <- NULL
    # use regular critical value
    cv <- qnorm(1-alpha/2)
    # estimate upper bound at 1 - alpha% confidence level
    ci_upper <- dr_ddd + cv * se_ddd
    # estimate lower bound at 95% confidence level
    ci_lower <- dr_ddd - cv * se_ddd
  }

  # ------------------------------------------------------------------------------
  # Return results
  # ------------------------------------------------------------------------------

  # Return null if inffunc is FALSE
  if (inffunc == FALSE){
    inf_func <- NULL
  }

  ret <- (list(ATT = dr_ddd,
              se = se_ddd,
              uci = ci_upper,
              lci = ci_lower,
              nboot = nboot,
              bT = bT,
              inf_func = inf_func,
              subgroup_counts = subgroup_counts
              ))

  return(ret)

}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

mboot_did <- function(linrep, nboot) {
    .Call('_triplediff_mboot_did', PACKAGE = 'triplediff', linrep, nboot)
}

checkPartitionUniqueness <- function(dta, idname, partitionName) {
    invisible(.Call('_triplediff_checkPartitionUniqueness', PACKAGE = 'triplediff', dta, idname, partitionName))
}

checkTreatmentUniqueness <- function(dta, idname, treatName) {
    invisible(.Call('_triplediff_checkTreatmentUniqueness', PACKAGE = 'triplediff', dta, idname, treatName))
}

checkWeightsUniqueness <- function(dta, idname) {
    invisible(.Call('_triplediff_checkWeightsUniqueness', PACKAGE = 'triplediff', dta, idname))
}

#' Function to generate a fake dataset for testing purposes only.
#' @description
#' Function to generate fake dataset to test internal procedures.
#'
#' @param seed Seed for reproducibility
#' @param num_ids Number of IDs
#' @param time Number of time periods
#' @param initial.year Initial year
#' @param treatment.year Treatment year
#'
#' @return A data.table with the following columns:
#'
#' - id: ID
#' - state: State variable
#' - year: Time variable
#' - partition: Partition variable
#' - x1: Covariate 1
#' - x2: Covariate 2
#' - treat: Treatment variable
#' - outcome: Outcome variable
#'
#' @keywords internal
generate_test_panel <- function(seed = 123,
                                num_ids = 100,
                                time = 2,
                                initial.year = 2019,
                                treatment.year = 2020) {
  # Flag for initial year and treatment year
  if (initial.year >= treatment.year) {
    stop("The initial year must be less than the treatment year")
  }

  if (time < 2) {
    stop("The time variable must be greater than 1")
  }

  if (num_ids < 1) {
    stop("The number of IDs must be greater than 0")
  }

  if ((treatment.year - initial.year +1 ) > time){
    stop("The difference between the treatment year and the initial year must be less or equal than the time variable")
  }

  # Set seed for reproducibility
  set.seed(seed)

  # Number of IDs
  num_ids <- num_ids

  # Number of observations per ID
  num_obs <- time

  # Generate IDs
  id = rep(1:num_ids, each = num_obs)

  # Generate x1 and x2 for each state
  x1 <- rep(rnorm(num_ids), each = num_obs)
  x2 <- rep(rnorm(num_ids), each = num_obs)

  # Generate state variable
  state = rep(sample(c("A", "B", "C", "D"), num_ids, replace = TRUE), each = num_obs)

  # generate time variable
  year = rep(seq(initial.year, initial.year+(time-1), 1), times = num_ids)

  # Generate partition variable
  partition = rep(rbinom(num_ids, 1, 0.5), each = num_obs)

  # Generate target group
  s_2 = as.numeric(state %in% c("A", "B"))

  # Generate post treatment variable
  t_2 = as.numeric(year >= treatment.year)

  # TWFE outcome variable
  y = 2 + 5*s_2 - 2 * partition + 3 * t_2 +
    4*s_2*partition + 2*s_2*t_2 + 3*partition*t_2 +
    1*s_2*partition*t_2 + rnorm(num_ids*num_obs)

  # Create a data.table
  dt <- data.table::data.table(
    id = id,
    state = state,
    year = year,
    partition = partition,
    x1 = x1,
    x2 = x2,
    treat = s_2,
    #post = t_2,
    outcome = y
  )
  return(dt)
}

# ---------------------------------------------------------------------
# Functions to generate data for 2 time periods
# ---------------------------------------------------------------------

fps <- function(psi_index, coefs, xvars){
  elem_wise_mult <- xvars %*% coefs
  return(psi_index * elem_wise_mult)
}

fps2 <- function(psi_index, coefs, xvars, c){
  elem_wise_mult <- c + xvars %*% coefs
  return(psi_index * elem_wise_mult)
}

freg <- function(coefs, xvars){
  return(210 + xvars %*% coefs)
}

get_long_data <- function(y_list, state, partition, covs){
  # ensure that y_list is a list of vectors
  if (!is.list(y_list) || any(!sapply(y_list, is.vector))) {
    stop("y_list must be a list of vectors")
  }

  # ensure that all y vectors have the same length
  y_length <- sapply(y_list, length)
  if (length(unique(y_length)) != 1) {
    stop("All y vectors must have the same length")
  }

  # check if covs has 4 columns
  if(ncol(covs) != 4){
    stop("covs must have 4 columns")
  }

  n <- y_length[1]
  t <- length(y_list)

  dt <- data.table(
    id = rep(1:n, t),
    state = rep(state, t),
    partition = rep(partition, t),
    time = rep(1:t, each = n),
    y = unlist(y_list),
    cov1 = rep(covs[, 1], t),
    cov2 = rep(covs[, 2], t),
    cov3 = rep(covs[, 3], t),
    cov4 = rep(covs[, 4], t)
  )

  return(dt)
}

# Mean and Std deviation of Z's without truncation
mean.z1 <- exp(0.25/2)
sd.z1 <- sqrt((exp(0.25) - 1) * exp(0.25))

mean.z2 <- 10
sd.z2 <- 0.54164

mean.z3 <- 0.21887
sd.z3 <-   0.04453

mean.z4 <- 402
sd.z4 <-  56.63891

# ---------------------------------------------------------------------
# Generate data for 2 time treatment periods
# ---------------------------------------------------------------------

#' Function that generates panel data with single treatment date assignment and two time periods.
#' @description
#' Generate panel data with a single treatment date and two periods
#'
#' @param size Integer. Number of units.
#' @param dgp_type Integer in \{1,2,3,4\}.
#'   1 = both nuisance functions correct;
#'   2 = only the outcome model correct;
#'   3 = only the propensity score correct;
#'   4 = both nuisance functions incorrect.
#'
#' @return A list with the following elements:
#' \describe{
#'   \item{data}{A \code{data.table} in long format with columns:
#'     \itemize{
#'       \item \code{id}: unit identifier
#'       \item \code{state}: state variable
#'       \item \code{time}: time variable
#'       \item \code{partition}: partition assignment
#'       \item \code{x1}, \code{x2}, \code{x3}, \code{x4}: covariates
#'       \item \code{y}: outcome variable
#'       \item \code{cluster}: cluster ID (no within-cluster correlation)
#'     }
#'   }
#'   \item{att}{True average treatment effect on the treated (ATT), set to 0.}
#'   \item{att.unf}{Oracle ATT computed under the unfeasible specification.}
#'   \item{eff}{Theoretical efficiency bound for the estimator.}
#' }
#'
#' @export
gen_dgp_2periods <- function(size, dgp_type){
  # Generate data for MC simulations
  # size: number of observations
  # dgp_type: type of DGP
  # Returns a data table in long format

  att <- 0
  # Gen covariates
  x1 <- stats::rnorm(size, mean = 0, sd = 1)
  x2 <- stats::rnorm(size, mean = 0, sd = 1)
  x3 <- stats::rnorm(size, mean = 0, sd = 1)
  x4 <- stats::rnorm(size, mean = 0, sd = 1)

  z1_tilde <- exp(x1/2)
  z2_tilde <- x2/(1 + exp(x1)) + 10
  z3_tilde <- (x1 * x3/25 + 0.6)^3
  z4_tilde <- (x1 + x4 + 20)^2


  z1 <- (z1_tilde - mean.z1) / sd.z1
  z2 <- (z2_tilde - mean.z2) / sd.z2
  z3 <- (z3_tilde - mean.z3) / sd.z3
  z4 <- (z4_tilde - mean.z4) / sd.z4

  x <- cbind(x1, x2, x3, x4)
  z <- cbind(z1, z2, z3, z4)

  w1 <- c(-1, .5, -.25, -.1)
  w2 <- c(-0.5, 2, .5, -.2)
  w3 <- c(3, -1.5, +.75, -.3)

  b1 <- c(27.4, 13.7, 13.7, 13.7)
  b2 <- c(6.85, 3.43, 3.43, 3.43)

  if(dgp_type == 1){
    # both models are functions of Z
    # index for pscores
    fps1 <- fps(0.2, w1, z)
    fps2 <- fps(0.2, w2, z)
    fps3 <- fps(0.05, w3, z)

    # index for outcome regression model
    freg1 <- freg(b1, z)
    freg0 <- freg(b2, z)
    eff <- 32.82

    # Generate groups
    exp_f1 <- exp(fps1)
    exp_f2 <- exp(fps2)
    exp_f3 <- exp(fps3)

    sum_exp_f <- exp_f1 + exp_f2 + exp_f3

    p1 <- exp_f1 / (1 + sum_exp_f)
    p2 <- exp_f2 / (1 + sum_exp_f)
    p3 <- exp_f3 / (1 + sum_exp_f)
    p4 <- 1 - p1 - p2 - p3

    # determine treatment group PA based on U and cumulative probabilities

    U <- runif(size)

    PA <- sapply(1:size, function(i) {
      if (U[i] <= p1[i]) {
        return(1)
      } else if (U[i] <= p1[i] + p2[i]) {
        return(2)
      } else if (U[i] <= 1 - p4[i]) {
        return(3)
      } else {
        return(4)
      }
    })

    # Generate state variable
    state <- ifelse(PA == 3 | PA == 4, 1, 0)

    # Generate partition
    partition <- ifelse(PA == 2 | PA == 4, 1, 0)

    # Generate potential outcomes

    unobs_het <- state*partition*freg1 + (1 - state)*partition*freg0
    or_lin <- state*freg1 + (1 - state)*freg0
    # v is the unobserved heterogeneity
    v <- stats::rnorm(size, mean = unobs_het, sd = 1)

    #Gen realized outcome at time 0
    y0 <- or_lin + v + stats::rnorm(size, mean = 0, sd = 1)

    # gen outcomes at time 1
    # First let's generate potential outcomes: y_1_potential
    y10 <- or_lin + v + stats::rnorm(size, mean = 0, sd = 1) + # This is the baseline
      or_lin # this is for the trend based on X

    y11 <- or_lin + v + stats::rnorm(size, mean = 0, sd = 1) +#This is the baseline
      or_lin + #this is for the trend based on X
      att # This is the treatment effects

    # Get unfeasible att
    att.unf <- (mean(state*partition*y11) - mean(state*partition*y10))/mean(state*partition)

    # Now we generate the realized outcomes
    y1 <- state*partition*y11 + (1 - state*partition)*y10

    # Generate the data assuming we always observe Z
    dt <- get_long_data(list(as.vector(y0), as.vector(y1)), state, partition, z)
    setorder(dt, "id", "time")
    # generate clusters (there's no actual within-cluster correlation) with sample with replacement by id
    # get unique ids
    unique_ids <- unique(dt$id)
    # sample cluster values for each unique id
    clusters <- data.table(id = unique_ids, cluster = sample(1:50, size = length(unique_ids), replace = TRUE))
    # merge the sampled clusters back into the original data table
    dt <- merge(dt, clusters, by = "id", all.x = TRUE)

    return(list(data = dt, att = 0, att.unf = att.unf, eff = eff))


  } else if(dgp_type == 2){
    # Pscore depends on X but Regressions depend on Z
    # index for pscores
    fps1 <- fps(0.2, w1, x)
    fps2 <- fps(0.2, w2, x)
    fps3 <- fps(0.05, w3, x)

    # index for outcome regression model
    freg1 <- freg(b1, z)
    freg0 <- freg(b2, z)
    eff <- 32.52

    # Generate groups
    exp_f1 <- exp(fps1)
    exp_f2 <- exp(fps2)
    exp_f3 <- exp(fps3)

    sum_exp_f <- exp_f1 + exp_f2 + exp_f3

    p1 <- exp_f1 / (1 + sum_exp_f)
    p2 <- exp_f2 / (1 + sum_exp_f)
    p3 <- exp_f3 / (1 + sum_exp_f)
    p4 <- 1 - p1 - p2 - p3

    # determine treatment group PA based on U and cumulative probabilities

    U <- runif(size)

    PA <- sapply(1:size, function(i) {
      if (U[i] <= p1[i]) {
        return(1)
      } else if (U[i] <= p1[i] + p2[i]) {
        return(2)
      } else if (U[i] <= 1 - p4[i]) {
        return(3)
      } else {
        return(4)
      }
    })

    # Generate state variable
    state <- ifelse(PA == 3 | PA == 4, 1, 0)

    # Generate partition
    partition <- ifelse(PA == 2 | PA == 4, 1, 0)

    # Generate potential outcomes

    unobs_het <- state*partition*freg1 + (1 - state)*partition*freg0
    or_lin <- state*freg1 + (1 - state)*freg0
    # v is the unobserved heterogeneity
    v <- stats::rnorm(size, mean = unobs_het, sd = 1)

    #Gen realized outcome at time 0
    y0 <- or_lin + v + stats::rnorm(size, mean = 0, sd = 1)

    # gen outcomes at time 1
    # First let's generate potential outcomes: y_1_potential
    y10 <- or_lin + v + stats::rnorm(size, mean = 0, sd = 1) + # This is the baseline
      or_lin # this is for the trend based on X

    y11 <- or_lin + v + stats::rnorm(size, mean = 0, sd = 1) +#This is the baseline
      or_lin + #this is for the trend based on X
      att # This is the treatment effects

    # Get unfeasible att
    att.unf <- (mean(state*partition*y11) - mean(state*partition*y10))/mean(state*partition)

    # Now we generate the realized outcomes
    y1 <- state*partition*y11 + (1 - state*partition)*y10

    # Generate the data assuming we always observe Z
    dt <- get_long_data(list(as.vector(y0), as.vector(y1)), state, partition, z)
    setorder(dt, "id", "time")
    # generate clusters (there's no actual within-cluster correlation) with sample with replacement by id
    # get unique ids
    unique_ids <- unique(dt$id)
    # sample cluster values for each unique id
    clusters <- data.table(id = unique_ids, cluster = sample(1:50, size = length(unique_ids), replace = TRUE))
    # merge the sampled clusters back into the original data table
    dt <- merge(dt, clusters, by = "id", all.x = TRUE)

    return(list(data = dt, att = 0, att.unf = att.unf, eff = eff))


  } else if(dgp_type == 3){
    # Pscore depends on Z but Regressions depend on X
    # index for pscores
    fps1 <- fps(0.2, w1, z)
    fps2 <- fps(0.2, w2, z)
    fps3 <- fps(0.05, w3, z)

    # index for outcome regression model
    freg1 <- freg(b1, x)
    freg0 <- freg(b2, x)
    eff <- 32.82

    # Generate groups
    exp_f1 <- exp(fps1)
    exp_f2 <- exp(fps2)
    exp_f3 <- exp(fps3)

    sum_exp_f <- exp_f1 + exp_f2 + exp_f3

    p1 <- exp_f1 / (1 + sum_exp_f)
    p2 <- exp_f2 / (1 + sum_exp_f)
    p3 <- exp_f3 / (1 + sum_exp_f)
    p4 <- 1 - p1 - p2 - p3

    # determine treatment group PA based on U and cumulative probabilities

    U <- runif(size)

    PA <- sapply(1:size, function(i) {
      if (U[i] <= p1[i]) {
        return(1)
      } else if (U[i] <= p1[i] + p2[i]) {
        return(2)
      } else if (U[i] <= 1 - p4[i]) {
        return(3)
      } else {
        return(4)
      }
    })

    # Generate state variable
    state <- ifelse(PA == 3 | PA == 4, 1, 0)

    # Generate partition
    partition <- ifelse(PA == 2 | PA == 4, 1, 0)

    # Generate potential outcomes

    unobs_het <- state*partition*freg1 + (1 - state)*partition*freg0
    or_lin <- state*freg1 + (1 - state)*freg0
    # v is the unobserved heterogeneity
    v <- stats::rnorm(size, mean = unobs_het, sd = 1)

    #Gen realized outcome at time 0
    y0 <- or_lin + v + stats::rnorm(size, mean = 0, sd = 1)

    # gen outcomes at time 1
    # First let's generate potential outcomes: y_1_potential
    y10 <- or_lin + v + stats::rnorm(size, mean = 0, sd = 1) + # This is the baseline
      or_lin # this is for the trend based on X

    y11 <- or_lin + v + stats::rnorm(size, mean = 0, sd = 1) +#This is the baseline
      or_lin + #this is for the trend based on X
      att # This is the treatment effects

    # Get unfeasible att
    att.unf <- (mean(state*partition*y11) - mean(state*partition*y10))/mean(state*partition)

    # Now we generate the realized outcomes
    y1 <- state*partition*y11 + (1 - state*partition)*y10

    # Generate the data assuming we always observe Z
    dt <- get_long_data(list(as.vector(y0), as.vector(y1)), state, partition, z)
    setorder(dt, "id", "time")
    # generate clusters (there's no actual within-cluster correlation) with sample with replacement by id
    # get unique ids
    unique_ids <- unique(dt$id)
    # sample cluster values for each unique id
    clusters <- data.table(id = unique_ids, cluster = sample(1:50, size = length(unique_ids), replace = TRUE))
    # merge the sampled clusters back into the original data table
    dt <- merge(dt, clusters, by = "id", all.x = TRUE)

    return(list(data = dt, att = 0, att.unf = att.unf, eff = eff))

  } else if (dgp_type == 4){
    # Both models are functions of X (misspecified)
    # index for pscores
    fps1 <- fps(0.2, w1, x)
    fps2 <- fps(0.2, w2, x)
    fps3 <- fps(0.05, w3, x)

    # index for outcome regression model
    freg1 <- freg(b1, x)
    freg0 <- freg(b2, x)
    eff <- 32.52

    # Generate groups
    exp_f1 <- exp(fps1)
    exp_f2 <- exp(fps2)
    exp_f3 <- exp(fps3)

    sum_exp_f <- exp_f1 + exp_f2 + exp_f3

    p1 <- exp_f1 / (1 + sum_exp_f)
    p2 <- exp_f2 / (1 + sum_exp_f)
    p3 <- exp_f3 / (1 + sum_exp_f)
    p4 <- 1 - p1 - p2 - p3

    # determine treatment group PA based on U and cumulative probabilities

    U <- runif(size)

    PA <- sapply(1:size, function(i) {
      if (U[i] <= p1[i]) {
        return(1)
      } else if (U[i] <= p1[i] + p2[i]) {
        return(2)
      } else if (U[i] <= 1 - p4[i]) {
        return(3)
      } else {
        return(4)
      }
    })

    # Generate state variable
    state <- ifelse(PA == 3 | PA == 4, 1, 0)

    # Generate partition
    partition <- ifelse(PA == 2 | PA == 4, 1, 0)

    # Generate potential outcomes

    unobs_het <- state*partition*freg1 + (1 - state)*partition*freg0
    or_lin <- state*freg1 + (1 - state)*freg0
    # v is the unobserved heterogeneity
    v <- stats::rnorm(size, mean = unobs_het, sd = 1)

    #Gen realized outcome at time 0
    y0 <- or_lin + v + stats::rnorm(size, mean = 0, sd = 1)

    # gen outcomes at time 1
    # First let's generate potential outcomes: y_1_potential
    y10 <- or_lin + v + stats::rnorm(size, mean = 0, sd = 1) + # This is the baseline
      or_lin # this is for the trend based on X

    y11 <- or_lin + v + stats::rnorm(size, mean = 0, sd = 1) +#This is the baseline
      or_lin + #this is for the trend based on X
      att # This is the treatment effects

    # Get unfeasible att
    att.unf <- (mean(state*partition*y11) - mean(state*partition*y10))/mean(state*partition)

    # Now we generate the realized outcomes
    y1 <- state*partition*y11 + (1 - state*partition)*y10

    # Generate the data assuming we always observe Z
    dt <- get_long_data(list(as.vector(y0), as.vector(y1)), state, partition, z)
    setorder(dt, "id", "time")
    # generate clusters (there's no actual within-cluster correlation) with sample with replacement by id
    # get unique ids
    unique_ids <- unique(dt$id)
    # sample cluster values for each unique id
    clusters <- data.table(id = unique_ids, cluster = sample(1:50, size = length(unique_ids), replace = TRUE))
    # merge the sampled clusters back into the original data table
    dt <- merge(dt, clusters, by = "id", all.x = TRUE)

    return(list(data = dt, att = 0, att.unf = att.unf, eff = eff))

  } else {
    stop("Invalid DGP type")
  }

}

# ---------------------------------------------------------------------
# Functions to generate data for multiple time periods
# ---------------------------------------------------------------------

#' Generate panel data with staggered treatment adoption (three periods)
#' @description
#' Generate panel data where units adopt treatment at different times across three periods.
#'
#' @param size Integer. Number of units to simulate.
#' @param dgp_type Integer in \{1,2,3,4\}. Only used when \code{include_covariates = TRUE}.
#'   1 = both nuisance functions correct (default);
#'   2 = only the outcome model correct;
#'   3 = only the propensity-score model correct;
#'   4 = both nuisance functions misspecified.
#' @param include_covariates Logical. If \code{TRUE} (default), generates covariates with
#'   transformations and uses \code{dgp_type} specification. If \code{FALSE}, uses constant
#'   covariates and fixed propensity score probabilities for a simpler DGP.
#'
#' @return A named list with components:
#' \describe{
#'   \item{data}{A \code{data.table} in long format with columns:
#'     \itemize{
#'       \item \code{id}: unit identifier
#'       \item \code{cohort}: first period when treatment is assigned
#'       \item \code{partition}: partition indicator
#'       \item \code{x1}, \code{x2}, \code{x3}, \code{x4}: covariates
#'       \item \code{cluster}: cluster identifier (no within-cluster correlation)
#'       \item \code{time}: time period index
#'       \item \code{y}: observed outcome
#'     }
#'   }
#'   \item{data_wide}{A \code{data.table} in wide format (one row per \code{id}) with columns:
#'     \itemize{
#'       \item \code{id}, \code{cohort}, \code{partition}, \code{x1}, \code{x2}, \code{x3}, \code{x4}, \code{cluster}
#'       \item \code{y_t0}, \code{y_t1}, \code{y_t2}: outcomes in periods 0, 1, and 2
#'     }
#'   }
#'   \item{ES_0_unf}{(Only if \code{include_covariates = TRUE}) Unfeasible (oracle) event-study parameter at time 0.}
#'   \item{prob_g2_p1}{(Only if \code{include_covariates = TRUE}) Proportion of units with \code{cohort == 2} and eligibility in period 1.}
#'   \item{prob_g3_p1}{(Only if \code{include_covariates = TRUE}) Proportion of units with \code{cohort == 3} and eligibility in period 1.}
#' }
#'
#' @export
gen_dgp_mult_periods <- function(size, dgp_type = 1, include_covariates = TRUE){

  # Validate parameters
  if (!include_covariates && dgp_type != 1) {
    warning("dgp_type is ignored when include_covariates = FALSE. Using simple DGP specification.")
  }

  # Generates covariates
  Xsi.ps = 0.4 # This scale things in pscore

  if (include_covariates) {
    x1 <- stats::rnorm(size, mean = 0, sd = 1)
    x2 <- stats::rnorm(size, mean = 0, sd = 1)
    x3 <- stats::rnorm(size, mean = 0, sd = 1)
    x4 <- stats::rnorm(size, mean = 0, sd = 1)
  } else {
    x1 <- rep(1, size)
    x2 <- rep(1, size)
    x3 <- rep(1, size)
    x4 <- rep(1, size)
  }

  x <- cbind(x1, x2, x3, x4)

  # Transform to Z only if covariates are included
  if (include_covariates) {
    z1_tilde <- exp(x1/2)
    z2_tilde <- x2/(1 + exp(x1)) + 10
    z3_tilde <- (x1 * x3/25 + 0.6)^3
    z4_tilde <- (x1 + x4 + 20)^2

    z1 <- (z1_tilde - mean.z1) / sd.z1
    z2 <- (z2_tilde - mean.z2) / sd.z2
    z3 <- (z3_tilde - mean.z3) / sd.z3
    z4 <- (z4_tilde - mean.z4) / sd.z4

    z <- cbind(z1, z2, z3, z4)
  } else {
    # For no covariates, z = x (constants)
    z <- x
  }

  w1 <- c(-1, 0.5, -0.25, -0.1) # for (2,A), (2,B)
  w2 <- c(-0.5, 1, -0.1, -0.25) # for (3,A), (3,B)
  w3 <- c(-0.25, 0.1, -1, -0.1) # for (0,A), (0,B)

  b1 <- c(27.4, 13.7, 13.7, 13.7)

  #-----------------------------------------------------------------------------
  # Generate groups and partitions
  # DGP has 3 timing groups, G=0,2,3, three timing periods, t=1,2,3,
  # and two partitions for each group, P=0,1 (P=0 is not eligible for treatment)
  # Thus, in total, we have 6 Group-Partition combinations

  if (!include_covariates) {
    #---------------------------------------------------------------------------
    # SIMPLE DGP: No covariates (constants only)
    #---------------------------------------------------------------------------
    # Use fixed probability matrix for propensity scores
    probs_pscore <- matrix(cbind(0.20, 0.15, 0.30, 0.20, 0.05, 0.10),
                           nrow=size, ncol=6, byrow = TRUE)
    colnames(probs_pscore) <- c("pi_2A", "pi_2B", "pi_3A", "pi_3B", "pi_0A", "pi_0B")

    # Sample group types
    group_types <- apply(probs_pscore, 1,
                         function(pvec) sample(x = seq(1,6), size=1, prob=pvec))

    # Translate group types into partitions and cohorts
    partition <- ifelse(group_types %in% c(1,3,5), 1, 0)
    cohort <- ifelse(group_types %in% c(1,2),
                     2,
                     ifelse(group_types %in% c(3,4),
                            3,
                            0))

    #---------------------------------------------------------------------------
    # Create indexes for potential outcomes (simple specification)
    #---------------------------------------------------------------------------
    # Index for PO (using simpler linear index without Z transformation)
    index_lin <- 210 + 27.4*x1 + 13.7*(x2 + x3 + x4)
    index_partition <- partition * index_lin
    # Index for unobserved heterogeneity
    index_unobs_het <- cohort * index_lin + index_partition
    # Index for Conditional Parallel Trends
    index_trend <- index_lin
    # Generate unobserved heterogeneity
    v <- stats::rnorm(size, mean = index_unobs_het, sd = 1)
    # Index to violates the parallel trends assumption
    index_pt_violation <- v/10

    # Index for per-period ATT(g,t)
    index_att_g2 <- 10
    index_att_g3 <- 25

    #---------------------------------------------------------------------------
    # Generate the potential outcomes
    #---------------------------------------------------------------------------
    # Baseline index
    baseline_t1 <- index_lin + index_partition + v

    # Generate untreated outcome at time 1
    y_t1 <- baseline_t1 + stats::rnorm(size)

    # Generate potential outcomes at time 2
    baseline_t2 <- baseline_t1 +
      index_pt_violation + # Violate the CPT
      index_trend # trend based on X

    y_t2_never <- baseline_t2 + stats::rnorm(size)
    y_t2_g2 <- baseline_t2 + stats::rnorm(size) +
      index_att_g2 * partition # This is the treatment effects for cohort 2

    # Generate potential outcomes at time 3
    baseline_t3 <- baseline_t1 +
      2*index_trend + # this is for the trend based on X 
      2*index_pt_violation # Violate the CPT

    y_t3_never <- baseline_t3 + stats::rnorm(size)
    y_t3_g2 <- baseline_t3 + stats::rnorm(size) +
      2*index_att_g2 * partition # This is the treatment effects for cohort 2 at time 3
    y_t3_g3 <- baseline_t3 + stats::rnorm(size) +
      index_att_g3 * partition # This is the treatment effects for cohort 3 at time 3

    #---------------------------------------------------------------------------
    # Get the realized outcomes
    #---------------------------------------------------------------------------
    y_t2 <- ifelse(cohort == 2 & partition==1, y_t2_g2, y_t2_never)
    y_t3 <- ifelse(cohort == 2 & partition==1, y_t3_g2,
                   ifelse(cohort == 3 & partition==1, y_t3_g3,
                          y_t3_never))

    #---------------------------------------------------------------------------
    # Put all data in a data frame
    #---------------------------------------------------------------------------
    dta_wide <- data.frame(cbind(id = 1:size,
                                 y_t1, y_t2 , y_t3,
                                 cohort , partition,
                                 x))
    # Transform data into long format
    dta <- get_long_data(list(as.vector(y_t1), as.vector(y_t2), as.vector(y_t3)), cohort, partition, x)
    setorder(dta, "id", "time")
    # generate clusters
    unique_ids <- unique(dta$id)
    clusters <- data.table(id = unique_ids, cluster = sample(1:50, size = length(unique_ids), replace = TRUE))
    dta <- merge(dta, clusters, by = "id", all.x = TRUE)

    #---------------------------------------------------------------------------
    # Return the data (no unfeasible parameters for simple DGP)
    #---------------------------------------------------------------------------
    return(list(data = as.data.table(dta),
                data_wide = as.data.table(dta_wide)))

  } else if(dgp_type == 1){
    # both models are functions of Z
    # index for pscores

    # Generalize Propensity score (3 groups and two partitions)
    pi_2A <- exp(fps2(Xsi.ps, w1, z, 1.25))
    pi_2B <- exp(fps2(-Xsi.ps, w1, z, -0.5))
    pi_3A <- exp(fps2(Xsi.ps, w2, z, 2))
    pi_3B <- exp(fps2(-Xsi.ps, w2, z, -1.25))
    pi_0A <- exp(fps2(Xsi.ps, w3, z, -0.5))
    # Get the sum of the exponentials
    sum_pi <- 1 + pi_2A + pi_2B + pi_3A + pi_3B + pi_0A
    # Construct the generalized pscore
    pi_2A <- pi_2A/sum_pi
    pi_2B <- pi_2B/sum_pi
    pi_3A <- pi_3A/sum_pi
    pi_3B <- pi_3B/sum_pi
    pi_0A <- pi_0A/sum_pi
    pi_0B <- 1 - (pi_2A + pi_2B + pi_3A + pi_3B + pi_0A)

    # Bind them into a matrix
    probs_pscore <- cbind(pi_2A, pi_2B, pi_3A, pi_3B, pi_0A, pi_0B)

    # Sample group types (1 and 3 are the group of interest)
    group_types <- apply(probs_pscore, 1,
                         function(pvec) sample(seq(1,6), size=1, prob=pvec))

    # Translate group types into partitions and cohorts
    partition <- ifelse(group_types %in% c(1,3,5), 1, 0)
    cohort <- ifelse(group_types %in% c(1,2),
                     2,
                     ifelse(group_types %in% c(3,4),
                            3,
                            0))
    #-----------------------------------------------------------------------------
    # Now create the indexes for the potential outcomes
    #-----------------------------------------------------------------------------
    # Index for PO
    index_lin <- freg(b1, z)
    index_partition <- partition * index_lin
    # Index for unobserved heterogeneity
    index_unobs_het <- cohort * index_lin + index_partition
    #Index for Conditional Parallel Trends (this does not vary with treatment group)
    index_trend <- index_lin
    # Generate unobserved heterogeneity
    # (fixed effects that are correlated with X, partition, and cohort)
    v <- stats::rnorm(size, mean = index_unobs_het, sd = 1)
    # Index to violates the parallel trends assumption
    index_pt_violation <- v/10

    # Index for per-period ATT(g,t)
    index_att_g2 <- 10
    index_att_g3 <- 25
    #-----------------------------------------------------------------------------
    # Generate the potential outcomes
    # Baseline index
    baseline_t1 <- index_lin + index_partition + v

    #Generate untreated outcome at time 1 (that is observed for all units)
    y_t1 <- baseline_t1 + stats::rnorm(size)

    # Generate potential outcomes at time 2
    baseline_t2 <- baseline_t1 +
      index_pt_violation + #Violate the CPT
      index_trend #this is for the trend based on X

    y_t2_never <- baseline_t2 + stats::rnorm(size)
    y_t2_g2 <- baseline_t2 + stats::rnorm(size) +
      index_att_g2 * partition # This is the treatment effects for cohort 2
    y_t2_g3 <-  baseline_t2 + stats::rnorm(size)

    # Generate potential outcomes at time 3
    baseline_t3 <- baseline_t1 +
      2*index_trend + #this is for the trend based on X
      2*index_pt_violation #Violate the CPT

    y_t3_never <- baseline_t3 + stats::rnorm(size)
    y_t3_g2 <- baseline_t3 + stats::rnorm(size) +
      2*index_att_g2 * partition # This is the treatment effects for cohort 2
    y_t3_g3 <- baseline_t3 + stats::rnorm(size) +
      index_att_g3 * partition # This is the treatment effects for cohort 2
    #-----------------------------------------------------------------------------
    # Get the realized outcomes now
    # y_t2<- y_t2_g2*(cohort == 2) + y_t2_g3*(cohort==3) + y_t2_never*(cohort==0)
    # y_t3<- y_t3_g2*(cohort == 2) + y_t3_g3*(cohort==3) + y_t3_never*(cohort==0)

    y_t2<-  ifelse(cohort == 2 & partition==1, y_t2_g2, y_t2_never)
    y_t3<- ifelse(cohort == 2 & partition==1, y_t3_g2,
                  ifelse(cohort == 3 & partition==1, y_t3_g3,
                         y_t3_never))
    #-----------------------------------------------------------------------------
    # Get the event study parameter
    # Get unfeasible ATT(g,t) (assuming we see the Potential outcomes)
    att_g2_t2_unf <- (mean((group_types==1)*y_t2_g2) - mean((group_types==1)*y_t2_never))/
      mean((group_types==1))
    att_g2_t3_unf <- (mean((group_types==1)*y_t3_g2) - mean((group_types==1)*y_t3_never))/
      mean((group_types==1))
    att_g3_t2_unf <- (mean((group_types==3)*y_t2_g3) - mean((group_types==3)*y_t2_never))/
      mean((group_types==3))
    att_g3_t3_unf <- (mean((group_types==3)*y_t3_g3) - mean((group_types==3)*y_t3_never))/
      mean((group_types==3))

    # Get unfeasible Event-Study (ES) (assuming we see the Potential outcomes and pscores)
    # get weights
    prob_g2_p1 <- mean(pi_2A/(pi_2A + pi_3A))
    prob_g3_p1 <- mean(pi_3A/(pi_2A + pi_3A))
    # Get ES
    ES_0_unf <- att_g2_t2_unf * prob_g2_p1 + att_g3_t3_unf * prob_g3_p1
    ES_1_unf <- att_g2_t3_unf
    # ------------------------------------------------------------------------------
    # Put all data in a data frame
    dta_wide <- data.frame(cbind(id = 1:size,
                                 y_t1 = y_t1,
                                 y_t2 = y_t2,
                                 y_t3 = y_t3,
                                 cohort = cohort,
                                 partition = partition,
                                 z = z)) # we always observe z
    # Transform data into long format
    # Generate the data assuming we always observe Z
    dta <- get_long_data(list(as.vector(y_t1), as.vector(y_t2), as.vector(y_t3)), cohort, partition, z)
    setorder(dta, "id", "time")
    # generate clusters (there's no actual within-cluster correlation) with sample with replacement by id
    # get unique ids
    unique_ids <- unique(dta$id)
    # sample cluster values for each unique id
    clusters <- data.table(id = unique_ids, cluster = sample(1:50, size = length(unique_ids), replace = TRUE))
    # merge the sampled clusters back into the original data table
    dta <- merge(dta, clusters, by = "id", all.x = TRUE)
    #-----------------------------------------------------------------------------
    # Return the data
    return(list(data = as.data.table(dta),
                data_wide = as.data.table(dta_wide),
                ES_0_unf = ES_0_unf,
                prob_g2_p1 = prob_g2_p1,
                prob_g3_p1 = prob_g3_p1))

  } else if(dgp_type == 2){
    # Pscore depends on X but Regressions depend on Z
    # index for pscores

    # Generalize Propensity score (3 groups and two partitions)
    pi_2A <- exp(fps2(Xsi.ps, w1, x, 1.25))
    pi_2B <- exp(fps2(-Xsi.ps, w1, x, -0.5))
    pi_3A <- exp(fps2(Xsi.ps, w2, x, 2))
    pi_3B <- exp(fps2(-Xsi.ps, w2, x, -1.25))
    pi_0A <- exp(fps2(Xsi.ps, w3, x, -0.5))
    # Get the sum of the exponentials
    sum_pi <- 1 + pi_2A + pi_2B + pi_3A + pi_3B + pi_0A
    # Construct the generalized pscore
    pi_2A <- pi_2A/sum_pi
    pi_2B <- pi_2B/sum_pi
    pi_3A <- pi_3A/sum_pi
    pi_3B <- pi_3B/sum_pi
    pi_0A <- pi_0A/sum_pi
    pi_0B <- 1 - (pi_2A + pi_2B + pi_3A + pi_3B + pi_0A)

    # Bind them into a matrix
    probs_pscore <- cbind(pi_2A, pi_2B, pi_3A, pi_3B, pi_0A, pi_0B)

    # Sample group types (1 and 3 are the group of interest)
    group_types <- apply(probs_pscore, 1,
                         function(pvec) sample(seq(1,6), size=1, prob=pvec))

    # Translate group types into partitions and cohorts
    partition <- ifelse(group_types %in% c(1,3,5), 1, 0)
    cohort <- ifelse(group_types %in% c(1,2),
                     2,
                     ifelse(group_types %in% c(3,4),
                            3,
                            0))
    #-----------------------------------------------------------------------------
    # Now create the indexes for the potential outcomes
    #-----------------------------------------------------------------------------
    # Index for PO
    index_lin <- freg(b1, z)
    index_partition <- partition * index_lin
    # Index for unobserved heterogeneity
    index_unobs_het <- cohort * index_lin + index_partition
    #Index for Conditional Parallel Trends (this does not vary with treatment group)
    index_trend <- index_lin
    # Generate unobserved heterogeneity
    # (fixed effects that are correlated with X, partition, and cohort)
    v <- stats::rnorm(size, mean = index_unobs_het, sd = 1)
    # Index to violates the parallel trends assumption
    index_pt_violation <- v/10

    # Index for per-period ATT(g,t)
    index_att_g2 <- 10
    index_att_g3 <- 25
    #-----------------------------------------------------------------------------
    # Generate the potential outcomes
    # Baseline index
    baseline_t1 <- index_lin + index_partition + v

    #Generate untreated outcome at time 1 (that is observed for all units)
    y_t1 <- baseline_t1 + stats::rnorm(size)

    # Generate potential outcomes at time 2
    baseline_t2 <- baseline_t1 +
      index_pt_violation + #Violate the CPT
      index_trend #this is for the trend based on X

    y_t2_never <- baseline_t2 + stats::rnorm(size)
    y_t2_g2 <- baseline_t2 + stats::rnorm(size) +
      index_att_g2 * partition # This is the treatment effects for cohort 2
    y_t2_g3 <-  baseline_t2 + stats::rnorm(size)

    # Generate potential outcomes at time 3
    baseline_t3 <- baseline_t1 +
      2*index_trend + #this is for the trend based on X
      2*index_pt_violation #Violate the CPT

    y_t3_never <- baseline_t3 + stats::rnorm(size)
    y_t3_g2 <- baseline_t3 + stats::rnorm(size) +
      2*index_att_g2 * partition # This is the treatment effects for cohort 2
    y_t3_g3 <- baseline_t3 + stats::rnorm(size) +
      index_att_g3 * partition # This is the treatment effects for cohort 2
    #-----------------------------------------------------------------------------
    # Get the realized outcomes now
    # y_t2<- y_t2_g2*(cohort == 2) + y_t2_g3*(cohort==3) + y_t2_never*(cohort==0)
    # y_t3<- y_t3_g2*(cohort == 2) + y_t3_g3*(cohort==3) + y_t3_never*(cohort==0)

    y_t2<-  ifelse(cohort == 2 & partition==1, y_t2_g2, y_t2_never)
    y_t3<- ifelse(cohort == 2 & partition==1, y_t3_g2,
                  ifelse(cohort == 3 & partition==1, y_t3_g3,
                         y_t3_never))
    #-----------------------------------------------------------------------------
    # Get the event study parameter
    # Get unfeasible ATT(g,t) (assuming we see the Potential outcomes)
    att_g2_t2_unf <- (mean((group_types==1)*y_t2_g2) - mean((group_types==1)*y_t2_never))/
      mean((group_types==1))
    att_g2_t3_unf <- (mean((group_types==1)*y_t3_g2) - mean((group_types==1)*y_t3_never))/
      mean((group_types==1))
    att_g3_t2_unf <- (mean((group_types==3)*y_t2_g3) - mean((group_types==3)*y_t2_never))/
      mean((group_types==3))
    att_g3_t3_unf <- (mean((group_types==3)*y_t3_g3) - mean((group_types==3)*y_t3_never))/
      mean((group_types==3))

    # Get unfeasible Event-Study (ES) (assuming we see the Potential outcomes and pscores)
    # get weights
    prob_g2_p1 <- mean(pi_2A/(pi_2A + pi_3A))
    prob_g3_p1 <- mean(pi_3A/(pi_2A + pi_3A))
    # Get ES
    ES_0_unf <- att_g2_t2_unf * prob_g2_p1 + att_g3_t3_unf * prob_g3_p1
    ES_1_unf <- att_g2_t3_unf
    # ------------------------------------------------------------------------------
    # Put all data in a data frame
    dta_wide <- data.frame(cbind(id = 1:size,
                                 y_t1, y_t2 , y_t3,
                                 cohort , partition,
                                 z)) # we always observe z
    # Transform data into long format
    dta <- get_long_data(list(as.vector(y_t1), as.vector(y_t2), as.vector(y_t3)), cohort, partition, z)
    setorder(dta, "id", "time")
    # generate clusters (there's no actual within-cluster correlation) with sample with replacement by id
    # get unique ids
    unique_ids <- unique(dta$id)
    # sample cluster values for each unique id
    clusters <- data.table(id = unique_ids, cluster = sample(1:50, size = length(unique_ids), replace = TRUE))
    # merge the sampled clusters back into the original data table
    dta <- merge(dta, clusters, by = "id", all.x = TRUE)
    #-----------------------------------------------------------------------------
    # Return the data
    return(list(data = as.data.table(dta),
                data_wide = as.data.table(dta_wide),
                ES_0_unf = ES_0_unf,
                prob_g2_p1 = prob_g2_p1,
                prob_g3_p1 = prob_g3_p1))

  } else if(dgp_type == 3){
    # Pscore depends on Z but Regressions depend on X
    # index for pscores

    # Generalize Propensity score (3 groups and two partitions)
    pi_2A <- exp(fps2(Xsi.ps, w1, z, 1.25))
    pi_2B <- exp(fps2(-Xsi.ps, w1, z, -0.5))
    pi_3A <- exp(fps2(Xsi.ps, w2, z, 2))
    pi_3B <- exp(fps2(-Xsi.ps, w2, z, -1.25))
    pi_0A <- exp(fps2(Xsi.ps, w3, z, -0.5))
    # Get the sum of the exponentials
    sum_pi <- 1 + pi_2A + pi_2B + pi_3A + pi_3B + pi_0A
    # Construct the generalized pscore
    pi_2A <- pi_2A/sum_pi
    pi_2B <- pi_2B/sum_pi
    pi_3A <- pi_3A/sum_pi
    pi_3B <- pi_3B/sum_pi
    pi_0A <- pi_0A/sum_pi
    pi_0B <- 1 - (pi_2A + pi_2B + pi_3A + pi_3B + pi_0A)

    # Bind them into a matrix
    probs_pscore <- cbind(pi_2A, pi_2B, pi_3A, pi_3B, pi_0A, pi_0B)

    # Sample group types (1 and 3 are the group of interest)
    group_types <- apply(probs_pscore, 1,
                         function(pvec) sample(seq(1,6), size=1, prob=pvec))

    # Translate group types into partitions and cohorts
    partition <- ifelse(group_types %in% c(1,3,5), 1, 0)
    cohort <- ifelse(group_types %in% c(1,2),
                     2,
                     ifelse(group_types %in% c(3,4),
                            3,
                            0))
    #-----------------------------------------------------------------------------
    # Now create the indexes for the potential outcomes
    #-----------------------------------------------------------------------------
    # Index for PO
    index_lin <- freg(b1, x)
    index_partition <- partition * index_lin
    # Index for unobserved heterogeneity
    index_unobs_het <- cohort * index_lin + index_partition
    #Index for Conditional Parallel Trends (this does not vary with treatment group)
    index_trend <- index_lin
    # Generate unobserved heterogeneity
    # (fixed effects that are correlated with X, partition, and cohort)
    v <- stats::rnorm(size, mean = index_unobs_het, sd = 1)
    # Index to violates the parallel trends assumption
    index_pt_violation <- v/10

    # Index for per-period ATT(g,t)
    index_att_g2 <- 10
    index_att_g3 <- 25
    #-----------------------------------------------------------------------------
    # Generate the potential outcomes
    # Baseline index
    baseline_t1 <- index_lin + index_partition + v

    #Generate untreated outcome at time 1 (that is observed for all units)
    y_t1 <- baseline_t1 + stats::rnorm(size)

    # Generate potential outcomes at time 2
    baseline_t2 <- baseline_t1 +
      index_pt_violation + #Violate the CPT
      index_trend #this is for the trend based on X

    y_t2_never <- baseline_t2 + stats::rnorm(size)
    y_t2_g2 <- baseline_t2 + stats::rnorm(size) +
      index_att_g2 * partition # This is the treatment effects for cohort 2
    y_t2_g3 <-  baseline_t2 + stats::rnorm(size)

    # Generate potential outcomes at time 3
    baseline_t3 <- baseline_t1 +
      2*index_trend + #this is for the trend based on X
      2*index_pt_violation #Violate the CPT

    y_t3_never <- baseline_t3 + stats::rnorm(size)
    y_t3_g2 <- baseline_t3 + stats::rnorm(size) +
      2*index_att_g2 * partition # This is the treatment effects for cohort 2
    y_t3_g3 <- baseline_t3 + stats::rnorm(size) +
      index_att_g3 * partition # This is the treatment effects for cohort 2
    #-----------------------------------------------------------------------------
    # Get the realized outcomes now
    # y_t2<- y_t2_g2*(cohort == 2) + y_t2_g3*(cohort==3) + y_t2_never*(cohort==0)
    # y_t3<- y_t3_g2*(cohort == 2) + y_t3_g3*(cohort==3) + y_t3_never*(cohort==0)

    y_t2<-  ifelse(cohort == 2 & partition==1, y_t2_g2, y_t2_never)
    y_t3<- ifelse(cohort == 2 & partition==1, y_t3_g2,
                  ifelse(cohort == 3 & partition==1, y_t3_g3,
                         y_t3_never))
    #-----------------------------------------------------------------------------
    # Get the event study parameter
    # Get unfeasible ATT(g,t) (assuming we see the Potential outcomes)
    att_g2_t2_unf <- (mean((group_types==1)*y_t2_g2) - mean((group_types==1)*y_t2_never))/
      mean((group_types==1))
    att_g2_t3_unf <- (mean((group_types==1)*y_t3_g2) - mean((group_types==1)*y_t3_never))/
      mean((group_types==1))
    att_g3_t2_unf <- (mean((group_types==3)*y_t2_g3) - mean((group_types==3)*y_t2_never))/
      mean((group_types==3))
    att_g3_t3_unf <- (mean((group_types==3)*y_t3_g3) - mean((group_types==3)*y_t3_never))/
      mean((group_types==3))

    # Get unfeasible Event-Study (ES) (assuming we see the Potential outcomes and pscores)
    # get weights
    prob_g2_p1 <- mean(pi_2A/(pi_2A + pi_3A))
    prob_g3_p1 <- mean(pi_3A/(pi_2A + pi_3A))
    # Get ES
    ES_0_unf <- att_g2_t2_unf * prob_g2_p1 + att_g3_t3_unf * prob_g3_p1
    ES_1_unf <- att_g2_t3_unf
    # ------------------------------------------------------------------------------
    # Put all data in a data frame
    dta_wide <- data.frame(cbind(id = 1:size,
                                 y_t1, y_t2 , y_t3,
                                 cohort , partition,
                                 z)) # we always observe z
    # Transform data into long format
    dta <- get_long_data(list(as.vector(y_t1), as.vector(y_t2), as.vector(y_t3)), cohort, partition, z)
    setorder(dta, "id", "time")
    # generate clusters (there's no actual within-cluster correlation) with sample with replacement by id
    # get unique ids
    unique_ids <- unique(dta$id)
    # sample cluster values for each unique id
    clusters <- data.table(id = unique_ids, cluster = sample(1:50, size = length(unique_ids), replace = TRUE))
    # merge the sampled clusters back into the original data table
    dta <- merge(dta, clusters, by = "id", all.x = TRUE)
    #-----------------------------------------------------------------------------
    # Return the data
    return(list(data = as.data.table(dta),
                data_wide = as.data.table(dta_wide),
                ES_0_unf = ES_0_unf,
                prob_g2_p1 = prob_g2_p1,
                prob_g3_p1 = prob_g3_p1))

  } else if (dgp_type == 4){
    # Both models are functions of X (misspecified)
    # index for pscores

    # Generalize Propensity score (3 groups and two partitions)
    pi_2A <- exp(fps2(Xsi.ps, w1, x, 1.25))
    pi_2B <- exp(fps2(-Xsi.ps, w1, x, -0.5))
    pi_3A <- exp(fps2(Xsi.ps, w2, x, 2))
    pi_3B <- exp(fps2(-Xsi.ps, w2, x, -1.25))
    pi_0A <- exp(fps2(Xsi.ps, w3, x, -0.5))
    # Get the sum of the exponentials
    sum_pi <- 1 + pi_2A + pi_2B + pi_3A + pi_3B + pi_0A
    # Construct the generalized pscore
    pi_2A <- pi_2A/sum_pi
    pi_2B <- pi_2B/sum_pi
    pi_3A <- pi_3A/sum_pi
    pi_3B <- pi_3B/sum_pi
    pi_0A <- pi_0A/sum_pi
    pi_0B <- 1 - (pi_2A + pi_2B + pi_3A + pi_3B + pi_0A)

    # Bind them into a matrix
    probs_pscore <- cbind(pi_2A, pi_2B, pi_3A, pi_3B, pi_0A, pi_0B)

    # Sample group types (1 and 3 are the group of interest)
    group_types <- apply(probs_pscore, 1,
                         function(pvec) sample(seq(1,6), size=1, prob=pvec))

    # Translate group types into partitions and cohorts
    partition <- ifelse(group_types %in% c(1,3,5), 1, 0)
    cohort <- ifelse(group_types %in% c(1,2),
                     2,
                     ifelse(group_types %in% c(3,4),
                            3,
                            0))
    #-----------------------------------------------------------------------------
    # Now create the indexes for the potential outcomes
    #-----------------------------------------------------------------------------
    # Index for PO
    index_lin <- freg(b1, x)
    index_partition <- partition * index_lin
    # Index for unobserved heterogeneity
    index_unobs_het <- cohort * index_lin + index_partition
    #Index for Conditional Parallel Trends (this does not vary with treatment group)
    index_trend <- index_lin
    # Generate unobserved heterogeneity
    # (fixed effects that are correlated with X, partition, and cohort)
    v <- stats::rnorm(size, mean = index_unobs_het, sd = 1)
    # Index to violates the parallel trends assumption
    index_pt_violation <- v/10

    # Index for per-period ATT(g,t)
    index_att_g2 <- 10
    index_att_g3 <- 25
    #-----------------------------------------------------------------------------
    # Generate the potential outcomes
    # Baseline index
    baseline_t1 <- index_lin + index_partition + v

    #Generate untreated outcome at time 1 (that is observed for all units)
    y_t1 <- baseline_t1 + stats::rnorm(size)

    # Generate potential outcomes at time 2
    baseline_t2 <- baseline_t1 +
      index_pt_violation + #Violate the CPT
      index_trend #this is for the trend based on X

    y_t2_never <- baseline_t2 + stats::rnorm(size)
    y_t2_g2 <- baseline_t2 + stats::rnorm(size) +
      index_att_g2 * partition # This is the treatment effects for cohort 2
    y_t2_g3 <-  baseline_t2 + stats::rnorm(size)

    # Generate potential outcomes at time 3
    baseline_t3 <- baseline_t1 +
      2*index_trend + #this is for the trend based on X
      2*index_pt_violation #Violate the CPT

    y_t3_never <- baseline_t3 + stats::rnorm(size)
    y_t3_g2 <- baseline_t3 + stats::rnorm(size) +
      2*index_att_g2 * partition # This is the treatment effects for cohort 2
    y_t3_g3 <- baseline_t3 + stats::rnorm(size) +
      index_att_g3 * partition # This is the treatment effects for cohort 2
    #-----------------------------------------------------------------------------
    # Get the realized outcomes now
    # y_t2<- y_t2_g2*(cohort == 2) + y_t2_g3*(cohort==3) + y_t2_never*(cohort==0)
    # y_t3<- y_t3_g2*(cohort == 2) + y_t3_g3*(cohort==3) + y_t3_never*(cohort==0)

    y_t2<-  ifelse(cohort == 2 & partition==1, y_t2_g2, y_t2_never)
    y_t3<- ifelse(cohort == 2 & partition==1, y_t3_g2,
                  ifelse(cohort == 3 & partition==1, y_t3_g3,
                         y_t3_never))
    #-----------------------------------------------------------------------------
    # Get the event study parameter
    # Get unfeasible ATT(g,t) (assuming we see the Potential outcomes)
    att_g2_t2_unf <- (mean((group_types==1)*y_t2_g2) - mean((group_types==1)*y_t2_never))/
      mean((group_types==1))
    att_g2_t3_unf <- (mean((group_types==1)*y_t3_g2) - mean((group_types==1)*y_t3_never))/
      mean((group_types==1))
    att_g3_t2_unf <- (mean((group_types==3)*y_t2_g3) - mean((group_types==3)*y_t2_never))/
      mean((group_types==3))
    att_g3_t3_unf <- (mean((group_types==3)*y_t3_g3) - mean((group_types==3)*y_t3_never))/
      mean((group_types==3))

    # Get unfeasible Event-Study (ES) (assuming we see the Potential outcomes and pscores)
    # get weights
    prob_g2_p1 <- mean(pi_2A/(pi_2A + pi_3A))
    prob_g3_p1 <- mean(pi_3A/(pi_2A + pi_3A))
    # Get ES
    ES_0_unf <- att_g2_t2_unf * prob_g2_p1 + att_g3_t3_unf * prob_g3_p1
    ES_1_unf <- att_g2_t3_unf
    # ------------------------------------------------------------------------------
    # Put all data in a data frame
    dta_wide <- data.frame(cbind(id = 1:size,
                                 y_t1, y_t2 , y_t3,
                                 cohort , partition,
                                 z)) # we always observe z
    # Transform data into long format
    dta <- get_long_data(list(as.vector(y_t1), as.vector(y_t2), as.vector(y_t3)), cohort, partition, z)
    setorder(dta, "id", "time")
    # generate clusters (there's no actual within-cluster correlation) with sample with replacement by id
    # get unique ids
    unique_ids <- unique(dta$id)
    # sample cluster values for each unique id
    clusters <- data.table(id = unique_ids, cluster = sample(1:50, size = length(unique_ids), replace = TRUE))
    # merge the sampled clusters back into the original data table
    dta <- merge(dta, clusters, by = "id", all.x = TRUE)
    #-----------------------------------------------------------------------------
    # Return the data
    return(list(data = as.data.table(dta),
                data_wide = as.data.table(dta_wide),
                ES_0_unf = ES_0_unf,
                prob_g2_p1 = prob_g2_p1,
                prob_g3_p1 = prob_g3_p1))

  } else {
    stop("Invalid DGP type")
  }

}

# Testing if main function in generating output
test_that("Testing generation of output in main function", {

  # generating dataset without errors
  test_panel = generate_test_panel()
  # data from dgp 2 periods
  data_2periods <- gen_dgp_2periods(size = 1000, dgp_type = 1)[["data"]]
  # data from dgp multiple periods
  data_multperiods <- gen_dgp_mult_periods(size = 1000, dgp_type = 1)[["data"]]

  # ------------------------------
  # Performing tests
  # ------------------------------

  ddd_test <- ddd(yname = "outcome", tname = "year", idname = "id", gname = "treat",
                  pname = "partition", xformla = ~x1 + x2,
                  data = test_panel, control_group = NULL, base_period = NULL, est_method = "dr",
                  weightsname = NULL, boot = FALSE, nboot = NULL,
                  inffunc = FALSE, skip_data_checks = FALSE)

  # performing DR estimation with 2 periods
  ddd_dr = ddd(yname = "y", tname = "time", idname = "id", gname = "state",
               pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
               data = data_2periods, alpha = 0.05, est_method = "dr")

  ddd_ipw = ddd(yname = "y", tname = "time", idname = "id", gname = "state",
               pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
               data = data_2periods, alpha = 0.05, est_method = "ipw")

  ddd_reg = ddd(yname = "y", tname = "time", idname = "id", gname = "state",
               pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
               data = data_2periods, alpha = 0.05, est_method = "reg")

  # performing DR estimation with 2 periods and bootstrapping + clustered std. errors.
  ddd_boot_cluster = ddd(yname = "y", tname = "time", idname = "id", gname = "state",
                         pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
                         data = data_2periods, alpha = 0.05, boot = TRUE, nboot = 999,
                         est_method = "dr", cluster = "cluster", cband = TRUE)


  # performing DR estimation with multiple periods
  ddd_gt <- ddd(yname = "y", tname = "time", idname = "id",
                gname = "state", pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
                data = data_multperiods, control_group = "nevertreated", base_period = "universal",
                est_method = "dr")

  ddd_gt_ipw <- ddd(yname = "y", tname = "time", idname = "id",
                gname = "state", pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
                data = data_multperiods, control_group = "nevertreated", base_period = "universal",
                est_method = "ipw")

  ddd_gt_reg <- ddd(yname = "y", tname = "time", idname = "id",
                gname = "state", pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
                data = data_multperiods, control_group = "nevertreated", base_period = "universal",
                est_method = "reg")

  # performing DR estimation with multiple periods and bootstrapping
  ddd_gt_boot_cluster <- ddd(yname = "y", tname = "time", idname = "id",
                             gname = "state", pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
                             data = data_multperiods, control_group = "nevertreated", base_period = "universal",
                             est_method = "dr", boot = TRUE, nboot = 999, cluster = "cluster", cband = TRUE)

  # expecting results
  expect_output(summary(ddd_test))
  expect_output(print(ddd_test))
  expect_output(summary(ddd_dr))
  expect_output(summary(ddd_ipw))
  expect_output(summary(ddd_reg))
  expect_output(summary(ddd_boot_cluster))
  expect_output(summary(ddd_gt))
  expect_output(summary(ddd_gt_ipw))
  expect_output(summary(ddd_gt_reg))
  expect_output(summary(ddd_gt_boot_cluster))


})

# =============================================================================
# Tests that the per-unit cluster vector built in att_gt() is CORRECTLY ALIGNED
# to the rows of inf_func_mat -- the porting guide calls this "the crux".
#
# Why this file exists (adversarial gap found in the original suite):
#   - CHECK 2 / CHECK 5b compute their "manual" CRVE from o$cluster_vector, the
#     object's OWN field, so they are self-referential: a deterministically
#     WRONG cluster vector passes them (implementation and check share the error).
#   - CHECK 7 only tests shuffle-INVARIANCE (stability), which a deterministic
#     misalignment satisfies, so it cannot detect a wrong-but-stable mapping.
#   - The only original test that catches a scrambled id->cluster map is the
#     analytic-vs-bootstrap CHECK 1, which is skip_on_cran() with a 10% tolerance.
#
# These tests pin the reported SE against an INDEPENDENT ground-truth cluster
# vector, reconstructed from the user's known id->cl map joined onto the unit
# ordering carried by first_period_dta (built in preprocessing, NOT by the
# att_gt() cluster_vec line under test). They fail if att_gt() misaligns,
# permutes, or otherwise corrupts the cluster vector.
# =============================================================================

.fit_clustered <- function(d, cg = "nevertreated") {
  suppressWarnings(ddd(
    yname = "y", tname = "time", idname = "id", gname = "state",
    pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
    data = d, control_group = cg, base_period = "varying",
    est_method = "dr", cluster = "cl", boot = FALSE))
}

# independent ground-truth cluster vector aligned to inf_func_mat rows:
# join the KNOWN id->cl map onto first_period_dta$id (unit row order).
.truth_cluster_vec <- function(o, clmap) {
  clmap$cl[match(o$first_period_dta$id, clmap$id)]
}
.crve <- function(IF, cv, n) sqrt(diag(crossprod(rowsum(IF, cv))) / n) / sqrt(n)

.make_data <- function(seed = 123, size = 600, n_clusters = 25, unbalanced = FALSE) {
  set.seed(seed)
  d <- gen_dgp_mult_periods(size = size, dgp_type = 1)[["data"]]
  ids <- sort(unique(d$id))
  if (unbalanced) {
    set.seed(seed + 1)
    cl <- sample(seq_len(n_clusters), length(ids), replace = TRUE,
                 prob = c(rep(8, 3), rep(1, n_clusters - 3)))
  } else {
    cl <- rep(seq_len(n_clusters), length.out = length(ids))
  }
  clmap <- data.frame(id = ids, cl = cl)
  list(data = merge(d, clmap, by = "id"), clmap = clmap)
}

# -----------------------------------------------------------------------------
# (1) cluster_vector and SE match an INDEPENDENT ground-truth partition.
#     Kills: misaligned / permuted / scrambled cluster vector in att_gt().
# -----------------------------------------------------------------------------
test_that("att_gt() cluster vector is aligned to the user's id->cl map (balanced & unbalanced)", {
  for (unb in c(FALSE, TRUE)) {
    dd  <- .make_data(seed = 123, size = 600, n_clusters = 25, unbalanced = unb)
    o   <- .fit_clustered(dd$data)
    IF  <- as.matrix(o$inf_func_mat); n <- o$n
    truth <- .truth_cluster_vec(o, dd$clmap)

    # the object's cluster vector must equal the independent ground truth
    expect_identical(as.numeric(o$cluster_vector), as.numeric(truth))

    # and the reported SE must be the CRVE of that independent partition
    manual <- .crve(IF, truth, n)
    ok <- is.finite(o$se) & is.finite(manual)
    expect_equal(as.numeric(o$se[ok]), as.numeric(manual[ok]), tolerance = 1e-8)
  }
})

# -----------------------------------------------------------------------------
# (2) row-permutation: reported SE matches the SAME independent ground truth
#     (not merely "is invariant"). Kills order-dependent misalignment that a
#     pure invariance check would miss.
# -----------------------------------------------------------------------------
test_that("clustered SE matches the ground-truth CRVE even after input rows are shuffled", {
  dd <- .make_data(seed = 55, size = 600, n_clusters = 25, unbalanced = TRUE)
  set.seed(777)
  d_shuf <- dd$data[sample(seq_len(nrow(dd$data))), ]

  o   <- .fit_clustered(d_shuf)
  IF  <- as.matrix(o$inf_func_mat); n <- o$n
  truth <- .truth_cluster_vec(o, dd$clmap)         # independent of row order

  expect_identical(as.numeric(o$cluster_vector), as.numeric(truth))
  manual <- .crve(IF, truth, n)
  ok <- is.finite(o$se) & is.finite(manual)
  expect_equal(as.numeric(o$se[ok]), as.numeric(manual[ok]), tolerance = 1e-8)
})

# -----------------------------------------------------------------------------
# (3) sensitivity: relabeling clusters (a bijection on labels) must NOT change
#     the SE; reassigning units to a DIFFERENT partition MUST change it. Guards
#     against the SE being accidentally independent of the clustering.
# -----------------------------------------------------------------------------
test_that("clustered SE is invariant to cluster relabeling but sensitive to repartitioning", {
  dd <- .make_data(seed = 7, size = 600, n_clusters = 20, unbalanced = TRUE)
  base <- .fit_clustered(dd$data)

  # (a) relabel: cl -> (n_clusters + 1 - cl). Same partition, different labels.
  d_relabel <- dd$data
  d_relabel$cl <- (max(dd$clmap$cl) + 1L) - d_relabel$cl
  o_relabel <- .fit_clustered(d_relabel)
  ok <- is.finite(base$se) & is.finite(o_relabel$se)
  expect_equal(as.numeric(base$se[ok]), as.numeric(o_relabel$se[ok]), tolerance = 1e-8)

  # (b) repartition: give every unit its own cluster -> clustered SE collapses
  #     to the i.i.d. analytic SE and DIFFERS from the grouped-cluster SE.
  d_singleton <- dd$data
  d_singleton$cl <- d_singleton$id
  o_singleton <- .fit_clustered(d_singleton)
  o_iid <- ddd(
    yname = "y", tname = "time", idname = "id", gname = "state",
    pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
    data = dd$data, control_group = "nevertreated", base_period = "varying",
    est_method = "dr", cluster = NULL, boot = FALSE)
  ok2 <- is.finite(o_singleton$se) & is.finite(o_iid$se)
  expect_equal(as.numeric(o_singleton$se[ok2]), as.numeric(o_iid$se[ok2]), tolerance = 1e-8)

  ok3 <- is.finite(base$se) & is.finite(o_singleton$se)
  expect_gt(max(abs(base$se[ok3] - o_singleton$se[ok3]) / o_singleton$se[ok3]), 0.02)
})

# Testing that results with analytical standard errors and bootstrap are comparable
test_that("multiplication works", {
  # generating dataset without errors
  test_panel = generate_test_panel()

  # ------------------------------
  # Performing tests
  # ------------------------------

  ddd_boostrap <- ddd(yname = "outcome", tname = "year", idname = "id", gname = "treat",
                 pname = "partition", xformla = ~x1 + x2,
                  data = test_panel, control_group = NULL, base_period = NULL, est_method = "dr",
                  weightsname = NULL, boot = TRUE, nboot = 1000, cband = TRUE,
                  inffunc = FALSE, skip_data_checks = FALSE)

  ddd_analytical <- ddd(yname = "outcome", tname = "year", idname = "id", gname = "treat",
                  pname = "partition", xformla = ~x1 + x2,
                  data = test_panel, control_group = NULL, base_period = NULL, est_method = "dr",
                  weightsname = NULL, boot = FALSE, nboot = NULL,
                  inffunc = FALSE, skip_data_checks = FALSE)

  # Check that point estimates are the same
  expect_equal(ddd_analytical$ATT, ddd_boostrap$ATT)

  # Check that standard errors are comparable
  expect_equal(ddd_analytical$se, ddd_boostrap$se, tolerance = 0.5)
})

# Testing clustered standard error is working correctly
test_that("clustered standard errors are working correctly", {
  # generating dataset without errors
  test_panel = gen_dgp_2periods(size = 5000, dgp_type = 1)$data

  # ------------------------------
  # Performing tests
  # ------------------------------

  att_nocluster <- ddd(yname = "y", tname = "time", idname = "id", gname = "state",
                pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4, base_period = "universal",
                data = test_panel, control_group = "nevertreated", est_method = "dr")

  att_cluster <-  ddd(yname = "y", tname = "time", idname = "id", gname = "state",
                      pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
                      data = test_panel, control_group = "nevertreated", boot = TRUE, nboot = 1000, cband = TRUE,
                      base_period = "universal", est_method = "dr", cluster = "cluster")

  # Check that point estimates are the same
  expect_equal(att_nocluster$ATT, att_cluster$ATT)

  # Check that standard errors are different
  expect_false(isTRUE(all.equal(att_nocluster$se, att_cluster$se)))
})

library(data.table)
# Function to generate a test panel dataset
# Dataset only for testing purposes and not for validation of any statistical property
generate_test_panel <- function(seed = 123,
                                num_ids = 50,
                                time = 2,
                                initial.year = 2019,
                                treatment.year = 2020) {
  # Set seed for reproducibility
  set.seed(seed)

  # Number of IDs
  num_ids <- num_ids

  # Number of observations per ID
  num_obs <- time

  # Generate IDs
  id = rep(1:num_ids, each = num_obs)

  # Generate x1 and x2 for each state
  x1 <- rep(rnorm(num_ids), each = num_obs)
  x2 <- rep(rnorm(num_ids), each = num_obs)

  # Generate state variable
  state = rep(sample(c("A", "B", "C", "D"), num_ids, replace = TRUE), each = num_obs)

  # generate time variable
  year = rep(seq(initial.year, initial.year+(time-1), 1), times = num_ids)

  # Generate partition variable
  partition = rep(rbinom(num_ids, 1, 0.5), each = num_obs)

  # Generate target group
  s_2 = as.numeric(state %in% c("A", "B"))

  # Generate post treatment variable
  t_2 = as.numeric(year >= treatment.year)

  # TWFE outcome variable
  y = 2 + 5*s_2 - 2 * partition + 3 * t_2 +
    4*s_2*partition + 2*s_2*t_2 + 3*partition*t_2 +
    1*s_2*partition*t_2 + rnorm(num_ids*num_obs)

  # Create a data.table
  dt <- data.table::data.table(
    id = id,
    state = state,
    year = year,
    partition = partition,
    x1 = x1,
    x2 = x2,
    treat = s_2,
    outcome = y
  )
  return(dt)
}

#' @title Summary
#' @description Summary of an aggddd object
#' @param object A aggddd object
#' @param ... Other params (required as generic function, but not used)
#' @export
#' @noRd
# Define new summary function
summary.agg_ddd <- function(object, ...){
  agg_ddd.obj <- object
  print(agg_ddd.obj)
}

# Main function for triplediff
NULL
#' Doubly Robust DDD estimators for the group-time average treatment effects.
#'
#' @description
#' \code{ddd} is the main function for computing the Doubly Robust DDD estimators for the ATT, with balanced panel data.
#' It can be used with covariates and/or under multiple time periods. At its core, \code{triplediff} employs
#' the doubly robust estimator for the ATT, which is a combination of the propensity score weighting and the outcome regression.
#' Furthermore, this package supports the application of machine learning methods for the estimation of the nuisance parameters.
#' @param yname The name of the outcome variable.
#' @param tname The name of the column containing the time periods.
#' @param idname The name of the column containing the unit id.
#' @param gname The name of the column containing the first period when a particular observation is treated. It is a positive number
#' for treated units and defines which group the unit belongs to. It takes value 0 or Inf for untreated units.
#' @param pname The name of the column containing the partition variable (e.g., the subgroup identifier). This is an indicator variable that is 1 for
#' the units eligible for treatment and 0 otherwise.
#' @param xformla The formula for the covariates to be included in the model. It should be of the form \code{~ x1 + x2}.
#' Default is \code{xformla = ~1} (no covariates).
#' @param data A data frame or data table containing the data.
#' @param control_group Valid for multiple periods only. The control group to be used in the estimation. Default is \code{control_group = "notyettreated"} which sets as control group the units that have not yet participated in the treatment.
#' The alternative is \code{control_group = "nevertreated"} which sets as control group the units that never participate in the treatment and does not change across groups or time periods.
#' @param base_period Valid for multiple periods. Choose between a "varying" or "universal" base period. Both yield the same post-treatment ATT(g,t) estimates.
#' Varying base period: Computes pseudo-ATT in pre-treatment periods by comparing outcome changes for a group to its comparison group from t-1 to t, repeatedly changing t.
#' Universal base period: Fixes the base period to (g-1), reporting average changes from t to (g-1) for a group relative to its comparison group, similar to event study regressions.
#' Varying base period reports ATT(g,t) right before treatment. Universal base period normalizes the estimate before treatment to be 0, adding one extra estimate in an earlier period.
#' @param est_method The estimation method to be used. Default is \code{"dr"} (doubly robust). It computes propensity score using logistic regression
#' and outcome regression using OLS. The alternative are \code{c("reg", "ipw")}.
#' @param panel Logical. If \code{TRUE} (default), the data is treated as panel data where each unit is observed in all time periods.
#' If \code{FALSE}, the data is treated as repeated cross-sections (RCS) where each observation may represent a different unit.
#' For RCS data, \code{idname} can be omitted or set to \code{NULL}, and the function will automatically create unique IDs for each observation.
#' @param allow_unbalanced_panel Logical. If \code{TRUE}, allows for unbalanced panel data where units may not be observed in all time periods.
#' Default is \code{FALSE}. Note: This parameter requires \code{panel = TRUE} and a valid \code{idname}.
#' @param weightsname The name of the column containing the weights. Default is \code{NULL}. As part of data processing, weights are enforced to be normalized
#' and have mean 1 across all observations.
#' @param boot Logical. If \code{TRUE}, the function computes standard errors using the multiplier bootstrap. Default is \code{FALSE}.
#' @param nboot The number of bootstrap samples to be used. Default is \code{NULL}. If \code{boot = TRUE}, the default is \code{nboot = 999}.
#' @param cluster The name of the variable to be used for clustering. The maximum number of cluster variables is 1. Default is \code{NULL}.
#' If \code{boot = TRUE}, the function computes the bootstrap standard errors clustering at the unit level setting as cluster variable the one in \code{idname}.
#' @param cband Logical. If \code{TRUE}, the function computes a uniform confidence band that covers all of the average treatment effects
#' with fixed probability \code{1-alpha}.  In order to compute uniform confidence bands, \code{boot} must also be set to `TRUE`.  The default is `FALSE`.
#' @param alpha The level of significance for the confidence intervals.  Default is \code{0.05}.
#' @param use_parallel Logical. If \code{TRUE}, the function runs in parallel processing. Valid only when \code{boot = TRUE}. Default is \code{FALSE}.
#' @param cores The number of cores to be used in the parallel processing. Default is \code{cores = 1}.
#' @param inffunc Logical. If \code{TRUE}, the function returns the influence function. Default is \code{FALSE}.
#' @param skip_data_checks Logical. If \code{TRUE}, the function skips data validation checks and proceeds directly to estimation.
#' This can improve performance when you are confident the data is correctly formatted. Default is \code{FALSE}.
#' Use with caution as skipping checks may lead to unexpected errors if data is malformed.
#'
#' @return A `ddd` object with the following basic elements:
#' \item{ATT}{The average treatment effect on the treated.}
#' \item{se}{The standard error of the ATT.}
#' \item{uci}{The upper confidence interval of the ATT.}
#' \item{lci}{The lower confidence interval of the ATT.}
#' \item{inf_func}{The estimate of the influence function.}
#'
#'
#' @examples
#' #----------------------------------------------------------
#' # Triple Diff with covariates and 2 time periods
#' #----------------------------------------------------------
#' set.seed(1234) # Set seed for reproducibility
#' # Simulate data for a two-periods DDD setup
#' df <- gen_dgp_2periods(size = 5000, dgp_type = 1)$data
#'
#' head(df)
#'
#' att_22 <- ddd(yname = "y", tname = "time", idname = "id", gname = "state",
#'               pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
#'              data = df, control_group = "nevertreated", est_method = "dr")
#'
#' summary(att_22)
#'
#' #----------------------------------------------------------
#' # Triple Diff with multiple time periods
#' #----------------------------------------------------------
#' data <- gen_dgp_mult_periods(size = 1000, dgp_type = 1)[["data"]]
#'
#' ddd(yname = "y", tname = "time", idname = "id",
#'      gname = "state", pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
#'      data = data, control_group = "nevertreated", base_period = "varying",
#'      est_method = "dr")
#'
#' @export
ddd <- function(yname,
                tname,
                idname = NULL,
                gname,
                pname,
                xformla,
                data,
                control_group = NULL,
                base_period = NULL,
                est_method = "dr",
                panel = TRUE,
                allow_unbalanced_panel = FALSE,
                weightsname = NULL,
                boot = FALSE,
                nboot = NULL,
                cluster = NULL,
                cband = FALSE,
                alpha = 0.05,
                use_parallel = FALSE,
                cores = 1,
                inffunc = FALSE,
                skip_data_checks = FALSE) {


  #------------------------------------------
  # Running initial arguments validation
  #------------------------------------------

  # Check if 'dta' is a data.table
  if (!"data.table" %in% class(data)) {
    # converting data to data.table
    dta <- data.table::as.data.table(data)
  } else {
    dta <- data
  }

  # Check what's the setting: 2 time periods or multiple time periods
  if (is.null(gname)) {
    stop("gname should be provided")
  } else {
    unique_gname_values <- dta[, uniqueN(get(gname))]
    unique_tname_values <- dta[, uniqueN(get(tname))]
    if (max(unique_gname_values,unique_tname_values) == 2) {
      multiple_periods <- FALSE
    } else if (max(unique_gname_values,unique_tname_values)  > 2) {
      multiple_periods <- TRUE
    } else {
      stop("Invalid gname. Please check your arguments.")
    }
  }

  #------------------------------------------
  # Validate idname parameter
  #------------------------------------------

  # If idname is not provided, assume RCS data
  if (is.null(idname)) {
    # Check for incompatible parameter combinations
    if (panel) {
      stop("idname is required when panel = TRUE. For repeated cross-section data, set panel = FALSE.")
    }
    if (allow_unbalanced_panel) {
      stop("idname is required when allow_unbalanced_panel = TRUE. For repeated cross-section data, set panel = FALSE and leave idname as NULL.")
    }
    # Set a placeholder - preprocessing will create the actual ID
    idname <- ".row_id"
  } else {
    # User provided idname - check if column exists
    if (!idname %in% names(dta)) {
      stop(paste0("Column '", idname, "' not found in data."))
    }
  }

  # Flag for est_method
  if (!(est_method %in% c("reg", "ipw", "dr"))) {
    warning("est_method = ", est_method, " is invalid or not supported. Using 'dr'.")
    est_method <- "dr"
  }

  # Flag for control group in multiple periods
  if (multiple_periods && is.null(control_group)) {
      warning("control_group should be provided for multiple time periods. Using 'nevertreated'")
      control_group <- "nevertreated"
  }

  # Flag for base period in multiple periods
  if (multiple_periods && is.null(base_period)) {
      warning("base_period should be provided for multiple time periods. Using 'varying'")
      base_period <- "varying"
  }

  #------------------------------------------
  # Run preprocess and validation checks
  #------------------------------------------

  if (skip_data_checks){
    if (!multiple_periods){
      dp <- run_nopreprocess_2periods(yname = yname,
                                      tname = tname,
                                      idname = idname,
                                      gname = gname,
                                      pname = pname,
                                      xformla = xformla,
                                      dta = dta,
                                      control_group = NULL,
                                      est_method = est_method,
                                      panel = panel,
                                      allow_unbalanced_panel = allow_unbalanced_panel,
                                      learners = NULL,
                                      n_folds = NULL,
                                      weightsname = weightsname,
                                      boot = boot,
                                      nboot = nboot,
                                      cluster = cluster,
                                      cband = cband,
                                      alpha = alpha,
                                      use_parallel = use_parallel,
                                      cores = cores,
                                      inffunc = inffunc)
    } else {
      stop("Triple Diff with multiple time periods is not yet supported.")
    }

  } else {
    if ((multiple_periods) && (est_method %in% c("dr", "reg", "ipw"))) {
      dp <- run_preprocess_multPeriods(yname = yname,
                                       tname = tname,
                                       idname = idname,
                                       gname = gname,
                                       pname = pname,
                                       xformla = xformla,
                                       dta = dta,
                                       control_group = control_group,
                                       base_period = base_period,
                                       est_method = est_method,
                                       panel = panel,
                                       allow_unbalanced_panel = allow_unbalanced_panel,
                                       learners = NULL,
                                       n_folds = NULL,
                                       weightsname = weightsname,
                                       boot = boot,
                                       nboot = nboot,
                                       cluster = cluster,
                                       cband = cband,
                                       alpha = alpha,
                                       use_parallel = use_parallel,
                                       cores = cores,
                                       inffunc = inffunc)

    } #else if ((multiple_periods) && (est_method == "dml")) {
      # dp <- run_preprocess_multPeriods(yname = yname,
      #                                  tname = tname,
      #                                  idname = idname,
      #                                  gname = gname,
      #                                  pname = pname,
      #                                  xformla = xformla,
      #                                  dta = dta,
      #                                  control_group = control_group,
      #                                  base_period = base_period,
      #                                  est_method = "dml",
      #                                  learners = learners,
      #                                  n_folds = n_folds,
      #                                  weightsname = weightsname,
      #                                  boot = boot,
      #                                  nboot = nboot,
      #                                  inffunc = inffunc)
      #stop("Triple Diff with multiple time periods and DML is not yet supported")}
    else if ((!multiple_periods) && (est_method %in% c("dr", "reg", "ipw"))) {
      dp <- run_preprocess_2Periods(yname = yname,
                                    tname = tname,
                                    idname = idname,
                                    gname = gname,
                                    pname = pname,
                                    xformla = xformla,
                                    dta = dta,
                                    control_group = NULL,
                                    est_method = est_method,
                                    panel = panel,
                                    allow_unbalanced_panel = allow_unbalanced_panel,
                                    learners = NULL,
                                    n_folds = NULL,
                                    weightsname = weightsname,
                                    boot = boot,
                                    nboot = nboot,
                                    cluster = cluster,
                                    cband = cband,
                                    alpha = alpha,
                                    use_parallel = use_parallel,
                                    cores = cores,
                                    inffunc = inffunc)
    } # else if ((!multiple_periods) && (est_method == "dml")) {
      # dp <- run_preprocess_2Periods(yname = yname,
      #                               tname = tname,
      #                               idname = idname,
      #                               gname = gname,
      #                               pname = pname,
      #                               xformla = xformla,
      #                               dta = dta,
      #                               control_group = NULL,
      #                               est_method = "dml",
      #                               learners = learners,
      #                               n_folds = n_folds,
      #                               weightsname = weightsname,
      #                               boot = boot,
      #                               nboot = nboot,
      #                               cluster = cluster,
      #                               cband = cband,
      #                               alpha = alpha,
      #                               use_parallel = use_parallel,
      #                               cores = cores,
      #                               inffunc = inffunc)}
  }

  #------------------------------------------
  # Run the estimation
  #------------------------------------------

  # multiple time periods case: dr or dml
  if (multiple_periods){
    # RUN DR for multiple periods
    if (est_method %in% c("dr", "reg", "ipw")){
      att_gt_dr <- att_gt(dp)
    } #else {
      # # RUN DML for multiple time periods
      # stop("DML for multiple time periods is not yet supported")
      # # TODO: IMPLEMENT att_gt_dml PROCEDURE AND ADJUST PARAMETERS
      # # att_gt_dml <- att_gt_dml(dp)
    #}
  } else {
    if (est_method %in% c("dml")){
        # RUN DML for 2 time periods
        # att_dml <- att_dml(dp)
      stop("DML estimation method is not yet supported.")
    } else {
      if (panel){
        # RUN DR for 2 time periods (balanced panel only)
        att_dr <- att_dr(dp)
      } else {
        # RUN DR for RCS 2 time periods (true RCS or unbalanced panel)
        att_dr <- att_dr_rc(dp)
      }
    }
  }


  #------------------------------------------
  # Return the results
  #------------------------------------------

  # record the call
  call.params <- match.call()
  # Record all arguments used in the function
  # Capture all arguments except 'data'
  arg_names <- setdiff(names(formals()), "data")
  args <- mget(arg_names, sys.frame(sys.nframe()))
  argu <- list(
    yname = args$yname,
    pname = args$pname,
    control_group = args$control_group,
    est_method = est_method,
    multiple_periods = multiple_periods,
    # learners = args$learners,
    # n_folds = args$n_folds,
    panel = args$panel, # getting from args because it could change in the pre process
    allow_unbalanced_panel = args$allow_unbalanced_panel, # getting from args because it could change in the pre process
    cband = dp$cband, # getting from dp because it could change in the pre process
    cluster = args$cluster,
    boot = dp$boot, # getting from dp because it could change in the pre process
    alpha = dp$alpha, # getting from dp because it could change in the pre process
    nboot = dp$nboot, # getting from dp because it could change in the pre process
    cores = dp$cores # getting from dp because it could change in the pre process
  )

  if (!multiple_periods){
    if (est_method %in% c("dr", "reg", "ipw")){
        ret <- list(
          ATT = att_dr$ATT,
          se = att_dr$se,
          lci = att_dr$lci,
          uci = att_dr$uci,
          nboot = att_dr$nboot, # this is not included under argu because it could change in the estimation process
          bT = att_dr$bT,
          att_inf_func = att_dr$inf_func,
          subgroup_counts = att_dr$subgroup_counts,
          call.params = call.params,
          argu = argu
        )
  } #else {
   #      ret <- list(
   #        ATT = att_dml$ATT,
   #        se = att_dml$se,
   #        lci = att_dml$lci,
   #        uci = att_dml$uci,
   #        subgroup_counts = att_dml$subgroup_counts,
   #        call.params = call.params,
   #        argu = argu
   #      )
   # }
  }# RETURNING LIST FOR 2 PERIODS CASE with DML

  # multiple time periods case: dr or dml
  if (multiple_periods){
    if (est_method %in% c("dr", "reg", "ipw")){
      ret <- list(
        ATT = att_gt_dr$ATT, # this is a vector of attgt
        se = att_gt_dr$se, # this is a vector of std. error for each attgt
        lci = att_gt_dr$lci, # this a vector of lower bound of CI for each attgt
        uci = att_gt_dr$uci, # this is a vector of upper bound of CI for each attgt
        groups = att_gt_dr$groups,
        periods = att_gt_dr$periods,
        tlist = att_gt_dr$tlist,
        glist = att_gt_dr$glist,
        cohort_size = att_gt_dr$cohort_size,
        n = att_gt_dr$n,
        bT = att_gt_dr$bT,
        inf_func_mat = att_gt_dr$inf_func_mat,
        first_period_dta = att_gt_dr$first_period_dta,
        cluster_vector = att_gt_dr$cluster_vector,
        cluster_var    = att_gt_dr$cluster_var,
        call.params = call.params,
        argu = argu
      )
    }
    # } else {
    #   ret <- list(
    #     ATT = att_gt_dml$ATT,
    #     se = att_gt_dml$se,
    #     lci = att_gt_dml$lci,
    #     uci = att_gt_dml$uci,
    #     att.inf.func = att_gt_dml$inf.func,
    #     call.params = call.params,
    #     argu = argu
    #   )
    # }
  }#RETURNING LIST FOR MULTIPLE PERIODS CASE

  # define a class
  class(ret) <- "ddd"

  return(ret)
  }

#' Doubly robust DDD estimator for ATT, with panel data and multiple periods
#'
#' This function implements a doubly robust estimator for assessing the average
#' treatment effect on the treated (ATT) using a triple differences (DDD) approach
#' in panel data settings with staggered treatment adoption. The function takes preprocessed
#' data structured specifically for this analysis.
#'
#' @importFrom Matrix Matrix
#' @param did_preprocessed A list containing preprocessed data and specifications for the DDD estimation.
#'        Expected elements include:
#'        - `preprocessed_data`: A data table containing the data with variables needed for the analysis.
#'        - `xformula`: The formula for the covariates to be included in the model. It should be of the form \code{~ x1 + x2}.
#'        Default is \code{xformla = ~1} (no covariates).
#'        - `control_group`: A character string indicating the control group. Default is \code{"notyettreated"}.
#'        - `n`: The number of unique ID observations in the data.
#'        - `cohorts`: A vector containing the number of treated cohorts
#'        - `time_periods`: A vector containing the number of time periods
#'        - `tlist`: A vector containing the time periods
#'        - `glist`: A vector containing the groups
#'        - `base_period`: A character string indicating the base period. Default is \code{"universal"}.
#'        - `cohort_size`: A vector containing the size of each cohort
#'        - `boot`: Logical. If \code{TRUE}, the function computes the bootstrap standard errors. Default is \code{FALSE}.
#'        - `nboot`: The number of bootstrap samples to be used. Default is \code{NULL}. If \code{boot = TRUE}, the default is \code{nboot = 999}.
#'        - `use_parallel`: Boolean of whether or not to use parallel processing in the multiplier bootstrap, default is \code{use_parallel=FALSE}
#'        - `cores`: the number of cores to use with parallel processing, default is \code{cores=1}
#'        - `cband`: Boolean of whether or not to compute simultaneous confidence bands, default is \code{cband=FALSE}
#'        - `alpha` The level of significance for the confidence intervals.  Default is \code{0.05}.
#'
#' @keywords internal
#' @return A list with the estimated ATT, standard error, upper and lower confidence intervals, and influence function.
#' @noRd
NULL
# ------------------------------------------------------------------------------

att_gt <- function(did_preprocessed){
  data <- did_preprocessed$preprocessed_data
  xformla <- did_preprocessed$xformula
  control_group <- did_preprocessed$control_group
  n <- did_preprocessed$n
  cohorts <- did_preprocessed$cohorts
  time_periods <- did_preprocessed$time_periods
  tlist <- did_preprocessed$tlist
  glist <- did_preprocessed$glist
  panel <- did_preprocessed$panel
  allow_unbalanced_panel <- did_preprocessed$allow_unbalanced_panel
  base_period <- did_preprocessed$base_period
  boot <- did_preprocessed$boot
  alpha <- did_preprocessed$alpha
  cohort_size <- did_preprocessed$cohort_size
  use_parallel <- did_preprocessed$use_parallel # to perform bootstrap
  cores <- did_preprocessed$cores # to perform bootstrap
  cband <- did_preprocessed$cband # to perform bootstrap + simult. conf. band

  orig_data <- copy(data)
  id_index <- unique(data$id)
  attgt_list <- list()

  counter <- 1
  tlist_length <- length(tlist)
  tfac <- 0
  if (base_period != "universal") {
    tlist_length <- tlist_length - 1
    tfac <- 1
  }

  # influence function matrix to be populated
  inf_func_mat <- Matrix::Matrix(data=0,nrow=n, ncol=cohorts*(time_periods - tfac), sparse=TRUE)
  se_gt_ddd_nyt <- rep(NA, cohorts * (time_periods - tfac)) # standard errors for not yet treated control group
  # If "nevertreated", create the control column that gonna indicate the comparison group for each iteration
  if (control_group == "nevertreated") {
    data[, control := as.integer(first_treat == 0)]
  }


  for (g in 1:cohorts){
    data[, treat := as.integer(first_treat == glist[g])]

    for (t in 1:tlist_length){
      pret <- t
      if (base_period == "universal"){
        pret <- tail(which(tlist < glist[g]), 1)
      }

      # creating control in case of "not yet treated option"
      if (control_group == "notyettreated") {
        # max_period <- tlist[max(t, pret) + tfac] + anticipation
        max_period <- tlist[max(t, pret) + tfac]
        data[, control := as.integer((first_treat == 0) | (first_treat > max_period & first_treat != glist[g]))]
      }


      # check if in post-treatment period
      if ((glist[g] <= tlist[(t + tfac)])) {

        # update pre-period if in post-treatment period to be period (g-delta-1)
        pret <- tail(which(tlist < glist[g]), 1)

        # print a warning message if there are no pre-treatment period
        if (length(pret) == 0) {
          warning(paste0("There are no pre-treatment periods for the group first treated at ", glist[g], "\nUnits from this group are dropped"))

          # if there are not pre-treatment periods, move on to next iteration
          break
        }
      }

      # normalize in period (g-1) to be equal to 0 when based period is universal and move on to next iteration
      if (base_period == "universal") {
        if (tlist[pret] == tlist[t + tfac]) {
          attgt_list[[counter]] <- list(att = 0, group = glist[g], year = tlist[t + tfac], post = 0)
          inf_func_mat[, counter] <- rep(0, n)
          counter <- counter + 1
          next
        }
      }

      # post treatment dummy variable
      post_treat <- 1*(glist[g] <= tlist[t + tfac])

      # filter data with only 2 periods
      cohort_data <- data[period %in% c(tlist[t + tfac], tlist[pret])]

      # ============================================
      # BRANCH: PANEL vs RCS/UNBALANCED
      # ============================================

      if (panel && !allow_unbalanced_panel) {
          # ==========================================
          # PANEL DATA PATH (balanced panels only)
          # ==========================================
        
          # get total number of units
          # n_size = uniqueN(cohort_data[, id])
        
          # filter data for treated and control groups in each (g,t) cell. Save index

          # index of unit in current cell (g,t) when treat = 1 and control = 1
          index_units_in_gt <- cohort_data[, treat == 1 | control == 1]

          # filter by units selected in current cell (g,t)
          cohort_data <- cohort_data[index_units_in_gt]

          # save number of units after filtering
          size_gt = uniqueN(cohort_data[, id])

          # Identify all available control states in this cohort.
          available_controls <- sort(unique(cohort_data$first_treat[cohort_data$control == 1 ]))
          available_controls <- available_controls[available_controls != glist[g]]

        if (length(available_controls) == 1) {
          # creating subgroup variable
          # 4 if (eligible ==1 & enabled == 1); 3 if (eligible ==0 & enabled == 1); 2 if (eligible ==1 & enabled == 0); 1 if (eligible ==0 & enabled == 0)
          # Create subgroup variable based on the control_group option
          cohort_data[, subgroup := NA_integer_]
          cohort_data[(treat == 1 & partition == 1), subgroup := 4]
          cohort_data[(treat == 1 & partition == 0), subgroup := 3]
          cohort_data[(treat == 0 & partition == 1), subgroup := 2]
          cohort_data[(treat == 0 & partition == 0), subgroup := 1]

          # add post treatment dummy variable if period is equal to t + tfac
          cohort_data[, post := 0]

          # Ypre is always the baseline, even in pre-treatment periods were baseline could be in later periods with respect to t.
          pseudo_post = ifelse(post_treat == 1, tlist[max(t, pret) + tfac], tlist[t + tfac])
          cohort_data[(period == pseudo_post), post := 1]

          # cohort_data[(period == tlist[max(t, pret) + tfac]), post := 1]
          # Calculate the size of each subgroup in the 'subgroup' column
          subgroup_counts <- cohort_data[, .(count = .N/2), by = subgroup][order(-subgroup)]

          # Reassign dp object and run att_dr
          did_preprocessed$preprocessed_data <- cohort_data
          did_preprocessed$subgroup_counts <- subgroup_counts
          did_preprocessed$boot <- FALSE # forcing false to avoid bootstrap inside att_dr() function.
          did_preprocessed$inffunc <- TRUE # forcing true to recover influence function

          # run att_dr based on 2 time periods subdata
          attgt_inf_func <- att_dr(did_preprocessed)

          # adjust influence function
          attgt_inf_func$inf_func <- (n/size_gt) * attgt_inf_func$inf_func
          # save results in a list
          attgt_list[[counter]] <- list(att = attgt_inf_func$ATT, group = glist[g], year = tlist[t + tfac], post = post_treat)

          # recover influence function
          inff <- rep(0, n)
          # avoid repetition in index
          inff[index_units_in_gt[seq(1, length(index_units_in_gt), by = 2)]] <- attgt_inf_func$inf_func
          # save in influence function matrix
          inf_func_mat[, counter] <- inff
          # update counter
          counter  <- counter + 1
        } else {
          # Here we have more than one control group available -> notyettreated

          # Initialize local containers.
          ddd_over_controls_res <- list()
          inf_mat_local <- NULL

          # Loop over each available control state.
          for (ctrl in available_controls) {
            # Subset data: keep treated units (first_treat == glist[g]) and control units (first_treat == ctrl)

            index_units_in_gt_ctrl <- cohort_data[, first_treat == glist[g] | first_treat == ctrl]
            subset_data <- cohort_data[index_units_in_gt_ctrl]
            size_gt_ctrl <- uniqueN(subset_data[, id])

            # Create subgroup variable based on the control_group option
            subset_data[, subgroup := NA_integer_]
            subset_data[(treat == 1 & partition == 1), subgroup := 4]
            subset_data[(treat == 1 & partition == 0), subgroup := 3]
            subset_data[(treat == 0 & partition == 1), subgroup := 2]
            subset_data[(treat == 0 & partition == 0), subgroup := 1]

            # add post treatment dummy variable if period is equal to t + tfac
            subset_data[, post := 0]

            # Ypre is always the baseline, even in pre-treatment periods were baseline could be in later periods with respect to t.
            pseudo_post = ifelse(post_treat == 1, tlist[max(t, pret) + tfac], tlist[t + tfac])
            subset_data[(period == pseudo_post), post := 1]

            # subset_data[(period == tlist[max(t, pret) + tfac]), post := 1]
            # Calculate the size of each subgroup in the 'subgroup' column
            subgroup_counts <- subset_data[, .(count = .N/2), by = subgroup][order(-subgroup)]

            # Reassign dp object and run att_dr
            did_preprocessed$preprocessed_data <- subset_data
            did_preprocessed$subgroup_counts <- subgroup_counts
            did_preprocessed$boot <- FALSE # forcing false to avoid bootstrap inside att_dr() function.
            did_preprocessed$inffunc <- TRUE # forcing true to recover influence function


            # run att_dr based on 2 time periods subdata
            ddd_out <- triplediff::att_dr(did_preprocessed)

            # computing the aggregate influence function for every g'>t: RIF_{g,g',t}
            # adjust influence function (rescale + recentered)
            ddd_out$inf_func <- ((size_gt/size_gt_ctrl)*ddd_out$inf_func) #+ 1*ddd_out$ATT

            # Save the ddd output in the local list using the control state as name.
            ddd_over_controls_res[[as.character(ctrl)]] <- ddd_out

            # recover influence function
            inff <- rep(0, n)
            # avoid repetition in index
            inff[index_units_in_gt_ctrl[seq(1, length(index_units_in_gt_ctrl), by = 2)]] <- ddd_out$inf_func

            # Append this vector as a new column in the local matrix.
            if (is.null(inf_mat_local)) {
              inf_mat_local <- matrix(inff, ncol = 1)
              colnames(inf_mat_local) <- as.character(ctrl)
            } else {
              inf_mat_local <- cbind(inf_mat_local, inff)
              colnames(inf_mat_local)[ncol(inf_mat_local)] <- as.character(ctrl)
            }
          } # end loop over available_controls

          # If no ddd estimations were obtained, skip this (g,t) cell.
          if (length(ddd_over_controls_res) == 0) next

          # Compute the naive ATT as the average of all ddd ATT estimates.
          att_vals <- sapply(ddd_over_controls_res, function(res) res$ATT)

          # -----------------------------------------
          # Compute the GMM-based estimator for ATT(g,t)
          # -----------------------------------------

          # computing the aggregate influence function for every g'>t: RIF_{g,g',t}
          OMEGA <- cov(inf_mat_local, use = "complete.obs")

          ones <- rep(1, length(ddd_over_controls_res))
          inv_OMEGA <- solve(OMEGA)
          w <- colSums(inv_OMEGA) / sum(inv_OMEGA)

          # Compute the GMM estimator for ATT(g,t)
          ATT_gmm <- sum(w * att_vals) / sum(w)
          # IF_gmm
          IF_gmm <- (inf_mat_local) %*% w
          # gmm_se
          gmm_se <- sqrt(1/ (n * sum(inv_OMEGA)))

          # save the results for this (g,t) cell
          attgt_list[[counter]] <- list(att = ATT_gmm, group = glist[g], year = tlist[t + tfac], post = post_treat)

          # save in influence function matrix
          inf_func_mat[, counter] <- IF_gmm

          # save the standard errors from GMM
          se_gt_ddd_nyt[counter] <- gmm_se

          # update counter
          counter  <- counter + 1

        } # end of GMM-based procedure

      } else { # else of if (panel)

        # ==========================================
        # RCS/UNBALANCED PANEL PATH
        # ==========================================

        # Get total number of observations (NOT unique IDs)
        # n_size = uniqueN(cohort_data[, id])

        # Filter for treated and control groups
        index_units_in_gt <- cohort_data[, treat == 1 | control == 1]
        cohort_data <- cohort_data[index_units_in_gt]

        # Save number of observations after filtering
        size_gt <- if (allow_unbalanced_panel) uniqueN(cohort_data$id) else nrow(cohort_data)
        # size_gt <- nrow(cohort_data)

        # Identify available control states
        available_controls <- sort(unique(cohort_data$first_treat[cohort_data$control == 1]))
        available_controls <- available_controls[available_controls != glist[g]]

        # ----------------------------------------
        # SINGLE CONTROL GROUP CASE (RCS)
        # ----------------------------------------
        if (length(available_controls) == 1) {

          # Get total sample size for influence function (number of unique IDs in original data)
          n_total <- uniqueN(orig_data$id)

          # Create subgroup variable
          cohort_data[, subgroup := NA_integer_]
          cohort_data[(treat == 1 & partition == 1), subgroup := 4]
          cohort_data[(treat == 1 & partition == 0), subgroup := 3]
          cohort_data[(treat == 0 & partition == 1), subgroup := 2]
          cohort_data[(treat == 0 & partition == 0), subgroup := 1]

          # Add post treatment dummy
          cohort_data[, post := 0]
          pseudo_post <- ifelse(post_treat == 1, tlist[max(t, pret) + tfac], tlist[t + tfac])
          cohort_data[(period == pseudo_post), post := 1]

          # Calculate subgroup counts - count unique IDs (handles both RCS and unbalanced panel)
          # For unbalanced panel, we need to count unique IDs in the cohort_data, not the full data
          subgroup_counts <- cohort_data[, .(count = uniqueN(id)), by = subgroup][order(-subgroup)]
          

          # Prepare did_preprocessed for att_dr_rc
          did_preprocessed$preprocessed_data <- cohort_data
          did_preprocessed$subgroup_counts <- subgroup_counts
          did_preprocessed$boot <- FALSE
          did_preprocessed$inffunc <- TRUE

          # Call att_dr_rc (RCS version)
          attgt_inf_func <- att_dr_rc(did_preprocessed)

          # Save ATT results
          attgt_list[[counter]] <- list(att = attgt_inf_func$ATT,
                                         group = glist[g],
                                         year = tlist[t + tfac],
                                         post = post_treat)

          # Populate influence function
          # Following did package approach: rescale BEFORE aggregating
          # This is important for unbalanced panels where IDs have different observation counts
          inff <- rep(0, n_total)  # Use total sample size

          # Rescale influence function (still at observation level)
          attgt_inf_func$inf_func <- (n_total / nrow(cohort_data)) * attgt_inf_func$inf_func

          # Aggregate influence function by ID (sum for repeated IDs)
          ids_in_cohort <- cohort_data$id
          aggte_inffunc <- suppressWarnings(stats::aggregate(attgt_inf_func$inf_func,
                                                              list(ids_in_cohort),
                                                              sum))

          # Match aggregated IDs to positions in unique ID index
          id_positions <- match(aggte_inffunc[, 1], id_index)
          inff[id_positions] <- aggte_inffunc[, 2]
          inf_func_mat[, counter] <- inff

          counter <- counter + 1

        # ----------------------------------------
        # MULTIPLE CONTROL GROUPS (GMM) CASE (RCS)
        # ----------------------------------------
        } else {

          # Get total sample size for influence function (number of unique IDs in original data)
          n_total <- uniqueN(orig_data$id)

          # Initialize containers
          ddd_over_controls_res <- list()
          inf_mat_local <- NULL

          # Loop over each available control state
          for (ctrl in available_controls) {

            # Subset data for this control group
            index_units_in_gt_ctrl <- cohort_data[, first_treat == glist[g] | first_treat == ctrl]
            subset_data <- cohort_data[index_units_in_gt_ctrl]
            size_gt_ctrl <- if (allow_unbalanced_panel) uniqueN(subset_data$id) else nrow(subset_data)

            # Create subgroup variable
            subset_data[, subgroup := NA_integer_]
            subset_data[(treat == 1 & partition == 1), subgroup := 4]
            subset_data[(treat == 1 & partition == 0), subgroup := 3]
            subset_data[(treat == 0 & partition == 1), subgroup := 2]
            subset_data[(treat == 0 & partition == 0), subgroup := 1]

            # Add post indicator
            subset_data[, post := 0]
            pseudo_post <- ifelse(post_treat == 1, tlist[max(t, pret) + tfac], tlist[t + tfac])
            subset_data[(period == pseudo_post), post := 1]

            # Calculate subgroup counts - count unique IDs (handles both RCS and unbalanced panel)
            subgroup_counts <- subset_data[, .(count = uniqueN(id)), by = subgroup][order(-subgroup)]

            # Prepare and run att_dr_rc
            did_preprocessed$preprocessed_data <- subset_data
            did_preprocessed$subgroup_counts <- subgroup_counts
            did_preprocessed$boot <- FALSE
            did_preprocessed$inffunc <- TRUE

            ddd_out <- att_dr_rc(did_preprocessed)

            # Save results
            ddd_over_controls_res[[as.character(ctrl)]] <- ddd_out

            # Populate influence function
            inff <- rep(0, n_total)  # Use total sample size

            # Rescale influence function (still at observation level)
            ddd_out$inf_func <- (n_total / nrow(subset_data)) * ddd_out$inf_func

            # Aggregate influence function by ID (sum for repeated IDs)
            ids_in_subset <- subset_data$id
            aggte_inffunc_ctrl <- suppressWarnings(stats::aggregate(ddd_out$inf_func,
                                                                     list(ids_in_subset),
                                                                     sum))

            # Match aggregated IDs to positions in unique ID index
            id_positions_ctrl <- match(aggte_inffunc_ctrl[, 1], id_index)
            inff[id_positions_ctrl] <- aggte_inffunc_ctrl[, 2]

            if (is.null(inf_mat_local)) {
              inf_mat_local <- matrix(inff, ncol = 1)
              colnames(inf_mat_local) <- as.character(ctrl)
            } else {
              inf_mat_local <- cbind(inf_mat_local, inff)
              colnames(inf_mat_local)[ncol(inf_mat_local)] <- as.character(ctrl)
            }
          } # end loop over available_controls

          # Skip if no estimations obtained
          if (length(ddd_over_controls_res) == 0) next

          # Compute GMM estimator
          att_vals <- sapply(ddd_over_controls_res, function(res) res$ATT)
          OMEGA <- cov(inf_mat_local, use = "complete.obs")
          inv_OMEGA <- solve(OMEGA)
          w <- colSums(inv_OMEGA) / sum(inv_OMEGA)

          ATT_gmm <- sum(w * att_vals) / sum(w)
          IF_gmm <- (inf_mat_local) %*% w
          gmm_se <- sqrt(1 / (n_total * sum(inv_OMEGA)))

          # Save results
          attgt_list[[counter]] <- list(att = ATT_gmm,
                                         group = glist[g],
                                         year = tlist[t + tfac],
                                         post = post_treat)
          inf_func_mat[, counter] <- IF_gmm
          se_gt_ddd_nyt[counter] <- gmm_se

          counter <- counter + 1

        } # end of GMM-based procedure (RCS)

      } # end of RCS/unbalanced path

    } # end of tlist loop
  } # end of glist loop

  # recover original arguments in did_preprocessed
  did_preprocessed$preprocessed_data <- orig_data
  did_preprocessed$boot <- boot # restoring boot argument

  # PREPROCESS attgt_list AND inf_func_mat
  attgt_res <- process_attgt(attgt_list)
  groups <- attgt_res$group
  periods <- attgt_res$periods
  att_gt_ddd <- attgt_res$att
  
  # ---- per-unit cluster vector aligned to the rows of inf_func_mat -------------
  # inf_func_mat rows are in id_index = unique(data$id) order (see top of att_gt()).
  # Map one cluster id per unit onto that exact ordering. `data` still carries the
  # `cluster` column from preprocessing.
  cluster      <- did_preprocessed$cluster           # cluster variable NAME (or NULL)
  cluster_vec  <- NULL
  cluster_analytic <- FALSE
  if (!is.null(cluster) && length(cluster) > 0) {
    cl_map      <- unique(data[, .(id, cluster)])     # data.table: one row per id
    cluster_vec <- cl_map$cluster[match(id_index, cl_map$id)]
    # only usable if it aligns with inf_func_mat and has no missing ids
    if (length(cluster_vec) == n && !anyNA(cluster_vec)) {
      cluster_analytic <- !boot                       # analytic branch only when not bootstrapping
    } else {
      cluster_vec <- NULL
    }
  }

  # Get analytical errors: This is analogous to cluster robust standard errors at the unit level
  n <- did_preprocessed$n
  if (cluster_analytic) {
    # Cluster-robust analytical variance: cluster sums of the influence function,
    # mirroring the cluster-sum aggregation of the multiplier bootstrap
    # (Callaway & Sant'Anna 2021, Remark 10). Same 1/n scaling as the i.i.d. V.
    Sc <- rowsum(as.matrix(inf_func_mat), cluster_vec)   # n_clusters x k cluster sums
    V  <- Matrix::Matrix(crossprod(Sc) / n)
  } else {
    V <- Matrix::t(inf_func_mat) %*% inf_func_mat / n
  }
  se_gt_ddd <- sqrt(Matrix::diag(V) / n)

  # If clustering was requested without the bootstrap but the cluster-robust V could
  # not be formed (e.g. cluster ids unavailable), warn that SEs are NOT cluster-robust.
  if (!is.null(cluster) && length(cluster) > 0 && !boot && !cluster_analytic) {
    warning("Clustered standard errors could not be computed analytically; the reported standard errors do NOT account for clustering. Set boot = TRUE for the cluster-robust multiplier bootstrap.")
  }

  # Zero standard error replaced by NA
  se_gt_ddd[se_gt_ddd <= sqrt(.Machine$double.eps)*10] <- NA

  # if control group is "notyettreated", we need to adjust the standard errors
  if (control_group == "notyettreated" && !cluster_analytic) {
    # adjust standard errors for not yet treated control group
    se_gt_ddd[!is.na(se_gt_ddd_nyt)] <- se_gt_ddd_nyt[!is.na(se_gt_ddd_nyt)]
  }
  # When cluster_analytic, the GMM (not-yet-treated) cells take their cluster-robust SE
  # from the cluster-sum V above (IF_gmm is already stored in inf_func_mat). This is
  # exact: without clustering sum(IF_gmm^2)/n == 1/sum(inv_OMEGA) == the old gmm_se.

  # Identify entries of main diagonal V that are zero or NA
  zero_na_sd_entry <- unique(which(is.na(se_gt_ddd)))

  # compute bootstrap standard errors
  bT <- NULL
  if (boot){
    # perform multiplier bootstrap
    boot_result <- mboot(inf_func_mat, did_preprocessed=did_preprocessed, use_parallel=use_parallel, cores=cores)
    bT <- boot_result$bT # sup-t confidence band
    # save bootstrap standard errors
    if(length(zero_na_sd_entry)>0) {
      se_gt_ddd[-zero_na_sd_entry] <- boot_result$se[-zero_na_sd_entry]
    } else {
      se_gt_ddd <- boot_result$se
    }
  }

  # Zero standard error replaced by NA
  se_gt_ddd[se_gt_ddd <= sqrt(.Machine$double.eps)*10] <- NA

  #-----------------------------------------------------------------------------
  # compute confidence intervals / bands
  #-----------------------------------------------------------------------------

  # compute critical values for point-wise interval
  cv <- qnorm(1 - alpha/2)

  # in case uniform confidence bands are requested
  if (boot){
    if (cband){
      # get critical value to compute uniform confidence bands
      cv <- boot_result$unif_crit_val
      if(cv >= 7){
        warning("Simultaneous critical value is arguably `too large' to be realible. This usually happens when number of observations per group is small and/or there is no much variation in outcomes.")
      }
    }
  }

  # compute confidence intervals/bands
  ci_upper <- att_gt_ddd + cv * se_gt_ddd
  ci_lower <- att_gt_ddd - cv * se_gt_ddd

  # ------------------------------------------------------------------------------
  # Return results
  # ------------------------------------------------------------------------------

  # we need this for aggregation
  # For balanced panel: use first period data (all IDs are present in first period)
  # For unbalanced panel or RCS: aggregate by ID to get one observation per unique ID
  if (panel && !allow_unbalanced_panel) {
    # Balanced panel: all IDs present in first period
    first_period_dta <- orig_data[period == tlist[1]]
  } else {
    # Unbalanced panel or RCS: take one observation per unique ID
    # This ensures we have one row per ID for aggregation weights
    first_period_dta <- orig_data[, .SD[1], by = id]
  }
  ret <- (list(ATT = att_gt_ddd,
               se = se_gt_ddd,
               uci = ci_upper,
               lci = ci_lower,
               groups = groups,
               periods = periods,
               tlist = tlist,
               glist = glist,
               cohort_size = cohort_size,
               n = n,
               bT = bT,
               inf_func_mat = inf_func_mat,
               first_period_dta = first_period_dta,
               cluster_vector = cluster_vec,   # per-unit clusters aligned to inf_func_mat rows (or NULL)
               cluster_var    = if (!is.null(cluster_vec)) cluster else NULL
              ))

  return(ret)
}

#' @import data.table
#' @importFrom stats model.matrix
#' @import BMisc
#' @export
NULL
#--------------------------------------------------
# Helper function to check partition-specific collinearity
# Returns a list with:
#   - collinear_vars: named list mapping dropped variables to the partition(s) where they're collinear
#   - all_collinear: vector of all variables that should be dropped globally
check_partition_collinearity <- function(data, subgroup_col, cov_cols, tol = 1e-6) {
  # Get unique subgroups (should be 1, 2, 3, 4)
  subgroups <- sort(unique(data[[subgroup_col]]))

  # We check comparisons: 4 vs 3, 4 vs 2, 4 vs 1
  # These are the partitions used in the DDD estimation
  comparison_groups <- subgroups[subgroups < 4]

  # Track which variables are collinear in which partition
  partition_collinear <- list()

  for (comp_group in comparison_groups) {
    # Subset data to subgroup 4 and the comparison group
    subset_data <- data[data[[subgroup_col]] %in% c(4, comp_group), ]

    # Get covariate matrix for this subset
    cov_subset <- as.matrix(subset_data[, cov_cols, with = FALSE])

    # Skip if no observations (shouldn't happen, but be safe)
    if (nrow(cov_subset) == 0) next

    # Use QR decomposition to detect collinearity
    qr_subset <- qr(cov_subset, tol = tol)
    rank_subset <- qr_subset$rank

    # Get indices of linearly independent columns
    non_collinear_indices <- qr_subset$pivot[seq_len(rank_subset)]

    # Find collinear variables in this subset
    collinear_in_subset <- setdiff(cov_cols, cov_cols[non_collinear_indices])

    if (length(collinear_in_subset) > 0) {
      partition_name <- paste0("subgroup 4 vs ", comp_group)
      for (var in collinear_in_subset) {
        if (is.null(partition_collinear[[var]])) {
          partition_collinear[[var]] <- partition_name
        } else {
          partition_collinear[[var]] <- c(partition_collinear[[var]], partition_name)
        }
      }
    }
  }

  return(list(
    collinear_vars = partition_collinear,
    all_collinear = names(partition_collinear)
  ))
}

#--------------------------------------------------
# Function to pre-process the data to use on ddd estimator

run_nopreprocess_2periods <- function(yname,
                                      tname,
                                      idname,
                                      gname,
                                      pname,
                                      xformla = ~1,
                                      dta,
                                      control_group = NULL,
                                      est_method = "dr",
                                      panel = TRUE,
                                      allow_unbalanced_panel = FALSE,
                                      learners = NULL,
                                      n_folds = NULL,
                                      weightsname = NULL,
                                      boot = FALSE,
                                      nboot = NULL,
                                      cluster = NULL,
                                      cband = FALSE,
                                      alpha = 0.05,
                                      use_parallel = FALSE,
                                      cores = NULL,
                                      inffunc = FALSE){

  arg_names <- setdiff(names(formals()), "dta")
  args <- mget(arg_names, sys.frame(sys.nframe()))

  # Flag for alpha > 0.10
  if (alpha > 0.10) {
    warning("alpha = ", alpha, " is too high. Using alpha = 0.05 as default.")
    alpha <- 0.05
    args$alpha <- alpha
  }

  # For dml, we only allow analytical standard errors. User needs to specify the number fo folds.
  # if (est_method == "dml"){
  #   # allowing bootstrap only for dr.
  #   if (boot == TRUE){
  #     warning("Bootstrapping is not allowed for DML. Setting boot = FALSE.")
  #     boot <- FALSE
  #     args$boot <- boot
  #   }
  #   # number of folds
  #   if (n_folds < 2 || is.null(n_folds)) {
  #     warning("For DML estimation, n_folds must be at least 2. Setting n_folds = 2.")
  #     n_folds <- 2
  #     args$n_folds <- n_folds
  #   }
  # }

  # setting default bootstrap reps
  if (boot == TRUE){
    if (is.null(nboot)){
      warning("Number of bootstrap samples not specified. Defaulting to 999 reps.")
      nboot <- 999
      args$nboot <- nboot
    }

    if(!cband){
      warning("cband = FALSE. Setting cband = TRUE for bootstrapped standard errors.")
      cband <- TRUE
      args$cband <- cband
    }
  }

  # Flags for cluster variable
  if (!is.null(cluster)) {
    # dropping idname from cluster
    if (idname %in% cluster) {
      cluster <- setdiff(cluster, idname)
    }

    # flag if cluster variable is in dataset
    if (!is.element(cluster, base::colnames(dta))) {
      stop("cluster = ",cluster,  " could not be found in the data provided. Please check your arguments")
    }

    # check if user is providing more than 2 cluster variables (different than idname)
    if (length(cluster) > 1) {
      stop("You can only provide 1 cluster variable additionally to the one provided in idname. Please check your arguments")
    }

    # check if bootstrap is on
    if (!boot){
      warning("Clustered SEs are only available when boot=TRUE. Setting boot=TRUE and cband=TRUE for bootstrapped standard errors.")
      boot <- TRUE
      args$boot <- boot
      cband <- TRUE
      args$cband <- cband
    }
  }


  # set weights
  base::ifelse(is.null(weightsname), weights <- rep(1, nrow(dta)), weights <- dta[[weightsname]])
  # Check for missing values in the weights vector
  if (anyNA(weights)) {
    stop("There are missing values in the weights column. Please check them.")
  }
  # enforcing normalization of weights
  weights <- weights/mean(weights)
  dta$weights <- weights


  # Flag for xformla
  if (is.null(xformla)) {
    xformla <- ~1
  }

  # Creating a post dummy variable based on tlist[2] (second period = post treatment)
  tlist <- dta[, sort(unique(get(tname)))]
  glist <- dta[, sort(unique(get(gname)))] # this has to be [0, G], second position in treated group.
  dta[, post := as.numeric(get(tname) == tlist[2])]

  # sort data based on idnam and tname
  data.table::setorderv(dta, c(idname, tname), c(1,1))

  #-------------------------------------
  # Handle RCS ID creation
  #-------------------------------------

  # If idname is the placeholder, create the actual ID column
  if (idname == ".row_id") {
    if (".row_id" %in% names(dta)) {
      stop("Column '.row_id' already exists in data. Please rename it or provide a different idname.")
    }
    # Create unique row IDs for RCS data
    dta[, .row_id := .I]
  }

  # -----------------------------------------------------------------------
  # Panel balance detection and handling
  # -----------------------------------------------------------------------

  # Detect if panel is balanced
  n_old <- uniqueN(dta[[idname]])
  row_orig <- dta[, .N]

  # Check if panel is balanced (all units observed in all periods)
  obs_per_unit <- dta[, .N, by = idname]
  n_periods <- uniqueN(dta[[tname]])
  is_balanced <- all(obs_per_unit$N == n_periods)

  # Initialize flag for true repeated cross-sections
  true_repeated_cross_sections <- FALSE

  if (panel) {
    if (!is_balanced) {
      if (allow_unbalanced_panel) {
        # Treat unbalanced panel as repeated cross-sections
        true_repeated_cross_sections <- TRUE
        warning("Panel is unbalanced. Treating as repeated cross-sections (allow_unbalanced_panel = TRUE).")
      } else {
        # Balance the panel by dropping units
        dta <- BMisc::make_balanced_panel(dta, idname, tname)
        n <- uniqueN(dta[[idname]])
        row_new <- dta[, .N]
        row_diff <- row_orig - row_new

        if (n < n_old) {
          warning(paste0("Dropped observations from ", n_old - n, " units ", "(", row_diff ," rows)" ," while converting to balanced panel."))
        }

        # Check if all data was dropped
        if (nrow(dta) == 0) {
          stop("All observations dropped to convert data to balanced panel. Please check your argument in 'idname' or set panel=FALSE for repeated cross-sections.")
        }
      }
    }
  } else {
    # panel = FALSE: treat as repeated cross-sections
    true_repeated_cross_sections <- TRUE
  }

  #-------------------------------------
  # Handle RCS ID creation
  #-------------------------------------

  # If idname is the placeholder, create the actual ID column in dta
  if (idname == ".row_id") {
    if (".row_id" %chin% names(dta)) {
      stop("Column '.row_id' already exists in data. Please rename it or provide a different idname.")
    }
    # Create unique row IDs for RCS data
    dta[, (idname) := .I]
  }

  #-------------------------------------
  # Generate the output for estimation
  # -------------------------------------

  cleaned_data <- data.table::data.table(id = dta[[idname]],
                                         y = dta[[yname]],
                                         post = dta$post,
                                         treat = dta[[gname]],
                                         period = dta[[tname]],
                                         partition = dta[[pname]],
                                         weights = dta$weights)
  idx_static_vars <- 8 # useful to perform the elimination of collinear variables
  # Add cluster column if cluster argument is provided
  if (!is.null(cluster)) {
    cleaned_data[, cluster := dta[[cluster]]]
    idx_static_vars <- 9
  }

  # creating subgroup variable
  # 4 if (partition ==1 & treat == 1); 3 if (partition ==0 & treat == 1); 2 if (partition ==1 & treat == 0); 1 if (partition ==0 & treat == 0)
  cleaned_data[, subgroup := fifelse(partition == 1 & treat == glist[2], 4,
                                     fifelse(partition == 0 & treat == glist[2], 3,
                                             fifelse(partition == 1 & treat == glist[1], 2, 1)))]

  # Flag for not enough observations for each subgroup
  # Calculate the size of each subgroup in the 'subgroup' column
  # For unbalanced panel: count unique IDs (each ID may appear multiple times)
  # For true RCS: count observations (each observation is unique)
  # For balanced panel: count observations / 2 (each unit appears twice)
  if (true_repeated_cross_sections) {
    if (panel && allow_unbalanced_panel) {
      # Unbalanced panel: count unique IDs per subgroup
      subgroup_counts <- cleaned_data[, .(count = uniqueN(id)), by = subgroup][order(-subgroup)]
    } else {
      # True RCS: count observations (each observation is unique)
      subgroup_counts <- cleaned_data[, .(count = .N), by = subgroup][order(-subgroup)]
    }
  } else {
    # Balanced panel: count observations and divide by 2
    subgroup_counts <- cleaned_data[, .(count = .N/2), by = subgroup][order(-subgroup)]
  }

  # adding covariates
  if (!is.null(xformla)) {
    cleaned_data <- cbind(cleaned_data, stats::model.matrix(xformla,
                                                            data = dta,
                                                            na.action = na.pass))
  }

  # Remove rows with missing values
  cleaned_data <- na.omit(cleaned_data)

  # Remove collinear variables
  # Convert the remaining columns to a matrix
  cov_m <- as.matrix(cleaned_data[, -c(1:idx_static_vars), with = FALSE])
  # Use the qr() function to detect collinear columns
  qr_m <- qr(cov_m, tol = 1e-6)
  # Get the rank of the matrix
  rank_m <- qr_m$rank
  # Get the indices of the non-collinear columns
  non_collinear_indices <- qr_m$pivot[seq_len(rank_m)]
  # Check if any covariates were dropped due to global collinearity (following DRDID approach)
  dropped_covariates_global <- setdiff(colnames(cov_m), colnames(cov_m)[non_collinear_indices])
  if (length(dropped_covariates_global) > 0) {
    warning("The following covariates were dropped due to global collinearity: ", paste(dropped_covariates_global, collapse = ", "))
  }
  # Drop the collinear columns from the data.table
  cleaned_data <- cleaned_data[, c(seq(1,idx_static_vars,1), non_collinear_indices + idx_static_vars), with = FALSE]

  # drop the intercept
  #cleaned_data[, (idx_static_vars+1) := NULL]
  cleaned_data[, "(Intercept)" := NULL]

  # Check for partition-specific collinearity (after global check)
  # Get remaining covariate column names (excluding static vars)
  cov_cols_remaining <- setdiff(names(cleaned_data), c("id", "y", "post", "treat", "period", "partition", "weights", "cluster", "subgroup"))
  if (length(cov_cols_remaining) > 0) {
    partition_check <- check_partition_collinearity(cleaned_data, "subgroup", cov_cols_remaining)

    if (length(partition_check$all_collinear) > 0) {
      # Build informative warning message
      partition_warnings <- sapply(names(partition_check$collinear_vars), function(var) {
        partitions <- paste(partition_check$collinear_vars[[var]], collapse = ", ")
        paste0("  - ", var, " (collinear in: ", partitions, ")")
      })
      warning("The following covariates were dropped due to partition-specific collinearity:\n",
              paste(partition_warnings, collapse = "\n"))

      # Drop these covariates globally
      cols_to_keep <- setdiff(names(cleaned_data), partition_check$all_collinear)
      cleaned_data <- cleaned_data[, ..cols_to_keep]
    }
  }

  # Update xformula to reflect dropped covariates (following DRDID approach)
  # Get final covariate column names after all collinearity checks
  final_cov_cols <- setdiff(names(cleaned_data), c("id", "y", "post", "treat", "period", "partition", "weights", "cluster", "subgroup"))
  if (length(final_cov_cols) > 0) {
    xformla <- stats::as.formula(paste("~", paste(final_cov_cols, collapse = " + ")))
  } else {
    xformla <- ~1
  }

  out <- list(preprocessed_data = cleaned_data,
              xformula = xformla,
              est_method = est_method,
              learners = learners,
              n_folds = n_folds,
              boot = boot,
              nboot = nboot,
              cluster = cluster,
              cband = cband,
              alpha = alpha,
              use_parallel = use_parallel,
              cores = cores,
              inffunc = inffunc,
              subgroup_counts = subgroup_counts,
              panel = panel,
              allow_unbalanced_panel = allow_unbalanced_panel,
              true_repeated_cross_sections = true_repeated_cross_sections)

  return(out)

}


# Preprocess function for 2 periods case
run_preprocess_2Periods <- function(yname,
                                   tname,
                                   idname,
                                   gname,
                                   pname,
                                   xformla = ~1,
                                   dta,
                                   control_group = NULL,
                                   est_method = "dr",
                                   panel = TRUE,
                                   allow_unbalanced_panel = FALSE,
                                   learners = NULL,
                                   n_folds = NULL,
                                   weightsname = NULL,
                                   boot = FALSE,
                                   nboot = NULL,
                                   cluster = NULL,
                                   cband = FALSE,
                                   alpha = 0.05,
                                   use_parallel = FALSE,
                                   cores = NULL,
                                   inffunc = FALSE){

  # Capture all arguments except 'data'
  arg_names <- setdiff(names(formals()), "dta")
  args <- mget(arg_names, sys.frame(sys.nframe()))

  #-------------------------------------
  # Error checking
  #-------------------------------------

  # Flag for alpha > 0.10
  if (alpha > 0.10) {
    warning("alpha = ", alpha, " is too high. Using alpha = 0.05 as default.")
    alpha <- 0.05
    args$alpha <- alpha
  }

  # For dml, we only allow analytical standard errors.
  # if (est_method == "dml" & boot == TRUE){
  #   warning("Bootstrapping is not allowed for DML. Setting boot = FALSE.")
  #   boot <- FALSE
  #   args$boot <- boot
  # }

  # setting default bootstrap reps
  if (boot){
    if (is.null(nboot)){
      warning("Number of bootstrap samples not specified. Defaulting to 999 reps.")
      nboot <- 999
      args$nboot <- nboot
    }

    if(!cband){
      warning("cband = FALSE. Setting cband=TRUE for bootstrapped standard errors.")
      cband <- TRUE
      args$cband <- cband
    }
  }

  #-------------------------------------
  # Handle RCS ID creation
  #-------------------------------------

  # If idname is the placeholder, create the actual ID column
  if (idname == ".row_id") {
    if (".row_id" %in% names(dta)) {
      stop("Column '.row_id' is an internal name. Please rename it or provide a different idname.")
    }
    # Create unique row IDs for RCS data
    dta[, .row_id := .I]
  }

  # Run argument checks
  validate_args_2Periods(args, dta)

  # Flags for cluster variable
  if (!is.null(cluster)) {
    # dropping idname from cluster
    if (idname %in% cluster) {
      cluster <- setdiff(cluster, idname)
    }

    # flag if cluster variable is in dataset
    if (!is.element(cluster, base::colnames(dta))) {
      stop("cluster = ",cluster,  " could not be found in the data provided. Please check your arguments.")
    }

    # check if user is providing more than 2 cluster variables (different than idname)
    if (length(cluster) > 1) {
      stop("You can only provide 1 cluster variable additionally to the one provided in idname. Please check your arguments.")
    }

    # Check that cluster variables do not vary over time within each unit
    if (length(cluster) > 0) {
      # Efficiently check for time-varying cluster variables
      clust_tv <- dta[, lapply(.SD, function(col) length(unique(col)) == 1), by = idname, .SDcols = cluster]
      # If any cluster variable varies over time within any unit, stop execution
      if (!all(unlist(clust_tv[, -1, with = FALSE]))) {
        stop("triplediff cannot handle time-varying cluster variables at the moment. Please check your cluster variable.")
      }
    }

    # check if bootstrap is on
    if (!boot){
      warning("Clustered SEs are only available when boot=TRUE. Setting boot=TRUE and cband=TRUE for bootstrapped standard errors.")
      boot <- TRUE
      args$boot <- boot
      cband <- TRUE
      args$cband <- cband

      # adding boot reps too
      if (is.null(nboot)){
        warning("Number of bootstrap samples not specified. Defaulting to 999 reps.")
        nboot <- 999
        args$nboot <- nboot
      }
    }
  }

  # Flag for parallel and cores
  if (boot){
    if ((use_parallel) && (is.null(cores))) {
      warning("Parallel processing is enabled but the number of cores is not specified. Using 1 core as default.")
      cores <- 1
      args$cores <- cores
    }
  }

  # set weights
  base::ifelse(is.null(weightsname), weights <- rep(1, nrow(dta)), weights <- dta[[weightsname]])
  # Check for missing values in the weights vector
  if (anyNA(weights)) {
    stop("There are missing values in the weights column. Please check them.")
  }
  # enforcing normalization of weights
  weights <- weights/mean(weights)
  dta$weights <- weights

  # Check if weights are unique by idname
  checkWeightsUniqueness(dta, idname)


  # Flag for xformla
  if (is.null(xformla)) {
    xformla <- ~1
  }

  tryCatch(
    expr = {
      # attempt to convert xfomrla argument to a formula
      xformla <- as.formula(xformla)
    },
    error = function(e) {
      # if an error is thrown, stop execution and throw an error
      stop("xformla is not a valid formula. Please check xformla", e$message)
    }
  )

  # Creating a post dummy variable based on tlist[2] (second period = post treatment)
  tlist <- dta[, sort(unique(get(tname)))]
  glist <- dta[, sort(unique(get(gname)))] # this has to be [0, G], second position in treated group.
  dta[, post := as.numeric(get(tname) == tlist[2])]

  # Checking if covariates are time invariant in panel data case
  # Only check for panel data (not RCS)
  if (panel) {
    # Create the model matrix
    cov_pre <- model.matrix(as.formula(xformla), data = dta[dta[["post"]] == 0])
    cov_post <- model.matrix(as.formula(xformla), data = dta[dta[["post"]] == 1])
    # Check if covariates are time invariant
    if (!all(cov_pre == cov_post)) {
      stop("Covariates are not time invariant. Please check xformla")
    }
  }

  # sort data based on idnam and tname
  data.table::setorderv(dta, c(idname, tname), c(1,1))

  # -----------------------------------------------------------------------
  # Panel balance detection and handling
  # -----------------------------------------------------------------------

  # Detect if panel is balanced
  n_old <- uniqueN(dta[[idname]])
  row_orig <- dta[, .N]

  # Check if panel is balanced (all units observed in all periods)
  obs_per_unit <- dta[, .N, by = idname]
  n_periods <- uniqueN(dta[[tname]])
  is_balanced <- all(obs_per_unit$N == n_periods)

  # Initialize flag for true repeated cross-sections
  true_repeated_cross_sections <- FALSE

  if (panel) {
    if (!is_balanced) {
      if (allow_unbalanced_panel) {
        # Treat unbalanced panel as repeated cross-sections
        true_repeated_cross_sections <- TRUE
        warning("Panel is unbalanced. Proceeding as such.")
      } else {
        # Balance the panel by dropping units
        dta <- BMisc::make_balanced_panel(dta, idname, tname)
        n <- uniqueN(dta[[idname]])
        row_new <- dta[, .N]
        row_diff <- row_orig - row_new

        # Check if all data was dropped
        if (nrow(dta) == 0) {
          stop("All observations dropped to convert data to balanced panel. Please check your argument in 'idname' or set panel=FALSE for repeated cross-sections.")
        }

        if (n < n_old) {
          warning(paste0("Dropped observations from ", n_old - n, " units ", "(", row_diff ," rows)" ," while converting to balanced panel."))
        }
      }
    }
  } else {
    # panel = FALSE: treat as repeated cross-sections
    true_repeated_cross_sections <- TRUE
  }

  #-------------------------------------
  # Generate the output for estimation
  # -------------------------------------

  cleaned_data <- data.table::data.table(id = dta[[idname]],
                                         y = dta[[yname]],
                                         post = dta$post,
                                         treat = dta[[gname]],
                                         period = dta[[tname]],
                                         partition = dta[[pname]],
                                         weights = dta$weights)

  # # For true RCS data with user-provided idname, replace with row IDs
  # if (true_repeated_cross_sections && idname != ".row_id") {
  #   cleaned_data[, id := .I]  # Use row index as ID for RCS
  # }

  idx_static_vars <- 8 # useful to perform the elimination of collinear variables
  # Add cluster column if cluster argument is provided
  if (!is.null(cluster)) {
    cleaned_data[, cluster := dta[[cluster]]]
    idx_static_vars <- 9
  }

  # creating subgroup variable
  # 4 if (partition ==1 & treat == 1); 3 if (partition ==0 & treat == 1); 2 if (partition ==1 & treat == 0); 1 if (partition ==0 & treat == 0)
  cleaned_data[, subgroup := fifelse(partition == 1 & treat == glist[2], 4,
                                     fifelse(partition == 0 & treat == glist[2], 3,
                                             fifelse(partition == 1 & treat == glist[1], 2, 1)))]

  # Flag for not enough observations for each subgroup
  # Calculate the size of each subgroup in the 'subgroup' column
  # For unbalanced panel: count unique IDs (each ID may appear multiple times)
  # For true RCS: count observations (each observation is unique)
  # For balanced panel: count observations / 2 (each unit appears twice)
  if (true_repeated_cross_sections) {
    if (panel && allow_unbalanced_panel) {
      # Unbalanced panel: count unique IDs per subgroup
      subgroup_counts <- cleaned_data[, .(count = uniqueN(id)), by = subgroup][order(-subgroup)]
    } else {
      # True RCS: count observations (each observation is unique)
      subgroup_counts <- cleaned_data[, .(count = .N), by = subgroup][order(-subgroup)]
    }
  } else {
    # Balanced panel: count observations and divide by 2
    subgroup_counts <- cleaned_data[, .(count = .N/2), by = subgroup][order(-subgroup)]
  }
  # Check if each subgroup has at least 5 observations. Check this threshold if needed.
  sufficient_obs <- all(subgroup_counts$count >= 5)
  # Stop the code if not all subgroups have at least 5 observations
  if (!sufficient_obs) {
    stop("Not enough observations in each subgroup to perform inference. Please check the distribution of 'gname' and 'pname' in the data.")
  }

  # Flag for small groups for inference
  # Calculate the size of each group in the 'treat' column
  # Subset 'gsize' to get the small groups based on "required size"
  gsize <- cleaned_data[, if(.N < length(BMisc::rhs_vars(xformla)) + 5) .N, by = "treat"]
  # If there are any small groups, stop execution and print an error message
  if (nrow(gsize) > 0) {
    stop("Either treatment or the comparison group in your dataset is very small. Inference is not feasible.")
  }

  # adding covariates
  if (!is.null(xformla)) {
    cleaned_data <- cbind(cleaned_data, stats::model.matrix(xformla,
                                                          data = dta,
                                                          na.action = na.pass))
  }

  # Check for missing values in several columns
  #missing_flags <- lapply(cleaned_data[, .(treat, post, y, weights)], anyNA)
  missing_flags <- lapply(cleaned_data[, .SD, .SDcols = c("treat", "post", "y", "weights")], anyNA)
  # all except id, y, post, treat, partition, period, subgroup, weights, cluster (if any)
  missing_X_flag <- base::anyNA(cleaned_data[,-c(1:idx_static_vars), with = FALSE])
  # Print warning messages if any missing values are found
  if (any(unlist(missing_flags))) {
    # Get the names of the TRUE values
    var.with.na <- names(missing_flags)[which(unlist(missing_flags))]
    # Print warning messages
    warning(paste("Some", var.with.na, "contain missing values (NA). We are dropping those observations"))
  }
  if(missing_X_flag) warning("Some covariates are missing (NA). We are dropping those observations")

  # Remove rows with missing values
  cleaned_data <- na.omit(cleaned_data)

  # Remove collinear variables
  # Convert the remaining columns to a matrix
  cov_m <- as.matrix(cleaned_data[, -c(1:idx_static_vars), with = FALSE])
  # Use the qr() function to detect collinear columns
  qr_m <- qr(cov_m, tol = 1e-6)
  # Get the rank of the matrix
  rank_m <- qr_m$rank
  # Get the indices of the non-collinear columns
  non_collinear_indices <- qr_m$pivot[seq_len(rank_m)]
  # Check if any covariates were dropped due to global collinearity (following DRDID approach)
  dropped_covariates_global <- setdiff(colnames(cov_m), colnames(cov_m)[non_collinear_indices])
  if (length(dropped_covariates_global) > 0) {
    warning("The following covariates were dropped due to global collinearity: ", paste(dropped_covariates_global, collapse = ", "))
  }
  # Drop the collinear columns from the data.table
  cleaned_data <- cleaned_data[, c(seq(1,idx_static_vars,1), non_collinear_indices + idx_static_vars), with = FALSE]

  # drop the intercept
  cleaned_data[, "(Intercept)" := NULL]

  # Check for partition-specific collinearity (after global check)
  # Get remaining covariate column names (excluding static vars)
  cov_cols_remaining <- setdiff(names(cleaned_data), c("id", "y", "post", "treat", "period", "partition", "weights", "cluster", "subgroup"))
  if (length(cov_cols_remaining) > 0) {
    partition_check <- check_partition_collinearity(cleaned_data, "subgroup", cov_cols_remaining)

    if (length(partition_check$all_collinear) > 0) {
      # Build informative warning message
      partition_warnings <- sapply(names(partition_check$collinear_vars), function(var) {
        partitions <- paste(partition_check$collinear_vars[[var]], collapse = ", ")
        paste0("  - ", var, " (collinear in: ", partitions, ")")
      })
      warning("The following covariates were dropped due to partition-specific collinearity:\n",
              paste(partition_warnings, collapse = "\n"))

      # Drop these covariates globally
      cols_to_keep <- setdiff(names(cleaned_data), partition_check$all_collinear)
      cleaned_data <- cleaned_data[, ..cols_to_keep]
    }
  }

  # Update xformula to reflect dropped covariates (following DRDID approach)
  # Get final covariate column names after all collinearity checks
  final_cov_cols <- setdiff(names(cleaned_data), c("id", "y", "post", "treat", "period", "partition", "weights", "cluster", "subgroup"))
  if (length(final_cov_cols) > 0) {
    xformla <- stats::as.formula(paste("~", paste(final_cov_cols, collapse = " + ")))
  } else {
    xformla <- ~1
  }

  out <- list(preprocessed_data = cleaned_data,
              xformula = xformla,
              tname = tname,
              est_method = est_method,
              learners = learners,
              n_folds = n_folds,
              boot = boot,
              nboot = nboot,
              cluster = cluster,
              cband = cband,
              alpha = alpha,
              use_parallel = use_parallel,
              cores = cores,
              inffunc = inffunc,
              subgroup_counts = subgroup_counts,
              panel = panel,
              allow_unbalanced_panel = allow_unbalanced_panel,
              true_repeated_cross_sections = true_repeated_cross_sections)

  return(out)
}

# Preprocess function for multiple periods case.
# @keywords internal
# @export
run_preprocess_multPeriods <- function(yname,
                                       tname,
                                       idname,
                                       gname,
                                       pname,
                                       xformla = ~1,
                                       dta,
                                       control_group,
                                       base_period,
                                       est_method = "dr",
                                       panel = TRUE,
                                       allow_unbalanced_panel = FALSE,
                                       learners = NULL,
                                       n_folds = NULL,
                                       weightsname = NULL,
                                       boot = FALSE,
                                       nboot = NULL,
                                       cluster = NULL,
                                       cband = FALSE,
                                       alpha = 0.05,
                                       use_parallel = FALSE,
                                       cores = NULL,
                                       inffunc = FALSE){

  # Capture all arguments except 'data'
  arg_names <- setdiff(names(formals()), "dta")
  args <- mget(arg_names, sys.frame(sys.nframe()))

  #-------------------------------------
  # Handle missing idname for RCS data
  #-------------------------------------

  # If idname is missing and panel=FALSE, set placeholder for RCS
  if (missing(idname) || is.null(idname)) {
    if (!panel) {
      idname <- ".row_id"
      args$idname <- idname
    } else {
      stop("idname is required when panel = TRUE. For repeated cross-section data, set panel = FALSE.")
    }
  }

  #-------------------------------------
  # Error checking
  #-------------------------------------

  # Flag for alpha > 0.10
  if (alpha > 0.10) {
    warning("alpha = ", alpha, " is too high. Using alpha = 0.05 as default.")
    alpha <- 0.05
    args$alpha <- alpha
  }

  # For dml, we only allow analytical standard errors.
  # if (est_method == "dml" & boot == TRUE){
  #   warning("Bootstrapping is not currently allowed for DML. Setting boot = FALSE.")
  #   boot <- FALSE
  #   args$boot <- boot
  # }

  # setting default bootstrap reps
  if (boot){
    if (is.null(nboot)){
      warning("Number of bootstrap samples not specified. Defaulting to 999 reps.")
      nboot <- 999
      args$nboot <- nboot
    }

    if(!cband){
      warning("cband = FALSE. Setting cband=TRUE for bootstrapped standard errors.")
      cband <- TRUE
      args$cband <- cband
    }
  }

  # Run argument checks
  validate_args_multPeriods(args, dta)

  # Flags for cluster variable
  if (!is.null(cluster)) {
    # dropping idname from cluster
    if (idname %in% cluster) {
      cluster <- setdiff(cluster, idname)
    }

    # flag if cluster variable is in dataset
    if (!is.element(cluster, base::colnames(dta))) {
      stop("cluster = ",cluster,  " could not be found in the data provided. Please check your arguments.")
    }

    # check if user is providing more than 2 cluster variables (different than idname)
    if (length(cluster) > 1) {
      stop("You can only provide 1 cluster variable additionally to the one provided in idname. Please check your arguments.")
    }

    # Check that cluster variables do not vary over time within each unit
    if (length(cluster) > 0) {
      # Efficiently check for time-varying cluster variables
      clust_tv <- dta[, lapply(.SD, function(col) length(unique(col)) == 1), by = idname, .SDcols = cluster]
      # If any cluster variable varies over time within any unit, stop execution
      if (!all(unlist(clust_tv[, -1, with = FALSE]))) {
        stop("triplediff cannot handle time-varying cluster variables at the moment. Please check your cluster variable.")
      }
    }

  }

  # Flag for parallel and cores
  if (boot){
    if ((use_parallel) && (is.null(cores))) {
      warning("Parallel processing is enabled but the number of cores is not specified. Using 1 core.")
      cores <- 1
      args$cores <- cores
    }
  }

  # set in-blank xformla if no covariates are provided
  if (is.null(xformla)) {
    xformla <- ~1
  }

  # keep relevant columns in data
  # For RCS placeholder, exclude idname from initial column selection (will be created later)
  if (idname == ".row_id") {
    cols_to_keep <- c(tname, yname, gname, pname, weightsname, cluster)
  } else {
    cols_to_keep <- c(idname, tname, yname, gname, pname, weightsname, cluster)
  }

  model_frame <- model.frame(xformla, data = dta, na.action = na.pass)
  # Subset the data.table to keep only relevant columns
  dta <- dta[, ..cols_to_keep]

  # Bind the model frame columns
  dta <- cbind(dta, as.data.table(model_frame))

  # Check if any covariates were missing
  n_orig <- dta[, .N]
  dta <- dta[complete.cases(dta)]
  n_new <- dta[, .N]
  n_diff <- n_orig - n_new
  if (n_diff != 0) {
    warning(paste0("Dropped ", n_diff, " rows from original data due to missing data."))
  }

  # set weights
  base::ifelse(is.null(weightsname), weights <- rep(1, n_new), weights <- dta[[weightsname]])
  # enforcing normalization of weights. At this point we already drop any missing values in weights.
  weights <- weights/mean(weights)
  dta$weights <- weights

  # get a list of dates from min to max
  tlist <- dta[, sort(unique(get(tname)))]

  # Coerce control group identified with Inf as zero
  dta[get(gname) == Inf, (gname) := 0]

  # Identify groups with treatment time bigger than the maximum treatment time
  # calculate the maximum treatment time
  max_treatment_time <- max(tlist, na.rm = TRUE)
  dta[, asif_never_treated := (get(gname) > max_treatment_time)]
  # replace NA values with FALSE in the logical vector
  dta[is.na(asif_never_treated), asif_never_treated := FALSE]
  # set gname to 0 for those groups considered as never treated
  dta[asif_never_treated == TRUE, (gname) := 0]
  # remove the temporary column
  dta[, asif_never_treated := NULL]

  # get list of treated groups (by time) from min to max
  glist <- sort(unique(dta[[gname]]))

  # Check if there is a never treated group
  if (0 %in% glist == FALSE) {
    if (control_group == "nevertreated") {
      stop("There is no available never-treated group")
    } else {
      # Drop all time periods with time periods >= latest treated
      latest_treated_time <- max(glist)
      # with anticipation?
      #latest_treated_time <- max(glist) - anticipation
      dta <- dta[get(tname) < latest_treated_time]

      tlist <- sort(unique(dta[[tname]]))
      glist <- sort(unique(dta[[gname]]))

      # don't compute ATT(g,t) for groups that are only treated at end
      # and only play a role as a comparison group
      glist <- glist[glist < max(glist)]
    }
  }

  # Get the first period
  first_period <- tlist[1]

  # Filter the treated groups
  # glist <- glist[glist > 0 & glist > (first_period + anticipation)]
  glist <- glist[glist > 0 & glist > first_period]

  # Check for groups treated in the first periods and drop them
  # identify groups treated in the first period
  dta[, treated_first_period := (get(gname) <= first_period) & (get(gname) != 0)]
  dta[is.na(treated_first_period), treated_first_period := FALSE]

  # count the number of units treated in the first period
  # For RCS placeholder, skip this check (will be handled after ID creation)
  if (idname == ".row_id") {
    nfirstperiod <- 0
  } else {
    nfirstperiod <- uniqueN(dta[treated_first_period == TRUE, get(idname)])
  }

  # handle units treated in the first period
  if (nfirstperiod > 0) {
    warning(paste0("Dropped ", nfirstperiod, " units that were already treated in the first period."))
    dta <- dta[get(gname) %in% c(0, glist)]

    # update tlist and glist
    tlist <- dta[, sort(unique(get(tname)))]
    glist <- dta[, sort(unique(get(gname)))]
    glist <- glist[glist > 0]

    # Drop groups treated in the first period or before
    first_period <- tlist[1]
    #glist <- glist[glist > first_period + anticipation]
    glist <- glist[glist > first_period]
  }
  # remove the temporary column
  dta[, treated_first_period := NULL]

  # ----------------------------------------------------------------
  # do we have missing data?
  # ----------------------------------------------------------------

  # Check for complete cases
  keepers <- complete.cases(dta)
  # For RCS placeholder, skip ID-based counting (will use row counts instead)
  if (idname == ".row_id") {
    n <- nrow(dta)
    n_keep <- sum(keepers)
  } else {
    n <- uniqueN(dta[[idname]])
    n_keep <- uniqueN(dta[keepers, get(idname)])
  }

  if (nrow(dta[keepers, ]) < nrow(dta)) {
    warning(paste0("Dropped ", (n - n_keep), " observations that had missing data."))
    dta <- dta[keepers, ]
  }

  # -----------------------------------------------------------------------
  # Panel balance detection and handling
  # -----------------------------------------------------------------------

  # Detect if panel is balanced
  # For RCS placeholder, skip balance checking (will be handled after ID creation)
  if (idname == ".row_id") {
    n_old <- nrow(dta)
    row_orig <- dta[, .N]
    is_balanced <- FALSE
  } else {
    n_old <- uniqueN(dta[[idname]])
    row_orig <- dta[, .N]

    # Check if panel is balanced (all units observed in all periods)
    obs_per_unit <- dta[, .N, by = idname]
    n_periods <- uniqueN(dta[[tname]])
    is_balanced <- all(obs_per_unit$N == n_periods)
  }

  # Initialize flag for true repeated cross-sections
  true_repeated_cross_sections <- FALSE

  if (panel) {
    if (!is_balanced) {
      if (allow_unbalanced_panel) {
        # Treat unbalanced panel as repeated cross-sections
        true_repeated_cross_sections <- FALSE
        warning("Panel is unbalanced. Proceeding as such.")
      } else {
        warning("Panel is unbalanced but allow_unbalanced_panel = FALSE. Forcing to convert data to balanced panel.")
        # Balance the panel by dropping units
        dta <- BMisc::make_balanced_panel(dta, idname, tname)
        n <- uniqueN(dta[[idname]])
        row_new <- dta[, .N]
        row_diff <- row_orig - row_new

        # Check if all data was dropped
        if (nrow(dta) == 0) {
          stop("All observations dropped to convert data to balanced panel. Please check your argument in 'idname' or set panel=FALSE for repeated cross-sections.")
        }

        if (n < n_old) {
          warning(paste0("Dropped observations from ", n_old - n, " units ", "(", row_diff ," rows)" ," while converting to balanced panel."))
        }
      }
    }
  } else {
    # panel = FALSE: treat as repeated cross-sections
    true_repeated_cross_sections <- TRUE
  }

  #-------------------------------------
  # Handle RCS ID creation
  #-------------------------------------

  # If idname is the placeholder, create the actual ID column in dta
  if (idname == ".row_id") {
    if (".row_id" %in% names(dta)) {
      stop("Column '.row_id' already exists in data. Please rename it or provide a different idname.")
    }
    # Create unique row IDs for RCS data
    dta[, .row_id := .I]
  }

  #-------------------------------------
  # Generate the output for estimation
  # ------------------------------------
  cleaned_data <- data.table::data.table(id = dta[[idname]],
                                         y = dta[[yname]],
                                         first_treat = dta[[gname]],
                                         period = dta[[tname]],
                                         partition = dta[[pname]],
                                         weights = dta$weights)

  # Add cluster column if cluster argument is provided
  if (!is.null(cluster)) {
    cleaned_data[, cluster := dta[[cluster]]]
  }

  # TODO: we may not need this as we already handled this above
  # # If drop all data, you do not have a panel.
  # if (nrow(cleaned_data) == 0) {
  #   stop("All observations dropped to convert data to balanced panel. Please check your argument in 'idname'.")
  # }

  # Compute sample size n = number of unique IDs
  # For panel: each unit appears multiple times
  # For RCS: each observation has unique ID after preprocessing
  n <- uniqueN(cleaned_data$id)

  # Check if groups is empty
  if(length(glist)==0){
    stop("No valid groups. The variable in 'gname' should be expressed as the time a unit is first treated (0 if never-treated).")
  }

  # Check for small comparison groups
  # Calculate the size of each group in the 'treat' column
  # For unbalanced panel: count unique IDs (each ID may appear in multiple periods)
  # For true RCS: count observations (each observation is unique)
  # For balanced panel: count observations / number of periods (each unit appears in all periods)
  if (true_repeated_cross_sections){
    # True RCS: count observations
      gsize <- cleaned_data[, .(N = .N), by = first_treat][order(-first_treat)]
  } else {
    if (panel && allow_unbalanced_panel) {
      # Unbalanced panel: count unique IDs per treatment group
      gsize <- cleaned_data[, .(N = uniqueN(id)), by = first_treat][order(-first_treat)]
    } else {
      # Balanced panel: count observations and divide by number of periods
      gsize <- cleaned_data[, .(N = .N / length(tlist)), by = first_treat][order(-first_treat)]
    }
  }

  # Calculate the required size
  reqsize <- length(BMisc::rhs_vars(xformla)) + 5
  # Identify groups to warn about
  small_groups <- gsize[N < reqsize]
  # Warn if some groups are small
  if (nrow(small_groups) > 0) {
    gpaste <- paste(small_groups[, first_treat], collapse = ",")
    warning(paste0("Be aware that there are some small groups in your dataset.\n  Check groups: ", gpaste, "."))

    if (0 %in% small_groups[, first_treat] & control_group == "nevertreated") {
      stop("Never treated group is too small, try setting control_group = \"notyettreated\"")
    }
  }

  # How many time periods
  nT <- length(tlist)
  # How many treated groups
  nG <- length(glist)

  # adding covariates
  if (!is.null(xformla)) {
    cleaned_data <- cbind(cleaned_data, stats::model.matrix(xformla,
                                                            data = dta,
                                                            na.action = na.pass))
  }

  # Remove collinear variables (following DRDID approach)
  # Determine the number of static columns (before covariates)
  # Static cols: id, y, first_treat, period, partition, weights, [cluster]
  idx_static_vars <- ifelse(is.null(cluster), 6, 7)

  # Convert the covariate columns (including intercept) to a matrix
  cov_m <- as.matrix(cleaned_data[, -c(1:idx_static_vars), with = FALSE])

  # Only check for collinearity if there are covariates beyond the intercept
  if (ncol(cov_m) > 1) {
    # Use the qr() function to detect collinear columns
    qr_m <- qr(cov_m, tol = 1e-6)
    # Get the rank of the matrix
    rank_m <- qr_m$rank
    # Get the indices of the non-collinear columns
    non_collinear_indices <- qr_m$pivot[seq_len(rank_m)]
    # Check if any covariates were dropped due to global collinearity
    dropped_covariates_global <- setdiff(colnames(cov_m), colnames(cov_m)[non_collinear_indices])
    if (length(dropped_covariates_global) > 0) {
      warning("The following covariates were dropped due to global collinearity: ", paste(dropped_covariates_global, collapse = ", "))
    }
    # Drop the collinear columns from the data.table
    cleaned_data <- cleaned_data[, c(seq(1, idx_static_vars, 1), non_collinear_indices + idx_static_vars), with = FALSE]
  }

  # drop the intercept
  cleaned_data[, "(Intercept)" := NULL]

  # Update xformula to reflect dropped covariates (following DRDID approach)
  # Get final covariate column names after collinearity check
  # Static cols in multi-period: id, y, first_treat, period, partition, weights, [cluster]
  final_cov_cols <- setdiff(names(cleaned_data), c("id", "y", "first_treat", "period", "partition", "weights", "cluster"))
  if (length(final_cov_cols) > 0) {
    xformla <- stats::as.formula(paste("~", paste(final_cov_cols, collapse = " + ")))
  } else {
    xformla <- ~1
  }

  # order dataset wrt idname and tname
  setorder(cleaned_data, "id", "period")

  out <- list(preprocessed_data = cleaned_data,
              xformula = xformla,
              tname = tname,
              est_method = est_method,
              boot = boot,
              nboot = nboot,
              cluster = cluster,
              cband = cband,
              alpha = alpha,
              use_parallel = use_parallel,
              cores = cores,
              control_group = control_group,
              base_period = base_period,
              n = n,
              cohorts = nG,
              time_periods = nT,
              cohort_size = gsize,
              tlist = tlist,
              glist = glist,
              panel = panel,
              allow_unbalanced_panel = allow_unbalanced_panel,
              true_repeated_cross_sections = true_repeated_cross_sections)

  return(out)
}

#' Process results inside att_gt_dr and att_gt_dml function
#' @param attgt_list A list of results from the `att_gt_dr` or `att_gt_dml` function.
#' @return A list with three vectors: `group`, `att`, and `periods` containing the group, average treatment effect, and time periods respectively.
#' @keywords internal
#' @export
process_attgt <- function(attgt_list){
  groups <- length(unique(unlist(BMisc::get_list_element(attgt_list, "group"))))
  time_periods <- length(unique(unlist(BMisc::get_list_element(attgt_list, "year"))))

  # empty vectors to hold results
  group <- c()
  att <- c()
  periods <- c()

  i <- 1
  # extract results and populate vectors
  for (g in 1:groups) {
    for (t in 1:time_periods) {
      group[i] <- attgt_list[[i]]$group
      periods[i] <- attgt_list[[i]]$year
      att[i] <- attgt_list[[i]]$att
      i <- i + 1
    }
  }

  return(list(group = group, att = att, periods = periods))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddd.R
\name{ddd}
\alias{ddd}
\title{Doubly Robust DDD estimators for the group-time average treatment effects.}
\usage{
ddd(
  yname,
  tname,
  idname = NULL,
  gname,
  pname,
  xformla,
  data,
  control_group = NULL,
  base_period = NULL,
  est_method = "dr",
  panel = TRUE,
  allow_unbalanced_panel = FALSE,
  weightsname = NULL,
  boot = FALSE,
  nboot = NULL,
  cluster = NULL,
  cband = FALSE,
  alpha = 0.05,
  use_parallel = FALSE,
  cores = 1,
  inffunc = FALSE,
  skip_data_checks = FALSE
)
}
\arguments{
\item{yname}{The name of the outcome variable.}

\item{tname}{The name of the column containing the time periods.}

\item{idname}{The name of the column containing the unit id.}

\item{gname}{The name of the column containing the first period when a particular observation is treated. It is a positive number
for treated units and defines which group the unit belongs to. It takes value 0 or Inf for untreated units.}

\item{pname}{The name of the column containing the partition variable (e.g., the subgroup identifier). This is an indicator variable that is 1 for
the units eligible for treatment and 0 otherwise.}

\item{xformla}{The formula for the covariates to be included in the model. It should be of the form \code{~ x1 + x2}.
Default is \code{xformla = ~1} (no covariates).}

\item{data}{A data frame or data table containing the data.}

\item{control_group}{Valid for multiple periods only. The control group to be used in the estimation. Default is \code{control_group = "notyettreated"} which sets as control group the units that have not yet participated in the treatment.
The alternative is \code{control_group = "nevertreated"} which sets as control group the units that never participate in the treatment and does not change across groups or time periods.}

\item{base_period}{Valid for multiple periods. Choose between a "varying" or "universal" base period. Both yield the same post-treatment ATT(g,t) estimates.
Varying base period: Computes pseudo-ATT in pre-treatment periods by comparing outcome changes for a group to its comparison group from t-1 to t, repeatedly changing t.
Universal base period: Fixes the base period to (g-1), reporting average changes from t to (g-1) for a group relative to its comparison group, similar to event study regressions.
Varying base period reports ATT(g,t) right before treatment. Universal base period normalizes the estimate before treatment to be 0, adding one extra estimate in an earlier period.}

\item{est_method}{The estimation method to be used. Default is \code{"dr"} (doubly robust). It computes propensity score using logistic regression
and outcome regression using OLS. The alternative are \code{c("reg", "ipw")}.}

\item{panel}{Logical. If \code{TRUE} (default), the data is treated as panel data where each unit is observed in all time periods.
If \code{FALSE}, the data is treated as repeated cross-sections (RCS) where each observation may represent a different unit.
For RCS data, \code{idname} can be omitted or set to \code{NULL}, and the function will automatically create unique IDs for each observation.}

\item{allow_unbalanced_panel}{Logical. If \code{TRUE}, allows for unbalanced panel data where units may not be observed in all time periods.
Default is \code{FALSE}. Note: This parameter requires \code{panel = TRUE} and a valid \code{idname}.}

\item{weightsname}{The name of the column containing the weights. Default is \code{NULL}. As part of data processing, weights are enforced to be normalized
and have mean 1 across all observations.}

\item{boot}{Logical. If \code{TRUE}, the function computes standard errors using the multiplier bootstrap. Default is \code{FALSE}.}

\item{nboot}{The number of bootstrap samples to be used. Default is \code{NULL}. If \code{boot = TRUE}, the default is \code{nboot = 999}.}

\item{cluster}{The name of the variable to be used for clustering. The maximum number of cluster variables is 1. Default is \code{NULL}.
If \code{boot = TRUE}, the function computes the bootstrap standard errors clustering at the unit level setting as cluster variable the one in \code{idname}.}

\item{cband}{Logical. If \code{TRUE}, the function computes a uniform confidence band that covers all of the average treatment effects
with fixed probability \code{1-alpha}.  In order to compute uniform confidence bands, \code{boot} must also be set to \code{TRUE}.  The default is \code{FALSE}.}

\item{alpha}{The level of significance for the confidence intervals.  Default is \code{0.05}.}

\item{use_parallel}{Logical. If \code{TRUE}, the function runs in parallel processing. Valid only when \code{boot = TRUE}. Default is \code{FALSE}.}

\item{cores}{The number of cores to be used in the parallel processing. Default is \code{cores = 1}.}

\item{inffunc}{Logical. If \code{TRUE}, the function returns the influence function. Default is \code{FALSE}.}

\item{skip_data_checks}{Logical. If \code{TRUE}, the function skips data validation checks and proceeds directly to estimation.
This can improve performance when you are confident the data is correctly formatted. Default is \code{FALSE}.
Use with caution as skipping checks may lead to unexpected errors if data is malformed.}
}
\value{
A \code{ddd} object with the following basic elements:
\item{ATT}{The average treatment effect on the treated.}
\item{se}{The standard error of the ATT.}
\item{uci}{The upper confidence interval of the ATT.}
\item{lci}{The lower confidence interval of the ATT.}
\item{inf_func}{The estimate of the influence function.}
}
\description{
\code{ddd} is the main function for computing the Doubly Robust DDD estimators for the ATT, with balanced panel data.
It can be used with covariates and/or under multiple time periods. At its core, \code{triplediff} employs
the doubly robust estimator for the ATT, which is a combination of the propensity score weighting and the outcome regression.
Furthermore, this package supports the application of machine learning methods for the estimation of the nuisance parameters.
}
\examples{
#----------------------------------------------------------
# Triple Diff with covariates and 2 time periods
#----------------------------------------------------------
set.seed(1234) # Set seed for reproducibility
# Simulate data for a two-periods DDD setup
df <- gen_dgp_2periods(size = 5000, dgp_type = 1)$data

head(df)

att_22 <- ddd(yname = "y", tname = "time", idname = "id", gname = "state",
              pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
             data = df, control_group = "nevertreated", est_method = "dr")

summary(att_22)

#----------------------------------------------------------
# Triple Diff with multiple time periods
#----------------------------------------------------------
data <- gen_dgp_mult_periods(size = 1000, dgp_type = 1)[["data"]]

ddd(yname = "y", tname = "time", idname = "id",
     gname = "state", pname = "partition", xformla = ~cov1 + cov2 + cov3 + cov4,
     data = data, control_group = "nevertreated", base_period = "varying",
     est_method = "dr")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate_data.R
\name{gen_dgp_2periods}
\alias{gen_dgp_2periods}
\title{Function that generates panel data with single treatment date assignment and two time periods.}
\usage{
gen_dgp_2periods(size, dgp_type)
}
\arguments{
\item{size}{Integer. Number of units.}

\item{dgp_type}{Integer in \{1,2,3,4\}.
1 = both nuisance functions correct;
2 = only the outcome model correct;
3 = only the propensity score correct;
4 = both nuisance functions incorrect.}
}
\value{
A list with the following elements:
\describe{
\item{data}{A \code{data.table} in long format with columns:
\itemize{
\item \code{id}: unit identifier
\item \code{state}: state variable
\item \code{time}: time variable
\item \code{partition}: partition assignment
\item \code{x1}, \code{x2}, \code{x3}, \code{x4}: covariates
\item \code{y}: outcome variable
\item \code{cluster}: cluster ID (no within-cluster correlation)
}
}
\item{att}{True average treatment effect on the treated (ATT), set to 0.}
\item{att.unf}{Oracle ATT computed under the unfeasible specification.}
\item{eff}{Theoretical efficiency bound for the estimator.}
}
}
\description{
Generate panel data with a single treatment date and two periods
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compute_aggregations.R
\name{compute_se_agg}
\alias{compute_se_agg}
\title{Take influence function and compute standard errors}
\usage{
compute_se_agg(
  influence_function,
  boot = FALSE,
  boot_std_errors = NA,
  cluster_vec = NULL
)
}
\arguments{
\item{influence_function}{An influence function}

\item{boot}{a boolean indicating whether bootstrapping was performed}

\item{boot_std_errors}{a vector of bootstrapped standard errors}
}
\value{
scalar standard error
}
\description{
Function to take an nx1 influence function and return
a standard error
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate_data.R
\name{generate_test_panel}
\alias{generate_test_panel}
\title{Function to generate a fake dataset for testing purposes only.}
\usage{
generate_test_panel(
  seed = 123,
  num_ids = 100,
  time = 2,
  initial.year = 2019,
  treatment.year = 2020
)
}
\arguments{
\item{seed}{Seed for reproducibility}

\item{num_ids}{Number of IDs}

\item{time}{Number of time periods}

\item{initial.year}{Initial year}

\item{treatment.year}{Treatment year}
}
\value{
A data.table with the following columns:
\itemize{
\item id: ID
\item state: State variable
\item year: Time variable
\item partition: Partition variable
\item x1: Covariate 1
\item x2: Covariate 2
\item treat: Treatment variable
\item outcome: Outcome variable
}
}
\description{
Function to generate fake dataset to test internal procedures.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mboot.R
\name{mboot}
\alias{mboot}
\title{Multiplier Bootstrap}
\usage{
mboot(inf_func, did_preprocessed, use_parallel = FALSE, cores = 1)
}
\arguments{
\item{inf_func}{an influence function}

\item{did_preprocessed}{A \code{dp} object obtained after preprocess}

\item{use_parallel}{Boolean of whether or not to use parallel processing in the multiplier
bootstrap, default is \code{use_parallel=FALSE}}

\item{cores}{the number of cores to use with parallel processing, default is \code{cores=1}}
}
\value{
A list with following elements
\item{bres}{results from each bootstrap iteration.}
\item{V}{variance matrix.}
\item{se}{standard errors.}
\item{crit_val}{a critical value for computing uniform confidence bands.}
}
\description{
This function take an influence function and use the
multiplier bootstrap to compute standard errors and critical values for
uniform confidence bands.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compute_aggregations.R
\name{compute_aggregation}
\alias{compute_aggregation}
\title{Compute Aggregated Treatment Effect Parameters}
\usage{
compute_aggregation(
  ddd_obj,
  type = "simple",
  cluster = NULL,
  balance_e = NULL,
  min_e = -Inf,
  max_e = Inf,
  na.rm = FALSE,
  boot = FALSE,
  nboot = NULL,
  cband = NULL,
  alpha = 0.05
)
}
\arguments{
\item{ddd_obj}{a ddd object (i.e., the results of the \code{\link[=ddd]{ddd()}} function)}

\item{type}{Which type of aggregated treatment effect parameter to compute.
\code{"simple"} just computes a weighted average of all
group-time average treatment effects with weights proportional to group
size.
\code{"eventstudy"} computes average effects across
different lengths of exposure to the treatment (event times). Here the overall effect averages the effect of the
treatment across the positive lengths of exposure. This is the default option;
\code{"group"} computes average treatment effects across different groups/cohorts; here
the overall effect averages the effect across different groups using group size as weights;
\code{"calendar"} computes average treatment effects across different
time periods, with weights proportional to the group size; here the overall effect averages the effect across each
time period.}

\item{cluster}{The name of the variable to be used for clustering. The maximum number of cluster variables is 1. Default is \code{NULL}.}

\item{balance_e}{If set (and if one computes event study), it balances
the sample with respect to event time.  For example, if \code{balance_e=2},
\code{agg_ddd} will drop groups that are not exposed to treatment for
at least three periods, the initial period \code{e=0} as well as the
next two periods, \code{e=1} and \code{e=2}.  This ensures that
the composition of groups does not change when event time changes.}

\item{min_e}{For event studies, this is the smallest event time to compute
dynamic effects for.  By default, \code{min_e = -Inf} so that effects at
all lengths of exposure are computed.}

\item{max_e}{For event studies, this is the largest event time to compute
dynamic effects for.  By default, \code{max_e = Inf} so that effects at
all lengths of exposure are computed.}

\item{na.rm}{Logical value if we are to remove missing Values from analyses. Defaults is FALSE.}

\item{boot}{Boolean for whether or not to compute standard errors using
the multiplier bootstrap.  If standard errors are clustered, then one
must set \code{boot=TRUE}. Default is value set in the ddd object.  If \code{boot = FALSE}, then analytical
standard errors are reported.}

\item{nboot}{The number of bootstrap iterations to use.  The default is the value set in the ddd object,
and this is only applicable if \code{boot=TRUE}.}

\item{cband}{Boolean for whether or not to compute a uniform confidence
band that covers all of the group-time average treatment effects
with fixed probability \code{1 - alpha}.  In order to compute uniform confidence
bands, \code{boot} must also be set to \code{TRUE}.  The default is
the value set in the ddd object}

\item{alpha}{The level of confidence for the confidence intervals.  The default is 0.05. Otherwise, it will
use the value set in the ddd object.}
}
\value{
Aggregation object (list) of class \code{\link{agg_ddd}}
}
\description{
Does the heavy lifting on computing aggregated group-time
average treatment effects
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate_data.R
\name{gen_dgp_mult_periods}
\alias{gen_dgp_mult_periods}
\title{Generate panel data with staggered treatment adoption (three periods)}
\usage{
gen_dgp_mult_periods(size, dgp_type = 1, include_covariates = TRUE)
}
\arguments{
\item{size}{Integer. Number of units to simulate.}

\item{dgp_type}{Integer in \{1,2,3,4\}. Only used when \code{include_covariates = TRUE}.
1 = both nuisance functions correct (default);
2 = only the outcome model correct;
3 = only the propensity-score model correct;
4 = both nuisance functions misspecified.}

\item{include_covariates}{Logical. If \code{TRUE} (default), generates covariates with
transformations and uses \code{dgp_type} specification. If \code{FALSE}, uses constant
covariates and fixed propensity score probabilities for a simpler DGP.}
}
\value{
A named list with components:
\describe{
\item{data}{A \code{data.table} in long format with columns:
\itemize{
\item \code{id}: unit identifier
\item \code{cohort}: first period when treatment is assigned
\item \code{partition}: partition indicator
\item \code{x1}, \code{x2}, \code{x3}, \code{x4}: covariates
\item \code{cluster}: cluster identifier (no within-cluster correlation)
\item \code{time}: time period index
\item \code{y}: observed outcome
}
}
\item{data_wide}{A \code{data.table} in wide format (one row per \code{id}) with columns:
\itemize{
\item \code{id}, \code{cohort}, \code{partition}, \code{x1}, \code{x2}, \code{x3}, \code{x4}, \code{cluster}
\item \code{y_t0}, \code{y_t1}, \code{y_t2}: outcomes in periods 0, 1, and 2
}
}
\item{ES_0_unf}{(Only if \code{include_covariates = TRUE}) Unfeasible (oracle) event-study parameter at time 0.}
\item{prob_g2_p1}{(Only if \code{include_covariates = TRUE}) Proportion of units with \code{cohort == 2} and eligibility in period 1.}
\item{prob_g3_p1}{(Only if \code{include_covariates = TRUE}) Proportion of units with \code{cohort == 3} and eligibility in period 1.}
}
}
\description{
Generate panel data where units adopt treatment at different times across three periods.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/att_dr.R
\name{att_dr}
\alias{att_dr}
\title{Doubly robust DDD estimator for ATT, with panel data and 2 periods}
\usage{
att_dr(did_preprocessed)
}
\arguments{
\item{did_preprocessed}{A list containing preprocessed data and specifications for the DDD estimation.
Expected elements include:
- \code{preprocessed_data}: A data table containing the data with variables needed for the analysis.
- \code{est_method}: The estimation method to be used. Default is \code{est_method = "dr"}.
- \code{xformula}: The formula for the covariates to be included in the model. It should be of the form \code{~ x1 + x2}.
Default is \code{xformla = ~1} (no covariates).
- \code{boot}: Logical. If \code{TRUE}, the function use the multiplier bootstrap to compute standard errors. Default is \code{FALSE}.
- \code{nboot}: The number of bootstrap samples to be used. Default is \code{NULL}. If \code{boot = TRUE}, the default is \code{nboot = 999}.
- \code{subgroup_counts}: A matrix containing the number of observations in each subgroup.
- \code{alpha} The level of significance for the confidence intervals.  Default is \code{0.05}.
- \code{inffunc}: Logical. If \code{TRUE}, the function returns the influence function. Default is \code{FALSE}.
- \code{use_parallel}: Boolean of whether or not to use parallel processing in the multiplier bootstrap, default is \code{use_parallel=FALSE}
- \code{cores}: the number of cores to use with parallel processing, default is \code{cores=1}
- \code{cband}: Boolean of whether or not to compute simultaneous confidence bands, default is \code{cband=FALSE}}
}
\value{
A list with the estimated ATT, standard error, upper and lower confidence intervals, and influence function.
}
\description{
This function implements a doubly robust estimator for assessing the average
treatment effect on the treated (ATT) using a triple differences (DDD) approach
in panel data settings across two time periods. The function takes preprocessed
data structured specifically for this analysis.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocess.R
\name{process_attgt}
\alias{process_attgt}
\title{Process results inside att_gt_dr and att_gt_dml function}
\usage{
process_attgt(attgt_list)
}
\arguments{
\item{attgt_list}{A list of results from the \code{att_gt_dr} or \code{att_gt_dml} function.}
}
\value{
A list with three vectors: \code{group}, \code{att}, and \code{periods} containing the group, average treatment effect, and time periods respectively.
}
\description{
Process results inside att_gt_dr and att_gt_dml function
}
\keyword{internal}
